/*=================================================================================  
 * calculator.y
 * Latest revision: 10-24-1998                                              
 * Copyright (C) 1998 by Thomas J. Nelson                                  
 * All rights reserved.                                                     
 * 
 * Each rule must start in column 1
 * To compile:  make calculator
 *     yacc -d calculator.y    
 *
 *    *** NOTE: *** This produces y.tab.c
 *     while bison -d calculator.y produces calculator.tab.c
 *     The makefile must use the correct c file or you will get
 *     unexplained crashes
 *
 *     flex calculator.l 
 *     cc -o calculator y.tab.c lex.yy.c -ll -lm
 *=================================================================================
 * After running flex, it is necessary to edit lex.yy.c:
 *     remove static from    static char yy_hold_char;              (line 211)
 *     remove static from    static char yy_n_chars;                (line 213)
 *     remove static from    static char *yy_c_buf_p                (line 219)
 *     remove static from    static int yyinput YY_PROTO(( void )); (line 447)
 *     remove static from    static int input YY_PROTO(( void ));   (line 449)
 *     remove static from    static int yyinput()                   (line 1169)
 *     remove static from    static int input()                     (line 1170)
 *
 *    At the macro starting at line 505, i.e. from:
 *
 *          #ifndef YY_INPUT
 *          #define YY_INPUT(buf,result,max_size) \
 *          . . .
 *          #endif
 * 
 *    the entire macro must be replaced with:
 *
 *          #ifndef YY_INPUT
 *          #define YY_INPUT(buf,result,max_size) \
 *              strncpy(buf,text,max_size); \
 *              result = strlen(buf);
 *          #endif
 *
 *    Add these lines at the top of the file after line 20:
 *
 *          extern char *text;
 *          int n; 
 *          int input(void);
 *          struct symtab* symlook(char *s);
 *
 *    Also, remove any exit()'s from line 1562 in the function:
 *
 *        static void yy_fatal_error( yyconst char msg[] )
 *          exit( YY_EXIT_FAILURE );
 *
 *    Remove -lfl from makefile to get rid of yylex undefined reference
 *    which is a bug in Bison
 *
 *=================================================================================
 * For false condition, characters are thrown away until desired stop point:
 *          while(yy_hold_char != ';') yyinput();
 *          yyinput();
 *     For true condition, set if_true to 1.
 *
 * yy_n_chars = no of chars in current evaluated string          
 * yyleng = no of chars in current token                         
 * yytext = current token                                        
 * Use input() for C, yyinput() for C++                          
 *   throws away next token & prepares for more input             
 * 
 * These 3 lines would throw away the rest of the current string 
 *   which is not what we want. (The yy_buffer_state typedef struct must also
 *   be included for the following line to work.)
 *          len = yy_n_chars - strlen(yy_current_buffer->yy_ch_buf) - 1;
 *          yy_c_buf_p += len;                                          
 *          yyinput();                                                    
 *
 * yy_hold_char is current character being input.
 *
 * The input string must end with \n\255 to parse correctly.  
 *================================================================================
 */

%{
#include<stdio.h>
#include<string.h>
#include<stdlib.h> 
#include<math.h>
#include "calculator.h"
#include "xdata.h"

void addfunc4(char *name, double(*func)(double, double, double, double));
void addfunc3(char *name, double(*func)(double, double, double));
void addfunc2(char *name, double(*func)(double,double));
void addfunc(char *name, double(*func)(double));
double drand(double a); 
double series(double x, double a, double b);
void skip_to_end(void);
double dmax(double a, double b);
double dmin(double a, double b);
struct symtab* symlook(char *s);
double int2(double x);
int input(void);
int yyinput(void);
extern int yylex(void);
void yyerror(char *s);
void yyrestart ( FILE *input_file );
enum{ N_STRING, N_NUMBER, N_VARIABLE, N_FUNCTION };

extern Globals     g;
extern Spreadsheet s;
extern double answer;
extern char yy_hold_char, yy_n_chars;
extern char *text;
extern char *yytext;

extern char *yy_c_buf_p;
extern int yy_init;          /* whether we need to initialize */
extern int yy_start;         /* start state number */

int if_true=0;
int need_refresh=0;
int lineno, k, row, col;
static char tempstring[128];
%}

%union {  double dval;
          struct symtab *symp;
       }

/*  tokens  */ 
%token <symp> NAME
%token <dval> NUMBER
%token Y X COL
%token IF THEN ELSE
%token EQ LE GE
%token PLUS_EQUALS MINUS_EQUALS TIMES_EQUALS DIVIDE_EQUALS
%token LOGICAL_AND LOGICAL_OR INCREMENT DECREMENT

/* left-associative operatore - Also assigns precedence, last %left statement 
    has the highest precedence.
*/

%left LOGICAL_AND LOGICAL_OR
%left '&' '|' '^' '%'
%left PLUS_EQUALS MINUS_EQUALS TIMES_EQUALS DIVIDE_EQUALS
%left '<' '>' EQ GE LE
%left '-' '+'
%left '*' '/' 
%left INCREMENT DECREMENT Y X COL
%nonassoc UMINUS '!' '~'
%type <dval> expression

%%

statement_list: 
        statement 
        |   statement_list statement
        |   '{' statement_list '}' statement
        ;
        
statement: 
        NAME '=' expression {  $1->value = $3; } ';'
        |   COL '=' expression    /* set y = expression, already corr. for start=1 */
                {  s.x[g.cc][g.rr] = $3;
                   need_refresh = 1;
                }';'
        |   Y '=' expression    /* set y[col] = expression, already corr. for start=1 */
                {  s.x[g.cc][g.rr] = $3;
                   need_refresh = 1;
                }';'

        |   NAME PLUS_EQUALS expression {  $1->value += $3; } ';'
        |   NAME MINUS_EQUALS expression { $1->value -= $3; }';'
        |   NAME TIMES_EQUALS expression { $1->value *= $3; }';'
        |   NAME DIVIDE_EQUALS expression
                        { if ($3 == 0.0)
                              yyerror((char*)"Division by zero");
                          else
                              $1->value /= $3;
                        }  ';'
        |   NAME INCREMENT { $1->value++; } ';'
        |   NAME DECREMENT { $1->value--; } ';'
        |   expression   { /* printf("= %g\n", $1); */ answer=$1; } ';'
        |   '\n'
        |   ';'
        |   IF '(' expression ')'
                {
                   if( $3 == 1)  if_true=1;
                               
                   else        {  if_true=0;
                                  skip_to_end();
                               }
                }
        |   ELSE
                {  
                   if(if_true) skip_to_end();
                } 
                
        /* Repeat above statement definitions terminated with \n */
                
        |   NAME '=' expression {  $1->value = $3; } '\n'
        |   NAME PLUS_EQUALS expression {  $1->value += $3; } '\n'
        |   NAME MINUS_EQUALS expression { $1->value -= $3; }'\n'
        |   NAME TIMES_EQUALS expression { $1->value *= $3; }'\n'
        |   NAME DIVIDE_EQUALS expression
                        { if ($3 == 0.0)
                              yyerror((char*)"Division by zero");
                          else
                              $1->value /= $3;
                        }  '\n'
        |   NAME INCREMENT { $1->value++; } '\n'
        |   NAME DECREMENT { $1->value--; } '\n'
        |   expression   { /* printf("= %g\n", $1);*/ answer=$1; } '\n'
        ;




expression: 
        expression '+' expression { $$ = $1 + $3; }
        |   expression '-' expression { $$ = $1 - $3; }
        |   expression '*' expression { $$ = $1 * $3; }
        |   expression '/' expression 
                        { if ($3 == 0.0)
                              yyerror((char*)"Division by zero");
                          else
                              $$ = $1 / $3; 
                        }
        |   '-' expression %prec UMINUS { $$ = -$2; }
        |   '!' expression %prec '!' { $$ = !$2; }
        |   '~' expression %prec '~' { $$ = (double)(~(int)$2); }
        |   '(' expression ')'        { $$ = $2; }
        |   X '[' expression ']' '[' expression ']'
            {   if( ((int)$3 >= 0) &&  ((int)$6 >= 0) )
                   $$ = s.x[(int)$3][(int)$6]; }
        |   X '[' expression  expression ']'{ $$ = s.x[(int)$3][(int)$4]; }
        |   expression '<' expression { $$ = $1 < $3; }
        |   expression '>' expression { $$ = $1 > $3; }
        |   expression LE expression { $$ = $1 <= $3; }
        |   expression GE expression { $$ = $1 >= $3; }
        |   expression EQ expression { $$ = $1 == $3; }
        |   expression LOGICAL_AND expression { $$ = $1 && $3; }
        |   expression LOGICAL_OR expression { $$ = $1 || $3; }
        |   expression '&' expression { $$ = (double)((int)$1 & (int)$3); }
        |   expression '|' expression { $$ = (double)((int)$1 | (int)$3); }
        |   expression '^' expression { $$ = (double)(pow($1,$3)); }
        |   expression '%' expression   /* modulus */
            {  if((int)$3)
                  $$ = ((double)(((int)$1) % ((int)$3))); 
               else
                  $$ = 0;
            }
        |   COL {  $$ = s.x[g.cc][g.rr]; }
        |   X {  $$ = g.rr+1; }  /* numbering starts at 1 in spreadsheet */
        |   Y {  $$ = s.x[g.cc][g.rr]; }
        |   NUMBER                   
        |   NAME {  $$ = $1->value;  }
        |   NAME '(' expression ')' 
                 {  
                    if($1->funcptr)
                       $$ = ($1->funcptr)($3);
                    else
                    { 
                       sprintf(tempstring, "%s is not a 1-argument function\n", $1->name);
                       yyerror(tempstring);
                       $$ = 0.0;
                    }
                 }
        |   NAME '(' expression ',' expression ')' 
                 {  
                    if($1->funcptr2)
                       $$ = ($1->funcptr2)($3,$5);
                    else
                    {  sprintf(tempstring, "%s is not a 2-argument function\n", $1->name);
                       yyerror(tempstring);
                       $$ = 0.0;
                    }
                 }
        |   NAME '(' expression ',' expression ',' expression ')' 
                 {  
                    if($1->funcptr3)
                       $$ = ($1->funcptr3)($3,$5,$7);
                    else
                    {  sprintf(tempstring, "%s is not a 3-argument function\n", $1->name);
                       yyerror(tempstring);
                       $$ = 0.0;
                    }
                 }
        |   NAME '(' X '[' expression ']' '[' expression ']' ':' X '[' expression ']' '[' expression ']' ')' 
                 {  
                    if($1->funcptr4)
                       $$ = ($1->funcptr4)($5,$8,$13,$16);
                    else
                    {  sprintf(tempstring, "%s is not a 4-argument function\n", $1->name);
                       yyerror(tempstring);
                       $$ = 0.0;
                    }
                 }
        |   NAME '(' expression  expression  ':'  expression  expression  ')' 
                 {  
                    if($1->funcptr4)
                       $$ = ($1->funcptr4)($3,$4,$6,$7);
                    else
                    {  sprintf(tempstring, "%s is not a 4-argument function\n", $1->name);
                       yyerror(tempstring);
                       $$ = 0.0;
                    }
                 }
        ;
%%


void initialize_evaluator(void)
{   
  
   /* 1-parameter functions - new functions can be added here */
   addfunc((char*)"abs",  fabs); 
   addfunc((char*)"acos", acos); 
   addfunc((char*)"asin", asin); 
   addfunc((char*)"atan", atan); 
   addfunc((char*)"cos",  cos); 
   addfunc((char*)"cosh", cosh); 
   addfunc((char*)"exp",  exp);
   addfunc((char*)"int",  int2); 
   addfunc((char*)"log",  log); 
   addfunc((char*)"log10",log10); 
   addfunc((char*)"rand", randomnumber); 
   addfunc((char*)"sin",  sin); 
   addfunc((char*)"sinh", sinh); 
   addfunc((char*)"sqrt", sqrt);
   addfunc((char*)"tan",  tan); 
   addfunc((char*)"tanh", tanh); 
   addfunc((char*)"erf",  erf); 
   addfunc((char*)"fact", factorial); 
   addfunc((char*)"random", randomnumber); 

#ifdef HAVE_ACOSH
   addfunc((char*)"acosh",acosh); 
#endif
#ifdef HAVE_CBRT
   addfunc((char*)"cbrt", cbrt); 
#endif

   /* 2-parameter functions - new functions can be added here */
   addfunc2((char*)"pow", pow); 
   addfunc2((char*)"min", dmin); 
   addfunc2((char*)"max", dmax); 
   addfunc2((char*)"binom", binom); 
   addfunc2((char*)"binomial", binom); 

   /* 3-parameter functions - new functions can be added here */
   addfunc3((char*)"series", series); 

   /* 4-parameter functions - new functions can be added here */
   addfunc4((char*)"sum", sum); 
   addfunc4((char*)"sumsq", sumsq); 

   lineno=1;
   yy_n_chars = 0;
}


void yyerror(char *s)
{  
   formula_error(s); 
   yyrestart(NULL);
}

void add_variable(char *name, double value)
{
   struct symtab *sp = symlook(name);
   sp->funcptr = NULL;
   sp->funcptr2 = NULL;
   sp->value = value;
   sp->type = N_VARIABLE;
}

void addfunc(char *name, double(*func)(double))
{
   struct symtab *sp = symlook(name);
   sp->funcptr = func;
   sp->funcptr2 = NULL;
   sp->funcptr3 = NULL;
   sp->funcptr4 = NULL;
   sp->type = N_FUNCTION;
}

void addfunc2(char *name, double(*func)(double, double))
{
   struct symtab *sp = symlook(name);
   sp->funcptr = NULL;
   sp->funcptr2 = func;
   sp->funcptr3 = NULL;
   sp->funcptr4 = NULL;
   sp->type = N_FUNCTION;
}

void addfunc3(char *name, double(*func)(double, double, double))
{
   struct symtab *sp = symlook(name);
   sp->funcptr = NULL;
   sp->funcptr2 = NULL;
   sp->funcptr3 = func;
   sp->funcptr4 = NULL;
   sp->type = N_FUNCTION;
}

void addfunc4(char *name, double(*func)(double, double, double, double))
{
   struct symtab *sp = symlook(name);
   sp->funcptr = NULL;
   sp->funcptr2 = NULL;
   sp->funcptr3 = NULL;
   sp->funcptr4 = func;
   sp->type = N_FUNCTION;
}


double read_variable(char *name)
{
   struct symtab *sp = symlook(name);
   return sp->value;
}


struct symtab* symlook(char *s)
{
   struct symtab *sp;
   for(sp=symtab; sp< &symtab[NSYMS]; sp++)
   {  
       if(sp->name && !strcmp(sp->name, s)) return sp;  /* is it already here */
#ifdef LINUX
       if(!sp->name){ sp->name=strdup(s); return sp; }  /* is it free         */
#else
       if(!sp->name)
       {   
           sp->name = (char*)malloc(strlen(s)+1);
           strcpy(sp->name, s); 
           return sp; 
       }  
#endif
   }
   yyerror((char*)"Too many symbols, ask a wizard to enlarge me");
   return NULL;
}



void skip_to_end(void)
{
   while(strchr("\n ",yy_hold_char)!=NULL) yyinput(); 
   int got_left_brace = yy_hold_char=='{';
   if(got_left_brace)
   {   while(yy_hold_char != '}') yyinput();
   }else
   {   while(strchr(";}\n",yy_hold_char)==NULL) yyinput(); 
   }
   yyinput();
   while(yy_hold_char=='\n') yyinput(); 
}

double dmax(double a, double b){ if(a>b) return a; return b; }
double dmin(double a, double b){ if(a<b) return a; return b; }
double drand(double a){ a=a; return rand(); }
double series(double x, double a, double b)
{
   return gmp_series_term(x, (int)a, (int)b );
}
double int2(double x) { return (double)(int)(x); }
