//--------------------------------------------------------------------//
// xdata.cc                                                           //
// Latest revision: 07-31-2023                                        //
//--------------------------------------------------------------------//
 
#include "xdata.h"

Globals g;
Spreadsheet s;

////  Don't set `transient' resource in fallbacks.
static const char *fallback_resources[] = 
{    
   "*.fontList: *helvetica-bold-r-normal--12*",
   "*.shadowThickness: 1",
   "*.DialogForm*shadowThickness: 1",
   "*.background:gray",   
   "*.foreground:black",
   "*.marginWidth:1",
   "*.marginHeight:1",

   "*.FileSelector*marginWidth: 1",
   "*.FileSelector*marginHeight: 1",
   "*.FileSelector*listVisibleItemCount: 25",
   "*.FileSelector*borderWidth: 0",
   "*.XmText*fontList: fixed",

   "*f214*acceleratorText: Ctrl+R",
   "*f14*acceleratorText: Ctrl+Q",
   "*f2*acceleratorText: Ctrl+S",
   "*f20a*acceleratorText: Ctrl+T",
   "*f20d*acceleratorText: Ctrl+E",
   NULL
};

   
//--------------------------------------------------------------------//

int main(int argc, char **argv)
{
   initialize_globals();
   XtAppContext app; 
   XtSetLanguageProc(NULL,NULL,NULL);
   g.main_widget = XtVaAppInitialize(&app, "Xdata", 
                        NULL, 0, &argc, argv, (char**)fallback_resources,  
                        XmNdeleteResponse, XmDO_NOTHING,
                        NULL);
   init_xlib(); 
   resize_main_widget(g.main_xsize, g.main_ysize);
   makewidgets(g.main_widget);
   xminitializemisc(g.main_widget);
   create_main_icon();
   create_spreadsheet(g.main_widget);
   if(argc>1) read(argv, argc);
   XtAppAddTimeOut(app,1000,timer_callback,NULL);
   XtAppMainLoop(app);
}



//--------------------------------------------------------------------//
// create_main_icon                                                   //
// Set the icon in case user iconifies program. Need a bigger one.    //
//--------------------------------------------------------------------//
void create_main_icon(void)
{
   Window win;
   XIconSize *size;
   int count, k, ok=0;
   int alloc1=0, n;
   Arg args[100];
   Pixmap icon;
   const uchar icon_data[] = {
   0x54, 0x01, 0xa8, 0x02, 0x50, 0x05, 0xa0, 0x0a, 
   0x50, 0x15, 0xa8, 0x2a, 0x14, 0x55, 0x0a, 0xaa, 
   0x05, 0x55, 0x82, 0x2a, 0x45, 0x15, 0xaa, 0x0a,
   0x54, 0x05, 0xa8, 0x02, 0x54, 0x01, 0xaa, 0x00,

   0x54, 0x01, 0xa8, 0x02, 0x50, 0x05, 0xa0, 0x0a, 
   0x50, 0x15, 0xa8, 0x2a, 0x14, 0x55, 0x0a, 0xaa, 
   0x05, 0x55, 0x82, 0x2a, 0x45, 0x15, 0xaa, 0x0a,
   0x54, 0x05, 0xa8, 0x02, 0x54, 0x01, 0xaa, 0x00,

   0x54, 0x01, 0xa8, 0x02, 0x50, 0x05, 0xa0, 0x0a, 
   0x50, 0x15, 0xa8, 0x2a, 0x14, 0x55, 0x0a, 0xaa, 
   0x05, 0x55, 0x82, 0x2a, 0x45, 0x15, 0xaa, 0x0a,
   0x54, 0x05, 0xa8, 0x02, 0x54, 0x01, 0xaa, 0x00,

   0x54, 0x01, 0xa8, 0x02, 0x50, 0x05, 0xa0, 0x0a, 
   0x50, 0x15, 0xa8, 0x2a, 0x14, 0x55, 0x0a, 0xaa, 
   0x05, 0x55, 0x82, 0x2a, 0x45, 0x15, 0xaa, 0x0a,
   0x54, 0x05, 0xa8, 0x02, 0x54, 0x01, 0xaa, 0x00,

   };

   win = RootWindow(g.display,DefaultScreen(g.display));

   ////  Make sure it is ok to create a 32x32 icon. XGetIconSizes returns 0
   ////  if preferred icon sizes are not set. Should ideally iterate through 
   ////  size[k].width_inc and size[k].height_inc.

   if(XGetIconSizes(g.display, win, &size, &count))
   {   alloc1=1;
       for(k=0;k<count;k++)  
       {    if(between(32, size[k].min_width, size[k].max_width) &&
               between(32, size[k].min_height, size[k].max_height)){ ok=1; break;}
       }
   }else ok=1;
   if(ok)
   {    icon = XCreateBitmapFromData(g.display, win, (char*)icon_data, 32, 32);
        n=0;
        if(icon != (Pixmap)None)
        {    XtSetArg(args[n], XmNiconic, False); n++;
             XtSetArg(args[n], XmNiconName, "Xdata"); n++;
             XtSetArg(args[n], XmNiconPixmap, icon); n++;
        }else fprintf(stderr, "Can't create icon\n");
        XtSetValues(g.main_widget, args, n);
   }else fprintf(stderr, "Not allowed to create a 32x32 icon!\n");
   if(alloc1) XFree(size);
}



//--------------------------------------------------------------------//
// resize_main_widget                                                 //
//--------------------------------------------------------------------//
void resize_main_widget(int xsize, int ysize)
{
  int n;
  Arg args[100];
  n=0;
  XtSetArg(args[n], XmNwidth, xsize); n++;
  XtSetArg(args[n], XmNheight, ysize); n++;
  XtSetValues(g.main_widget, args, n);
}


//--------------------------------------------------------------------//
// Make widgets                                                       //
// Call after set_motif_colors and before setup_colors.               //
// Widgets that exist for the life of the application.                //
//--------------------------------------------------------------------//
void makewidgets(Widget parent)
{ 
  int n;
  Arg args[100];
  Widget  menupane, main_window;

  n=0;
  XtSetArg(args[n], XmNmarginHeight, 0); n++;
  XtSetArg(args[n], XmNmarginWidth, 0); n++;

  main_window = XmCreateMainWindow(parent, (char*)"Main", args, n);

  n=0;
  XtSetArg(args[n], XmNmarginHeight, 0); n++;
  XtSetArg(args[n], XmNmarginWidth, 0); n++;
  g.menubar = XmCreateMenuBar(main_window, (char*)"MenuBar", args, n);

  //------Create 'File' pull-down menu---------------//

  g.menupane[0] = menupane = createmenu(g.menubar, (char*)"fileMenu", (char*)" File ", 'f');
  createmenuitem(menupane, (char*)"f1", (char*)"Read data...", 'L', executecb, (char*)"read");
  createmenuitem(menupane, (char*)"f2", (char*)"Save data...", 'S', executecb, (char*)"save");
  createmenuitem(menupane, (char*)"f2b",(char*)"Export...", 'E', executecb, (char*)"export");
  createmenuitem(menupane, (char*)"f2a",(char*)"Send to pipe", 'p', executecb, (char*)"pipe");
  createmenuitem(menupane, (char*)"f3", (char*)"Print data...", 'P', executecb, (char*)"print");
  createmenuseparator(menupane, (char*)"sep1");
  createmenuitem(menupane, (char*)"f14", (char*)"Quit", 'Q', quitcb, NULL);


  //------Create 'Data' pull-down menu--------------//

  menupane = createmenu(g.menubar, (char*)"dataMenu", (char*)" Data ", 'd');
  createmenuitem(menupane, (char*)"f15", (char*)"Copy", 'C', executecb, (char*)"copy");
  createmenuitem(menupane, (char*)"f16", (char*)"Paste", 'P', executecb, (char*)"paste");
  createmenuitem(menupane, (char*)"f16", (char*)"Swap 2 columns", 'P', executecb, (char*)"swap");
  createmenuseparator(menupane, (char*)"sep5");
  createmenuitem(menupane, (char*)"f19", (char*)"Insert cell", 'I', executecb, (char*)"insert");
  createmenuitem(menupane, (char*)"f20", (char*)"Delete cell", 'D', executecb, (char*)"delete");
  createmenuitem(menupane, (char*)"f20a",(char*)"Set column title", 'T', executecb, (char*)"title");
  createmenuitem(menupane, (char*)"f214",(char*)"Set row title", 'R', executecb, (char*)"rowtitle");
  createmenuitem(menupane, (char*)"f20b",(char*)"Set selected to number", 'D', executecb, (char*)"settonumber");
  createmenuitem(menupane, (char*)"f20c",(char*)"Set selected to string", 'D', executecb, (char*)"settostring");
  createmenuitem(menupane, (char*)"f20d",(char*)"Set formula...", 'F', executecb, (char*)"setformula");
  createmenuseparator(menupane, (char*)"sep6");
  createmenuitem(menupane, (char*)"f22", (char*)"Add rows", 'R', executecb, (char*)"addrows");
  createmenuitem(menupane, (char*)"f19", (char*)"Insert row", 'I', executecb, (char*)"insertrow");
  createmenuitem(menupane, (char*)"f23", (char*)"Delete column", 'D', executecb, (char*)"deletecolumn");
  createmenuitem(menupane, (char*)"f23", (char*)"Delete row", 'L', executecb, (char*)"deleterow");
  createmenuitem(menupane, (char*)"f23a",(char*)"Decimate", 'D', executecb, (char*)"decimate");
  createmenuitem(menupane, (char*)"f23b",(char*)"Remove spikes", 'R', executecb, (char*)"removespikes");
  createmenuseparator(menupane, (char*)"sep6");
  createmenuitem(menupane, (char*)"f24", (char*)"Select range", 'S', executecb, (char*)"select");
  createmenuitem(menupane, (char*)"f24", (char*)"Select row", 'S', executecb, (char*)"selectrow");
  createmenuitem(menupane, (char*)"f25", (char*)"Select column", 'T', executecb, (char*)"selectcolumn");
  createmenuitem(menupane, (char*)"f25", (char*)"Deselect row", 'T', executecb, (char*)"unselectrow");
  createmenuitem(menupane, (char*)"f25", (char*)"Deselect column", 'T', executecb, (char*)"unselectcolumn");
  createmenuitem(menupane, (char*)"f26", (char*)"Deselect all", 'E', executecb, (char*)"deselect");
  createmenuitem(menupane, (char*)"f26", (char*)"Find value in column", 'M', executecb, (char*)"find");
  createmenuitem(menupane, (char*)"f26", (char*)"Find maximum in column", 'M', executecb, (char*)"max");
  createmenuitem(menupane, (char*)"f26", (char*)"Find minimum in column", 'M', executecb, (char*)"min");
  createmenuseparator(menupane, (char*)"sep6a");
  createmenuitem(menupane, (char*)"f26", (char*)"Reorder columns", 'M', executecb, (char*)"reorder");
  createmenuseparator(menupane, (char*)"sep6a");
  createmenuitem(menupane, (char*)"f19", (char*)"Define variable", 'v', executecb, (char*)"define");
  createmenuitem(menupane, (char*)"f24", (char*)"Delete all data", 'r', executecb, (char*)"erase");
  createmenuitem(menupane, (char*)"f24a",(char*)"Reformat DIF data", 'R', executecb, (char*)"reformatdif");
  createmenuitem(menupane, (char*)"f24b",(char*)"Reformat as mass spectrum", 'M', executecb, (char*)"reformatmassspec");
  createmenuitem(menupane, (char*)"f23b",(char*)"Convert ms/ms to TIC", 'R', executecb, (char*)"msmstotic");
  createmenuseparator(menupane, (char*)"sep6b");
  createmenuitem(menupane, (char*)"f23c",(char*)"Logic operations", 'R', executecb, (char*)"logic");
//  createmenuitem(menupane, (char*)"f24c",(char*)"Analyze HDX data", 'M', executecb, (char*)"findintegerdifferences");

  //------Create 'Math' pull-down menu------------//

  menupane = createmenu(g.menubar, (char*)"mathMenu", (char*)" Math ", 'm');
  createmenuitem(menupane, (char*)"f18", (char*)"Formula editor", 'A', executecb, (char*)"formula");
  createmenuitem(menupane, (char*)"f18", (char*)"Add", 'A', executecb, (char*)"add");
  createmenuitem(menupane, (char*)"f27", (char*)"Integrate range of data points in col.", 'I', executecb, (char*)"integraterange");
  createmenuitem(menupane, (char*)"f27", (char*)"Integrate", 'I', executecb, (char*)"integrate");
  createmenuitem(menupane, (char*)"f28", (char*)"Differentiate", 'D', executecb, (char*)"differentiate");
  createmenuitem(menupane, (char*)"f29", (char*)"Divide", 'V', executecb, (char*)"divide");
  createmenuitem(menupane, (char*)"f31", (char*)"Integers", 'I', executecb, (char*)"ramp");
  createmenuitem(menupane, (char*)"f31", (char*)"1/x", 'I', executecb, (char*)"inverse");
  createmenuitem(menupane, (char*)"f32", (char*)"Linear regression", 'L', executecb, (char*)"linreg");
  createmenuitem(menupane, (char*)"f33", (char*)"Log e", 'G', executecb, (char*)"ln");
  createmenuitem(menupane, (char*)"f33", (char*)"Log10", 'G', executecb, (char*)"log10");
  createmenuitem(menupane, (char*)"f33", (char*)"ln", 'G', executecb, (char*)"ln");
  createmenuitem(menupane, (char*)"f30", (char*)"exp(x)", 'e', executecb, (char*)"exponential");
  createmenuitem(menupane, (char*)"f30", (char*)"10^x", 't', executecb, (char*)"10^x");
  createmenuitem(menupane, (char*)"f18", (char*)"Multiply", 'M', executecb, (char*)"multiply");
  createmenuitem(menupane, (char*)"f34", (char*)"Negatives to zero", 'N', executecb, (char*)"negatives");
  createmenuitem(menupane, (char*)"f38", (char*)"Normalize", 'N', executecb, (char*)"normalize");
  createmenuitem(menupane, (char*)"f38", (char*)"Normalize to max", 'N', executecb, (char*)"normalizemax");
  createmenuitem(menupane, (char*)"f34", (char*)"Random", 'N', executecb, (char*)"random");
  createmenuitem(menupane, (char*)"f31", (char*)"Reverse", 'I', executecb, (char*)"reverse");
  createmenuitem(menupane, (char*)"f35", (char*)"Shrink/expand", 'S', executecb, (char*)"shrink");
  createmenuitem(menupane, (char*)"f35", (char*)"Bin - average", 'B', executecb, (char*)"binaverage");
  createmenuitem(menupane, (char*)"f35", (char*)"Bin - total", 'B', executecb, (char*)"bintotal");
  createmenuitem(menupane, (char*)"f35", (char*)"Smooth", 'S', executecb, (char*)"smooth");
  createmenuitem(menupane, (char*)"f35a",(char*)"Square root", 'Q', executecb, (char*)"sqrt");
  createmenuitem(menupane, (char*)"f36", (char*)"Sort", 'O', executecb, (char*)"sort");
  createmenuitem(menupane, (char*)"f36", (char*)"Linked Sort", 'O', executecb, (char*)"linkedsort");
  createmenuitem(menupane, (char*)"f37", (char*)"Subtract", 'U', executecb, (char*)"subtract");
  createmenuseparator(menupane, (char*)"sep6");
  createmenuitem(menupane, (char*)"f36", (char*)"Factorial", 'O', executecb, (char*)"fact");
  createmenuitem(menupane, (char*)"f36", (char*)"Combinations", 'O', executecb, (char*)"comb");
  createmenuitem(menupane, (char*)"f36", (char*)"Permutations", 'O', executecb, (char*)"perm");
  createmenuseparator(menupane, (char*)"sep6");
  createmenuitem(menupane, (char*)"f36", (char*)"Sin", 'O', executecb, (char*)"sin");
  createmenuitem(menupane, (char*)"f36", (char*)"Cos", 'O', executecb, (char*)"cos");
  createmenuitem(menupane, (char*)"f36", (char*)"Tan", 'O', executecb, (char*)"tan");
  createmenuseparator(menupane, (char*)"sep6");
  createmenuitem(menupane, (char*)"f36c",(char*)"Error propagation (2 sets)", 'E', executecb, (char*)"errorpropagation");
  createmenuitem(menupane, (char*)"f36f",(char*)"Error propagation (set+value)", 'E', executecb, (char*)"errorpropagationsetvalue");
  createmenuitem(menupane, (char*)"f36d",(char*)"Error propagation (mean+error)", 'E', executecb, (char*)"errorpropagationfrommean");
  createmenuseparator(menupane, (char*)"sep6");
  createmenuitem(menupane, (char*)"f18", (char*)"Mod", 'M', executecb, (char*)"mod");
  createmenuitem(menupane, (char*)"f18", (char*)"Mod inverse", 'M', executecb, (char*)"modinv");
  createmenuitem(menupane, (char*)"f18", (char*)"Mod power", 'M', executecb, (char*)"modpwr");
//  createmenuitem(menupane, (char*)"f36e",(char*)"Error propagation ratio+error ", 'E', executecb, (char*)"errorpropagationratio");
//  createmenuitem(menupane, (char*)"f36g",(char*)"Analyze PKC data           ", 'E', executecb, (char*)"pkc");

  //------Create 'Advanced math' pull-down menu---------------//

  menupane = createmenu(g.menubar, (char*)"advancedmathMenu", (char*)" Advanced math ", 'A');
  createmenuitem(menupane, (char*)"f50", (char*)"Fourier transform", 'F',  executecb, (char*)"fft");
  createmenuitem(menupane, (char*)"f51", (char*)"Non-linear regression", 'N',  executecb, (char*)"nlr");
  createmenuitem(menupane, (char*)"f51a",(char*)"Wavelet transform", 'w',  executecb, (char*)"wavelet");
  createmenuitem(menupane, (char*)"f51", (char*)"Evaluate", 'N',  executecb, (char*)"evaluate");
  createmenuitem(menupane, (char*)"f51", (char*)"Graph", 'N',  executecb, (char*)"graph");
  createmenuitem(menupane, (char*)"f51", (char*)"Close graph", 'N',  executecb, (char*)"ungraph");
  createmenuitem(menupane, (char*)"f54", (char*)"Area under curve", 'U', executecb, (char*)"areaundercurve");
  createmenuitem(menupane, (char*)"f55", (char*)"Subtract baseline", 'B',  executecb, (char*)"subtractbaseline");
  createmenuitem(menupane, (char*)"f55", (char*)"Find peaks", 'P',  executecb, (char*)"findpeaks");

  //------Create 'Statistics' pull-down menu--------------//

  menupane = createmenu(g.menubar, (char*)"statisticsMenu", (char*)" Statistics ", 's');
  createmenuitem(menupane, (char*)"f38", (char*)"Mean, SD, etc", 'M', executecb, (char*)"stat");
  createmenuitem(menupane, (char*)"f39", (char*)"t test 1 column", 't', executecb, (char*)"ttest1");
  createmenuitem(menupane, (char*)"f39", (char*)"t test 2 columns equal variance", 't', executecb, (char*)"ttest2");
  createmenuitem(menupane, (char*)"f39", (char*)"t test 2 columns unequal variance", 't', executecb, (char*)"ttestunequal");
  createmenuitem(menupane, (char*)"f39a",(char*)"t test/paired samples", 'p', executecb, (char*)"tptest");
  createmenuitem(menupane, (char*)"f39b",(char*)"t test 1 var from mean + SD", 't', executecb, (char*)"ttestfrommean1");
  createmenuitem(menupane, (char*)"f39b",(char*)"t test 1 var from mean + SE", 't', executecb, (char*)"ttestfrommeanse1");
  createmenuitem(menupane, (char*)"f39b",(char*)"t test 2 var from mean + SD", 't', executecb, (char*)"ttestfrommean2");
  createmenuitem(menupane, (char*)"f39b",(char*)"t test 2 var from mean + SE", 't', executecb, (char*)"ttestfrommeanse2");
  createmenuitem(menupane, (char*)"f39", (char*)"f test", 'f', executecb, (char*)"ftest");
  createmenuitem(menupane, (char*)"f39c",(char*)"1-way ANOVA", 'f', executecb, (char*)"anova");
  createmenuitem(menupane, (char*)"f39c",(char*)"Mann Whitney U test", 'f', executecb, (char*)"mannwhitney");
  createmenuitem(menupane, (char*)"f40", (char*)"Chi square (Observed vs. expected)", 'C', executecb, (char*)"chisq1");
  createmenuitem(menupane, (char*)"f40", (char*)"Chi square (Set 1 vs. set 2)", 'C', executecb, (char*)"chisq2");
  createmenuitem(menupane, (char*)"f40a",(char*)"Chi square (2 x 2 matrix, Georgetown)", 'C', executecb, (char*)"chisq3");
  createmenuitem(menupane, (char*)"f44", (char*)"Combine two SE", 'S', executecb, (char*)"combinese");
  createmenuitem(menupane, (char*)"f45", (char*)"Combine three SE", 'S', executecb, (char*)"combinese3");
  createmenuseparator(menupane, (char*)"sep7");
  createmenuitem(menupane, (char*)"f41a",(char*)"Poisson (number)", 'p', executecb, (char*)"poisson3");
  createmenuitem(menupane, (char*)"f41", (char*)"Poisson (set+number)", 'p', executecb, (char*)"poisson");
  createmenuitem(menupane, (char*)"f42", (char*)"Poisson (two sets)", 'q', executecb, (char*)"poisson2");
  createmenuitem(menupane, (char*)"f42", (char*)"Binomial probability", 'q', executecb, (char*)"binomialprobability");
  createmenuseparator(menupane, (char*)"sep7");
  createmenuitem(menupane, (char*)"f43", (char*)"Stats on row", 'S', executecb, (char*)"rowstats");
  createmenuitem(menupane, (char*)"f36i",(char*)"Mean, SE/SD on all rows in a set of columns", 'E', executecb, (char*)"errorallsets");
  createmenuitem(menupane, (char*)"f36h",(char*)"Mean, SE/SD of column w/error set", 'E', executecb, (char*)"errorsets");
  createmenuitem(menupane, (char*)"f36i",(char*)"Statistics on all rows in a set of columns", 'E', executecb, (char*)"ttestallvalues");
  createmenuseparator(menupane, (char*)"sep7");
  createmenuitem(menupane, (char*)"f46", (char*)"Replace numbers w/mean", 'R', executecb, (char*)"replacewithmean");
  createmenuitem(menupane, (char*)"f46", (char*)"Replace numbers w/maximim", 'R', executecb, (char*)"replacewithmax");
  createmenuitem(menupane, (char*)"f46", (char*)"Replace numbers w/minimum", 'R', executecb, (char*)"replacewithmin");
  createmenuseparator(menupane, (char*)"sep7");
  createmenuitem(menupane, (char*)"f47", (char*)"P value from r", 'P', executecb, (char*)"pvaluefromr");
  createmenuitem(menupane, (char*)"f47", (char*)"P value of two linear regressions", 'P', executecb, (char*)"pvaluelinreg");


 //------Create 'About' pull-down menu---------------//

  menupane = createmenu(g.menubar, (char*)"aboutMenu", (char*)" About ", 'a');
  createmenuitem(menupane, (char*)"f63", (char*)"About the program", 'p', executecb, (char*)"abouttheprogram");

  //----Create 'Config' pull-down menu---------------//

  menupane = createmenu(g.menubar, (char*)"configMenu", (char*)" Config ", 'o');
  createmenuitem(menupane, (char*)"f69", (char*)"Configure...", 'C',  executecb, (char*)"configure");
 
  //------Create 'Help' pull-down menu---------------//

  menupane = createmenu(g.menubar, (char*)"helpMenu", (char*)" Help ", 'h');
  createmenuitem(menupane, (char*)"f70", (char*)"Help", 'H',  executecb, (char*)"help");
  createmenuitem(menupane, (char*)"f71", (char*)"Manual", 'H',  executecb, (char*)"manual");

  XtManageChild(main_window);
  XtManageChild(g.menubar);
  XtRealizeWidget(parent);

  ////  Create a graphics context 
 
  g.mainwindow=XtWindow(main_window);
    // This overwrites memory for some reason on 64 bit system
  XtVaGetValues(g.main_widget, 
        XmNbackground, &g.bcolor,
        XmNforeground, &g.fcolor,  NULL);         

  g.scolor = 8429728;            //// string color
  g.ecolor = 7129728;            //// equation color
  g.hexcolor = 10985634;         //// hex color
// g.main_fixedcolor = 9999000;
// g.main_fixedcolor = 0x686268;
 g.main_fixedcolor = 0x8f979f;  //// color for row=0 and col=0

  g.gc = XCreateGC(g.display, g.mainwindow, 0L,(XGCValues*)NULL);
  XSetFont(g.display, g.gc, g.fontstruct->fid);
  XSetForeground(g.display, g.gc, g.fcolor);
  XSetBackground(g.display, g.gc, g.bcolor);
}


//--------------------------------------------------------------------//
// createmenu                                                         //
//--------------------------------------------------------------------//
Widget createmenu(Widget parent, char *name, char *label, char mnemonic)
{

    int n;
    Arg args[100];
    Widget menu,w;
    XmString xmstring,xms;
    xms=0;
    char accelerator[50] = "Ctrl<Key>X";  // Substitute mnemonic for the 'X'
    accelerator[9] = mnemonic;

    menu = XmCreatePulldownMenu(parent, name, NULL, 0);
    XtVaSetValues(menu, 
        XmNtearOffModel, XmTEAR_OFF_ENABLED, 
        XmNmenuAccelerator, accelerator,            
        XmNaccelerator, accelerator,            
        XmNacceleratorText, "ASD",            
#ifdef MOTIF2
        XmNtearOffTitle, xms=XmStringCreateSimple(label),
#endif
        XmNmnemonic, mnemonic,            
        NULL);

#ifdef MOTIF2
    XmStringFree(xms);
#endif

    n = 0;
    XtSetArg(args[n], XmNsubMenuId, menu); n++;
    XtSetArg(args[n], XmNmnemonic, mnemonic); n++;
    XtSetArg(args[n], XmNtearOffModel, XmTEAR_OFF_ENABLED); n++;
    XtSetArg(args[n], XmNlabelString, xmstring=XmStringCreateSimple(label)); n++;
    w = XmCreateCascadeButton(parent,(char*)"cascadebutton",args,n);
    XtManageChild(w);
    XmStringFree(xmstring);
    return menu;
}


//--------------------------------------------------------------------//
// createmenuitem                                                     //
//--------------------------------------------------------------------//
Widget createmenuitem(Widget parent, char *name, char *label,
	char mnemonic, XtCallbackProc callback, const char *cbArg)
{
    Widget w;
    XmString xmstring;
    char accelerator[50] = "Ctrl<Key>X";  
    accelerator[9] = mnemonic;
   
    w = XtVaCreateManagedWidget(name, xmPushButtonWidgetClass, parent, 
    	    XmNlabelString, xmstring=XmStringCreateSimple(label),         
    	    XmNmnemonic, mnemonic, NULL);
    XtVaSetValues(w, 
            XmNaccelerator, accelerator, NULL);

    XtAddCallback(w, XmNactivateCallback, (XtCallbackProc)callback, (void *)cbArg);
    XmStringFree(xmstring);
    return w;
}



//--------------------------------------------------------------------//
//  addmainbutton                                                     //
//  add a button on left side                                         //
//--------------------------------------------------------------------//
int addmainbutton(Widget parent, char *name, char *command, XtCallbackProc cb)
{
  static int y=0;
  static int buttonno=0;
  static pushbuttoncbstruct data[100];   ////  Enough for 100 buttons
  Widget w;
  Arg args[100];
  int n,x;
  if(buttonno/2*2==buttonno) x=0; else x=55;
  n=0;
  XtSetArg(args[n], XmNx, x); n++;
  XtSetArg(args[n], XmNy, y); n++;
  XtSetArg(args[n], XmNheight, 24); n++;
  XtSetArg(args[n], XmNwidth, 60); n++;
  XtSetArg(args[n], XmNmarginWidth, 0); n++;
  XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
  XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
  XtSetArg(args[n], XmNrightAttachment, XmATTACH_NONE); n++;
  XtSetArg(args[n], XmNrecomputeSize, False); n++;
  XtSetArg(args[n], XmNtraversalOn, True); n++;
  if(!strcmp(name, "String") || !strcmp(name, "Hex")) 
      w = XmCreateToggleButton(parent, name, args, n);
  else
      w = XmCreatePushButton(parent, name, args, n);
 
  data[buttonno].button = buttonno;
  data[buttonno].widget = w;
  strcpy(data[buttonno].string, name);
  strcpy(data[buttonno].command, command);
  XtAddCallback(w, XmNactivateCallback, (XtCallbackProc)cb, (XtPointer)&data[buttonno]);
  XtAddCallback(w, XmNvalueChangedCallback, (XtCallbackProc)cb, (XtPointer)&data[buttonno]);
  XtManageChild(w);

  if(buttonno/2*2!=buttonno) y += 20;
  buttonno++;
  XtSetKeyboardFocus(w, s.spreadsheet_widget);
  return buttonno-1;
}


//--------------------------------------------------------------------//
// createmenuseparator                                                //
//--------------------------------------------------------------------//
Widget createmenuseparator(Widget parent, char *name)
{
    Widget button;
    int n;
    Arg args[100];
    n = 0;
    button = XmCreateSeparator(parent, name, args, n);
    XtManageChild(button);
    return button;
}



//--------------------------------------------------------------------//
// addstandardbuttons                                                 //
//--------------------------------------------------------------------//
void addstandardbuttons(Widget form, Widget buttons[3], int *data, void *ptr) 
{
  addOkCancelHelpButtons(form, buttons,
         (char*)"Ok", 15, 
         (char*)"Cancel", 125, 
         (char*)"Help", 230, data,
         ptr);
}


//--------------------------------------------------------------------//
// addOkCancelHelpButtons                                             //
// Must use a ptr to put data in callback - a variable would go out   //
// of scope when this function returns.                               //
//--------------------------------------------------------------------//
void addOkCancelHelpButtons(Widget form, Widget button[3], char *s1, int x1, 
   char *s2, int x2, char *s3, int x3, int *data, void *ptr)
{  
   Arg args[100];
   int n;

   n=0;
   XtSetArg(args[n],XmNtopAttachment,   XmATTACH_NONE); n++;
   XtSetArg(args[n],XmNbottomAttachment,XmATTACH_FORM); n++;
   XtSetArg(args[n],XmNleftAttachment,  XmATTACH_FORM); n++;
   XtSetArg(args[n],XmNrightAttachment, XmATTACH_NONE); n++;
   XtSetArg(args[n],XmNwidth,           100); n++;
   XtSetArg(args[n],XmNleftOffset,      x1); n++;
   XtSetArg(args[n],XmNbottomOffset,    2); n++;
   XtSetArg(args[n],XmNdefaultButton,   True); n++;
   XtSetArg(args[n],XmNsensitive,       True); n++;
   XtSetArg(args[n],XmNnavigationType,  XmEXCLUSIVE_TAB_GROUP); n++;
   XtSetArg(args[n],XmNinitialFocus,    True); n++;
   XtSetArg(args[n],XmNaccelerator,     10); n++;
   button[0] = XmCreatePushButton(form,s1,args,n);  // OK button
   XtManageChild(button[0]);

   n=0;
   XtSetArg(args[n],XmNtopAttachment,   XmATTACH_NONE); n++;
   XtSetArg(args[n],XmNbottomAttachment,XmATTACH_FORM); n++;
   XtSetArg(args[n],XmNleftAttachment,  XmATTACH_FORM); n++;
   XtSetArg(args[n],XmNrightAttachment, XmATTACH_NONE); n++;
   XtSetArg(args[n],XmNwidth,           100); n++;
   XtSetArg(args[n],XmNleftOffset,      x2); n++;
   XtSetArg(args[n],XmNbottomOffset,    2); n++;
   XtSetArg(args[n],XmNsensitive,       True); n++;
   XtSetArg(args[n],XmNdefaultButton,   False); n++;
   XtSetArg(args[n],XmNshowAsDefault,   False); n++;
   XtSetArg(args[n],XmNnavigationType,  XmEXCLUSIVE_TAB_GROUP); n++;
   button[1] = XmCreatePushButton(form,s2,args,n);  // Cancel button
   XtManageChild(button[1]);

   n=0;
   XtSetArg(args[n],XmNtopAttachment,   XmATTACH_NONE); n++;
   XtSetArg(args[n],XmNbottomAttachment,XmATTACH_FORM); n++;
   XtSetArg(args[n],XmNleftAttachment,  XmATTACH_FORM); n++;
   XtSetArg(args[n],XmNrightAttachment, XmATTACH_NONE); n++;
   XtSetArg(args[n],XmNwidth,           100); n++;
   XtSetArg(args[n],XmNleftOffset,      x3); n++;
   XtSetArg(args[n],XmNbottomOffset,    2); n++;
   XtSetArg(args[n],XmNsensitive,       True); n++;
   XtSetArg(args[n],XmNdefaultButton,   False); n++;
   XtSetArg(args[n],XmNshowAsDefault,   False); n++;
   XtSetArg(args[n],XmNnavigationType,  XmEXCLUSIVE_TAB_GROUP); n++;
   button[2] = XmCreatePushButton(form,s3,args,n);  // Help button
   XtManageChild(button[2]);


   XtAddCallback(button[0], XmNactivateCallback, (XtCallbackProc)okcb,
         (XtPointer)form);
   XtAddCallback(button[1], XmNactivateCallback, (XtCallbackProc)cancelcb,
         (XtPointer)ptr);
   XtAddCallback(button[2], XmNactivateCallback, (XtCallbackProc)helpcb,
         (XtPointer)data);

   ////  Make Enter & Escape Keys the same as Mouse Click On The Buttons.
   
   XtAddEventHandler(button[0], KeyPressMask, FALSE,
                (XtEventHandler)entercb, (XtPointer)form);
   XtAddEventHandler(button[1], KeyPressMask, FALSE,
                (XtEventHandler)entercb, (XtPointer)form);
   XtAddEventHandler(button[2], KeyPressMask, FALSE,
                (XtEventHandler)entercb, (XtPointer)form);

}


//--------------------------------------------------------------------//
// cancelcb - callback                                                //
//--------------------------------------------------------------------//
void cancelcb(Widget w, XtPointer client_data, XmACB *call_data)
{
   client_data=client_data; call_data=call_data; // keep compiler quiet
   g.getout = 1;
   g.dialogdone=XtParent(w);
   g.state = NORMAL;
   clickboxinfo *c = (clickboxinfo*) client_data;
   if(c!=NULL)
   {  
      if(c->callbacktype==1) editclosecb(w, client_data, call_data);
   }
   g.busy = 0;
}


//--------------------------------------------------------------------//
// init_xlib                                                          //
// Initialize miscellaneous Xlib functions.                           //
// Call after read_settings so want_colormaps & colors are known.     //
//--------------------------------------------------------------------//
void init_xlib(void)
{
  XVisualInfo          vis_template;
  int                  num_visuals; 

  if(setlocale(LC_ALL,"")==NULL)
       fprintf(stderr,"Cant set locale\n");
  if(XSetLocaleModifiers("")==NULL)
       fprintf(stderr, "Cant set locale modifiers\n");

  //-------Define the type of visual desired here---------------------//

  g.display = XtDisplay(g.main_widget);

 //------Set up fonts------------------------------------------------//

  if((g.fontstruct=XLoadQueryFont(g.display, g.font)) == NULL) 
  {   fprintf(stderr, "%s: display %s cannot load font %s\ntrying fixed font",
              g.appname, DisplayString(g.display), g.font);
      strcpy(g.font, (char*)"fixed");
      if((g.fontstruct=XLoadQueryFont(g.display, g.font)) == NULL) 
      {   fprintf(stderr, "%s: display %s cannot load fixed font\n",
              g.appname, DisplayString(g.display));
          exit(1);
      } 
  } 

  //------------------------------------------------------------------//

  g.screen = DefaultScreen(g.display); 
  g.rootwindow = RootWindow(g.display,g.screen);
  vis_template.screen = g.screen;

  //------Get a list of visuals for this screen-----------------------//

  //depth_wanted = DefaultDepth(g.display, g.screen);
  g.visual = DefaultVisual(g.display, g.screen);
  XGetVisualInfo(g.display, VisualScreenMask, &vis_template,
          &num_visuals);  
  g.screen = DefaultScreen(g.display);

}




//--------------------------------------------------------------------//
// xminitializemisc                                                   //
// Initialize miscellaneous Xm functions.                             //
// Must be called after create_drawing_area() so g.mainwindow &       //
//   g.display are defined.                                           //
// Must be called after init_xlib so g.fontstruct is defined.         //
//--------------------------------------------------------------------//
void xminitializemisc(Widget parent)
{

  ////  Get rid of those pesky messages that spew out on the screen
  ////  all the time.
  XtAppSetWarningHandler(XtWidgetToApplicationContext(parent), warningmessagecb);

}


//--------------------------------------------------------------------//
// addlabel                                                           //
// Print text on a Widget (used in dialog boxes)                      //
// Creates a Widget, don't forget to destroy it.                      //
// 'string' is the labelString resource specified in fallbacks        //
// 'position' can be LEFT, CENTER, or RIGHT.                          //
// Make boundaries as big as possible in case user changes string to  //
//  something wide. If x2-x1 is too small for the string, Motif will  //
//  enlarge the dialog box to gigantic size.                          //
//--------------------------------------------------------------------//
Widget addlabel(Widget parent, char *string, int position, double x1, double y1, 
   double x2, double y2) 
{
   Widget label;
   Arg args[100];
   int n=0;
   XtSetArg(args[n], XmNleftPosition, x1); n++;     // % of width from left
   XtSetArg(args[n], XmNtopPosition, y1); n++;      // % of height from top
   XtSetArg(args[n], XmNresizable, False); n++;
   XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNrightPosition, x2); n++;  
   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNbottomPosition, y2); n++;  
   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
   switch(position)
   {  case LEFT: XtSetArg(args[n], XmNalignment, XmALIGNMENT_BEGINNING); n++; break;
      case CENTER: XtSetArg(args[n], XmNalignment, XmALIGNMENT_CENTER); n++; break;
      case RIGHT: XtSetArg(args[n], XmNalignment, XmALIGNMENT_END); n++; break;
   }
   label = XmCreateLabel(parent, string, args, n);
   XtManageChild(label);   
   return label;
} 



//--------------------------------------------------------------------//
// event_loop                                                         //
// Wait here until user clicks OK or Cancel, so that the variables    //
// being filled don't go out of scope. The only other way to avoid    //
// this is to make all the variables global, which would make the     //
// program incompatible with the DOS version.                         //
// Compare the widget ID of the form clicked in case 2 dialogs        //
// are active.                                                        //
//--------------------------------------------------------------------//
void event_loop(Widget w)
{
  g.dialogdone = 0;
  g.getout=0;   
  while(g.dialogdone != w && g.getout==0)
     XtAppProcessEvent (XtWidgetToApplicationContext(g.main_widget), XtIMAll);
  g.dialogdone=0;
  g.busy = 0;
} 



                                                                           
//--------------------------------------------------------------------//
// list - display a single list (Motif)                               //
//--------------------------------------------------------------------//
void list(listinfo *l) 
{
    int wc = 0;
    Widget form, list, *widgetlist;
    widgetlist = new Widget[100];
    form = openwindow(l, &list, widgetlist, &wc);
   
    ////  Handler needed here to raise the window in some WMs. Must be in this exact spot.
    ////  Adding a clickboxmapcb doesnt work. Using on top hint doesn't work either
    XtAddEventHandler(XtParent(form),  StructureNotifyMask, False, (XtEH)(void*)widgetraisecb, (XtP)NULL);

    event_loop(form);
    closewindow(form, widgetlist, wc);
    XtUnmanageChild(form); 
    delete[] widgetlist;
}


//--------------------------------------------------------------------//
// additemtolist                                                      //
// 'w' is the list widget to add the item to.                         //
// 'l' is a ptr to the listinfo struct                                //
// 'form' is the form widget the list is displayed on.                //
// 'showpos' can be TOP or BOTTOM. If TOP it displays first item.     //
//--------------------------------------------------------------------//
void additemtolist(Widget w, Widget form, char *s, int showpos)
{
     XmString xms;
                 // Last param = position; 0 means append to list
     XmListAddItemUnselected(w, xms = XmStringCreateSimple(s), 0);
     if(showpos==BOTTOM) XmListSetBottomPos(w, 0);  // Last item is last visible item.
     XmStringFree(xms);      
     if(XtIsManaged(form)) XRaiseWindow(g.display, XtWindow(XtParent(form)));
}


//--------------------------------------------------------------------//
// closewindow                                                        //
// Unmanages & destroys all Widgets on an opened window. The calling  //
//   routine must maintain a dynamic array w[] of Widgets and the     //
//   Widget count wc, which are filled when openwindow() is called.   //
//--------------------------------------------------------------------//
void closewindow(Widget form, Widget *widgetlist, int wc)
{
    int k;
    XtUnmanageChild(form);
    for(k=0;k<wc;k++) XtDestroyWidget(widgetlist[k]);
    g.busy = 0;
}


//--------------------------------------------------------------------//
// openwindow  (Motif version)                                        //
// same as list except remains on screen permanently. Calling routine //
// calls closewindow() to get rid of it.                              //
// Pass the address of 'helptopic', not the helptopic no. itself,     //
// since the cb would go out of scope.                                //
// Returns the list Widget in 'listwidget' so caller can add to list. //
// 'listwidget' is needed for XmListAddItemUnselected.                //
// The return widget is needed as a parameter to closewindow().       //
// The calling routine must create and maintain a dynamic array w     //
//   of Widgets which is filled when openwindow() is called. This is  //
//   used to destroy the Widgets in closewindow().                    //
//--------------------------------------------------------------------//
Widget openwindow(listinfo *l, Widget *listwidget, Widget *w, int *count)
{
   int k, width=0, xsize, ysize, showtop=TOP, wc=0;
   Widget list, form, label, savebutton;
   Arg args[100];
   XmString xms;
   Cardinal n;
   const int LISTTOP = 8;
   
   xsize = 440;   // Minimum size
   ysize = l->itemstoshow*16 + 100;
   if(l->firstitem>1) showtop=BOTTOM;
   wc = *count;

   //------- Calculate width -----------------------------------------//

   for(k=0;k<l->size;k++) width=max(width,(int)strlen(l->info[k]));
   xsize = max(xsize,min(640,width*8));
   xsize = max(l->width, xsize);

   //------- Create a form dialog shell widget -----------------------//

   n=0;
   XtSetArg(args[n], XmNwidth, xsize); n++;
   XtSetArg(args[n], XmNheight, ysize); n++;
   if(l->transient)
   {  XtSetArg(args[n], XmNtransient, True); n++;}
   else
   {  XtSetArg(args[n], XmNtransient, False); n++;}
   XtSetArg(args[n], XmNresizable, False); n++;
   XtSetArg(args[n], XmNautoUnmanage, False); n++;
   XtSetArg(args[n], XmNkeyboardFocusPolicy, XmEXPLICIT); n++;
   XtSetArg(args[n], XmNnavigationType, XmEXCLUSIVE_TAB_GROUP); n++;
   XtSetArg(args[n], XmNdialogTitle, xms=XmStringCreateSimple(l->title)); n++;
   form = w[wc++] = XmCreateFormDialog(g.main_widget, (char*)"List", args, n);
   XmStringFree(xms);
   label = w[wc++] = addlabel(form, l->title, CENTER, 20, 1, 80, LISTTOP); 

   //--------Title and 'save' button----------------------------------//

   addstandardbuttons(form, &w[wc], &l->helptopic, NULL);
   wc += 3;    // addstandardbuttons adds 3 widgets
   savebutton = w[wc++] = XtVaCreateManagedWidget("Save",
         xmPushButtonWidgetClass,    form,
         XmNtopAttachment,           XmATTACH_NONE,
         XmNbottomAttachment,        XmATTACH_FORM,
         XmNleftAttachment,          XmATTACH_FORM,
         XmNrightAttachment,         XmATTACH_NONE,
         XmNwidth,                   100,
         XmNleftOffset,              335,
         XmNbottomOffset,            5,
         XmNdefaultButton,           False,
         NULL);

   //------- Put scrolled list on the form ---------------------------//

   XtSetArg(args[n], XmNbottomOffset, 34); n++;
   XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
   XtSetArg(args[n], XmNtopWidget, label); n++;    
   XtSetArg(args[n], XmNfractionBase, 100); n++;   // Use percentages
   XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
   XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
   XtSetArg(args[n], XmNresizable, False); n++;
   XtSetArg(args[n], XmNlistSizePolicy, XmCONSTANT); n++;
   XtSetArg(args[n], XmNselectionPolicy, XmSINGLE_SELECT); n++;
   XtSetArg(args[n], XmNscrollBarDisplayPolicy, XmAS_NEEDED); n++; 
   XtSetArg(args[n], XmNorientation, XmVERTICAL); n++;
   XtSetArg(args[n], XmNvisibleItemCount, l->itemstoshow); n++;
   list = w[wc++] = XmCreateScrolledList(form, (char*)"list", args, n);
   *listwidget = list;
   XtAddCallback(list, XmNsingleSelectionCallback, listcb, (XtPointer*)l);

   l->widget = list;
   XtAddCallback(savebutton, XmNactivateCallback, (XtCallbackProc)savecb,
         (XtPointer)l);

   for(k=0;k<l->size;k++) additemtolist(list, form, l->info[k], showtop);
   XtManageChild(list);
   XtManageChild(form);
   XRaiseWindow(g.display, XtWindow(XtParent(form)));
   *count = wc;
   return form;
} 




//--------------------------------------------------------------------//
// message (Motif version)                                            //
// valid modes                                                        //
// -----------                                                        //
// NONE                                                               //
// INFO                                                               //
// PROMPT                                                             //
// QUESTION                                                           //
// YESNOQUESTION                                                      //
// WARNING                                                            //
// ERROR                                                              //
// Space for the returnstring must be already allocated.              //
// 'maxstringsize' must be set to the buffer size of `returnstring' if//
//    mode is PROMPT.  The string will be truncated to this size.     //
// Returns one of:                                                    //
//   OK                                                               //
//   INFO = a string was entered                                      //
//   YES = user clicked Yes to yes/no question or Ok                  //
//   NO  = user clicked No button                                     //
//   CANCEL = user clicked Cancel button                              // 
//--------------------------------------------------------------------//
int message(char *heading){  return message(heading, NULL, INFO, 0); }
int message(char *heading, int mode)
{   if(mode==PROMPT) fprintf(stderr,"Error in message, line %d\n",__LINE__);
    return message(heading, NULL, mode, 50); 
}
int message(char *heading, char *returnstring, int mode, int maxstringsize)
{ 
   int alloc1=0, alloc3=0, n;
   XmString xms, xms2, xms3=NULL;
   Arg args[100];
   Widget w2=0;
   Widget www;
   if(mode==NONE) return 0;
   if(!g.want_messages) return 0;
   int ostate = g.state;
   g.state = MESSAGE;
   g.message_count++;  // callers can check this variable in case error occurs
   clickboxinfo c;
   c.title = new char[1024];
   c.helptopic   = 34;
   c.answer = 0;
   c.form = NULL;
   c.maxstringsize = maxstringsize;
   if(mode==PROMPT && returnstring!=NULL) strcpy(c.title, returnstring);
   n = 0; 

   xms = XmStringCreateLtoR(heading, XmFONTLIST_DEFAULT_TAG);
   XtSetArg(args[n], XmNmessageString, xms); n++;
   XtSetArg(args[n], XmNnoResize, False); n++;
   XtSetArg(args[n], XmNselectionLabelString, xms) ;n++;
   switch(mode)  
   {  case INFO:
         XtSetArg(args[n], XmNtitle, "Information"); n++;
         www = XmCreateInformationDialog(g.main_widget, (char*)"Information", args, n);
         break;
      case PROMPT:
         XtSetArg(args[n], XmNtitle, "Information needed"); n++;
         if(returnstring[0] && returnstring!=NULL)
         {   alloc3=1;
             XtSetArg(args[n], XmNtextString,xms3=XmStringCreateSimple(returnstring));n++;
         }
         www = XmCreatePromptDialog(g.main_widget, (char*)"Prompt", args, n);
         if(alloc3) XmStringFree(xms3); 
         break;
      case ERROR:
         XtSetArg(args[n], XmNtitle, "Error"); n++;
         www = XmCreateErrorDialog(g.main_widget, (char*)"Error", args, n);
         break;
      case QUESTION:         // Ok, cancel, help
         XtSetArg(args[n], XmNtitle, "Question"); n++;
         www = XmCreateQuestionDialog(g.main_widget, (char*)"Question", args, n);
         break;
      case YESNOQUESTION:    // Yes, no, cancel, help
         XtSetArg(args[n], XmNtitle, "Question"); n++;
         XtSetArg(args[n], XmNokLabelString, xms2=XmStringCreateSimple((char*)"Yes")); n++;
         www = XmCreateQuestionDialog(g.main_widget, (char*)"YesNoQuestion", args, n);
         n = 0;
         w2 = XmCreatePushButton(www, (char*)"No", args, n);
         XtManageChild(w2);
         alloc1 = 1;
         XtAddCallback(w2, XmNactivateCallback, nomsgcb, (XtPointer)&c);
         XmStringFree(xms2);
        break;
      case WARNING:
      default:
         XtSetArg(args[n], XmNtitle, "Warning"); n++;
         www = XmCreateWarningDialog(g.main_widget, (char*)"Warning", args, n);
         break;
   }
   if(mode==PROMPT)
      XtAddCallback(www, XmNokCallback, promptmsgcb, (XtPointer)&c);
   else
      XtAddCallback(www, XmNokCallback, okmsgcb, (XtPointer)&c);
   XtAddCallback(www, XmNcancelCallback, cancelmsgcb, (XtPointer)&c);

   ////  Handler needed here to raise the window in some WMs. Must be in this exact spot.
   ////  Adding a clickboxmapcb doesnt work. Using on top hint doesn't work either
   XtAddEventHandler(XtParent(www), EnterWindowMask |  StructureNotifyMask, False, (XtEH)(void*)widgetraisecb, (XtP)&c);

   XtManageChild(www);
   XmStringFree(xms);
   XRaiseWindow(g.display, XtWindow(www));
   while(!c.answer)
        XtAppProcessEvent(XtWidgetToApplicationContext(www),XtIMAll);
   if(mode==PROMPT && c.title[0] && c.title!=NULL && returnstring!=NULL)
   {   c.title[FILENAMELENGTH-2] = 0;
       strcpy(returnstring, c.title); 
   }
   if(alloc1) XtDestroyWidget(w2);
   XtDestroyWidget(www);
   delete[] c.title;
   g.state = ostate;
   g.busy = 0;
   return c.answer;
}

//--------------------------------------------------------------------------//
//  widgetraisecb                                                           //
//--------------------------------------------------------------------------//
void widgetraisecb(Widget w, XtP client_data, XEvent *event)
{
   Window win;
   //// Window focus_return_win;
   //// int returnstate=0;
   client_data=client_data; event=event;
   XRaiseWindow(g.display, XtWindow(w));
   XFlush(g.display);

   //// grab back the input focus that is stolen in cinnamon and
   //// some other window managers
   //// win = XGetInputFocus(g.display, &focus_return_win, &returnstate);
   //// 3rd param can be RevertToParent, RevertToPointerRoot, or RevertToNone)
   //// win=0 = focus goes to the window 
   //// win=1 = focus goes to wherever pointer is
   //// win=2 = crash (not often used) 
   win = 0;  // go to the damn window
   XSetInputFocus(g.display, win, RevertToParent, CurrentTime);
}


//--------------------------------------------------------------------// 
//  clickbox                                                          //
//--------------------------------------------------------------------// 
int clickbox(char* title, int startval, int minval,
      int maxval, void f(int answer), int helptopic)
{
   return clickbox(title,startval,minval,maxval,f,null,0,INTEGER,helptopic);
}

double clickbox(char* title, double fstartval, double fminval,
      double fmaxval, void f(double answer), int helptopic)
{
   int startval, minval, maxval, answer;
   double fanswer;
   double factor = pow(10,g.signif);
 
   startval = (int)(fstartval * factor);
   minval = (int)(fminval * factor);
   maxval = (int)(fmaxval * factor);
   answer = clickbox(title, startval, minval, maxval, null, f, g.signif, 
       DOUBLE, helptopic);
   fanswer = (double)answer / factor;
   return fanswer;
}


int clickbox(char* title, int startval, int minval, int maxval, void f1(int answer),
   void f2(double answer), int decimalpoints, int type, int helptopic)
{
   Widget form, slider, text, *w;
   clickboxinfo clickboxdata[2];
   int k, n, wc=0, xsize=340, ysize=200;
   Arg args[100];
   w = new Widget[100];
   char textlabel[FILENAMELENGTH];
   int answer;
   if(minval==maxval || startval<minval || startval>maxval)
   {    message((char*)"Value cannot currently be changed"); 
        g.getout=0;
        return startval; 
   }
   int ostate = g.state;
   g.state = MESSAGE;
   
   //------- Create a form dialog shell widget -----------------------//

   n=0;
   XtSetArg(args[n], XmNwidth, xsize); n++;
   XtSetArg(args[n], XmNheight, ysize); n++;
   XtSetArg(args[n], XmNresizable, False); n++;
   XtSetArg(args[n], XmNautoUnmanage, False); n++;
   XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
   XtSetArg(args[n], XmNtitle, title); n++;
   form = w[wc++] = XmCreateFormDialog(g.main_widget, (char*)"ClickboxForm", args, n);

   //--------Title for dialog box-------------------------------------//

   w[wc++] = addlabel(form,title,CENTER,1,2,99,15);

   //--------Slider---------------------------------------------------//

   n = 0;
   XtSetArg(args[n], XmNwidth, max(200,xsize-20)); n++;      
   XtSetArg(args[n], XmNleftPosition, 5); n++;     // % of width from left
   XtSetArg(args[n], XmNtopPosition, 40); n++;     // % of height from top
   XtSetArg(args[n], XmNbottomPosition, 75); n++;  // % of width from top
   XtSetArg(args[n], XmNrightPosition, 95); n++;   // % of height from left
   XtSetArg(args[n], XmNfractionBase, 100); n++;   // Use percentages
   XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNminimum, minval); n++; 
   XtSetArg(args[n], XmNmaximum, maxval); n++; 
   XtSetArg(args[n], XmNprocessingDirection,XmMAX_ON_RIGHT); n++;      
   XtSetArg(args[n], XmNhighlightOnEnter, True); n++;      
   XtSetArg(args[n], XmNincrement, 1); n++; 
   XtSetArg(args[n], XmNvalue, startval); n++; 
   XtSetArg(args[n], XmNorientation, XmHORIZONTAL); n++;      
   XtSetArg(args[n], XmNscaleHeight, 35); n++;      
#ifdef MOTIF2
   XtSetArg(args[n], XmNshowValue, XmNEAR_BORDER); n++;      
#endif
   XtSetArg(args[n], XmNdecimalPoints, decimalpoints); n++;      
   XtSetArg(args[n], XmNshowArrows, True); n++;      
   XtSetArg(args[n], XmNhighlightThickness, 1); n++;      
   slider = w[wc++] = XmCreateScale(form, (char*)"Value", args, n);
   XtManageChild(slider);


   //--------Text area------------------------------------------------//

   n = 0;
   XtSetArg(args[n], XmNcolumns, 40); n++;
   XtSetArg(args[n], XmNrows, 1); n++;
   XtSetArg(args[n], XmNvalue, textlabel); n++;
   XtSetArg(args[n], XmNleftPosition, 5); n++;     // % of width from left
   XtSetArg(args[n], XmNtopPosition, 20); n++;     // % of height from top
   XtSetArg(args[n], XmNbottomPosition, 38); n++;  // % of width from top
   XtSetArg(args[n], XmNrightPosition, 95); n++;   // % of height from left
   XtSetArg(args[n], XmNfractionBase, 100); n++;   // Use percentages
   XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNmarginHeight, 2); n++;
   XtSetArg(args[n], XmNmarginWidth, 2); n++;
   clickboxdata[0].title = textlabel;
   clickboxdata[0].answer = startval;
   clickboxdata[0].fanswer = startval;
   clickboxdata[0].startval = startval;
   clickboxdata[0].minval = minval;
   clickboxdata[0].maxval = maxval;
   clickboxdata[0].type = type;
   clickboxdata[0].wantpreview = 0;
   clickboxdata[0].decimalpoints = decimalpoints;
   clickboxdata[0].form = form;
   clickboxdata[0].widget[0] = slider;
   clickboxdata[0].f1 = null;
   clickboxdata[0].f2 = null;
   clickboxdata[0].f3 = null;
   clickboxdata[0].callbacktype = 0;
   if(type==INTEGER)
      itoa(startval, textlabel, 10);
   else
      gcvt(startval/pow(10, g.signif), g.signif, textlabel);
   text = w[wc++] = XmCreateText(form, textlabel, args, n);
   XtAddCallback(text, XmNvalueChangedCallback,
      (XtCallbackProc)clickboxcb, (XtPointer)&clickboxdata[0]);
   XtManageChild(text);
   
   //--------Now add a callback so slider can change text area--------//

   clickboxdata[1].title = textlabel;
   clickboxdata[1].answer = startval;
   clickboxdata[1].fanswer = startval;
   clickboxdata[1].startval = startval;
   clickboxdata[1].minval = minval;
   clickboxdata[1].maxval = maxval;
   clickboxdata[1].type = type;
   clickboxdata[1].wantpreview = 0;
   clickboxdata[1].form = form;
   clickboxdata[1].widget[0] = text;
   clickboxdata[1].decimalpoints = max(0,decimalpoints);
   clickboxdata[1].f1 = f1;
   clickboxdata[1].f2 = f2;
   clickboxdata[1].f3 = null;
   clickboxdata[1].callbacktype = 0;
   XtAddCallback(slider, XmNdragCallback,
      (XtCallbackProc)slidercb, (XtPointer)&clickboxdata[1]);
   XtAddCallback(slider, XmNvalueChangedCallback,
      (XtCallbackProc)slidercb, (XtPointer)&clickboxdata[1]);

   ////  Add callbacks so hitting Return activates Ok button
   for(k=0;k<wc;k++)
      XtAddEventHandler(w[k], KeyPressMask, FALSE, (XtEH)entercb, (XtP)form);


   //--------Ok, Cancel, Help buttons & their callbacks---------------//

   addstandardbuttons(form, &w[wc], &helptopic, NULL);
   wc += 3;    // addstandardbuttons adds 3 widgets

   ////  Handler needed here to raise the window in some WMs. Must be in this exact spot.
   ////  Adding a clickboxmapcb doesnt work. Using on top hint doesn't work either
   for(k=0;k<wc;k++)
      XtAddEventHandler(XtParent(w[k]), StructureNotifyMask, False, (XtEH)(void*)widgetraisecb, (XtP)NULL);

   XtManageChild(form);

   XtVaSetValues(form, XmNy, 200,NULL);  //// trying to raise window doesn't work


   event_loop(form);
   XmScaleGetValue(slider, &answer);
   XtUnmanageChild(form);
   for(k=0;k<wc;k++)
   {   XtRemoveEventHandler(w[k], KeyPressMask, FALSE, (XtEH)entercb, (XtP)form);
       XtDestroyWidget(w[k]);
   }
   delete[] w;
   g.state = ostate;
   return answer;
}


//--------------------------------------------------------------------//
// multiclick- puts integers into clickboxinfo struct by click-and-   //
// box         dragging with the mouse.                               //
//            'getout' must be reset to 0 by the calling routine.     //
//             Returns error code (0=OK)                              //
//             Calls f(), which is a calling-routine-specific func-   //
//             tion, whenever a value is changed or mouse is          //
//             clicked. This allows passed function f() to add new    //
//             buttons.                                               //
//             'a.boxtype' is 0=counter, 1=slider, 2=valslider        //
//             If 'a[0].wantpreview'==1, it adds "preview" button and //
//             only calls f() when clicked; otherwise, it calls f()   //
//             continuously.                                          //
//--------------------------------------------------------------------//
int multiclickbox(char* title, int noofboxes, clickboxinfo* a,
      void f(int values[10]), int helptopic)
{
   Widget form, slider[noofboxes], text[noofboxes], *w;
   clickboxinfo clickboxdata[2][noofboxes];
   int k, n, wc=0, xsize, y, ypct, ysize;
   int boxsize;
   Arg args[100];
   w = new Widget[100];
   char textlabel[FILENAMELENGTH];

   xsize = 425;
   ysize = 150+20*noofboxes;            // overall height in pixels

   const int LABELSTART=64;
   const int LABELEND=85; 
   
   ypct = 100/(noofboxes+2);             // % of y height for each box
   boxsize = (int)(0.9*ypct);            // height of each box in %

   //------- Create a form dialog shell widget -----------------------//

   n=0;
   XtSetArg(args[n], XmNwidth, xsize); n++;
   XtSetArg(args[n], XmNheight, ysize); n++;
   XtSetArg(args[n], XmNresizable, False); n++;
   XtSetArg(args[n], XmNautoUnmanage, False); n++;
   XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
   XtSetArg(args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
   XtSetArg(args[n], XmNtitle, title); n++;
   form = w[wc++] = XmCreateFormDialog(g.main_widget, (char*)"MultiClickboxForm", args, n);

   //--------Title for dialog box-------------------------------------//

   w[wc++] = addlabel(form,title,CENTER,1,1,99,15);

   //--------Slider---------------------------------------------------//

   y = 15;
   for(k=0;k<noofboxes;k++)
   {  
       a[0].answers[k] = a[k].startval;

       w[wc++] = addlabel(form,a[k].title,LEFT,LABELSTART,y+boxsize/4,
                 LABELEND,y+boxsize+boxsize/4);
       itoa(a[k].startval, textlabel, 10);

       n = 0;
       XtSetArg(args[n], XmNwidth, max(300,xsize-20)); n++;      
       XtSetArg(args[n], XmNleftPosition, 5); n++;         // % of width from left
       XtSetArg(args[n], XmNtopPosition, y); n++;          // % of height from top
       XtSetArg(args[n], XmNbottomPosition, y+boxsize); n++;    // % of width from top
       XtSetArg(args[n], XmNrightPosition, LABELSTART-1); n++;  // % of height from left
       XtSetArg(args[n], XmNfractionBase, 100); n++;       // Use percentages
       XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
       XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
       XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
       XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
       XtSetArg(args[n], XmNminimum, a[k].minval); n++; 
       XtSetArg(args[n], XmNmaximum, a[k].maxval); n++; 
       XtSetArg(args[n], XmNprocessingDirection,XmMAX_ON_RIGHT); n++;      
       XtSetArg(args[n], XmNhighlightOnEnter, True); n++;      
       XtSetArg(args[n], XmNincrement, 1); n++; 
       XtSetArg(args[n], XmNvalue, a[k].startval); n++; 
       XtSetArg(args[n], XmNorientation, XmHORIZONTAL); n++;      
       XtSetArg(args[n], XmNshowValue, XmNONE); n++;      
       XtSetArg(args[n], XmNdecimalPoints, 0); n++;      
       XtSetArg(args[n], XmNshowArrows, True); n++;      
       XtSetArg(args[n], XmNhighlightThickness, 1); n++;      
       slider[k] = w[wc++] = XmCreateScale(form, (char*)"MultiClickBoxScale", args, n);
       XtManageChild(slider[k]);
 
       //--------Text area------------------------------------------------//
 
       n = 0;
       XtSetArg(args[n], XmNcolumns, 20); n++;
       XtSetArg(args[n], XmNrows, 1); n++;
       XtSetArg(args[n], XmNvalue, textlabel); n++;
       XtSetArg(args[n], XmNleftPosition, LABELEND+1); n++;   // % of width from left
       XtSetArg(args[n], XmNtopPosition, y+boxsize/4); n++;   // % of height from top
       XtSetArg(args[n], XmNbottomPosition, y+boxsize); n++;  // % of width from top
       XtSetArg(args[n], XmNrightPosition, 98); n++;          // % of height from left
       XtSetArg(args[n], XmNfractionBase, 100); n++;          // Use percentages
       XtSetArg(args[n], XmNtopAttachment, XmATTACH_POSITION); n++;
       XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
       XtSetArg(args[n], XmNbottomAttachment, XmATTACH_POSITION); n++;
       XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
       XtSetArg(args[n], XmNmarginHeight, 2); n++;
       XtSetArg(args[n], XmNmarginWidth, 2); n++;
       strcpy(clickboxdata[0][k].title, textlabel);
       clickboxdata[0][k].answer = a[k].startval;
           // An answers[] array has been allocated only for the first one.
           // All the data are accessible to all the other sliders, and all
           // are aliased to the same original array.
       clickboxdata[0][k].answers = a[0].answers; 
       clickboxdata[0][k].k = k;
       clickboxdata[0][k].startval = a[k].startval;
       clickboxdata[0][k].minval = a[k].minval;
       clickboxdata[0][k].maxval = a[k].maxval;
       clickboxdata[0][k].type = INTEGER;
       clickboxdata[0][k].wantpreview = 0;
       clickboxdata[0][k].widget[0] = slider[k];
       clickboxdata[0][k].f1 = null;
       clickboxdata[0][k].f2 = null;
       clickboxdata[0][k].f3 = f;
       clickboxdata[0][k].callbacktype = 0;
       text[k] = w[wc++] = XmCreateText(form, textlabel, args, n);
       XtAddCallback(text[k], XmNvalueChangedCallback,
          (XtCallbackProc)clickboxcb, (XtPointer)&clickboxdata[0][k]);
       XtManageChild(text[k]);
 
       //--------Now add a callback so slider can change text area--------//
 
       strcpy(clickboxdata[1][k].title, textlabel);
       clickboxdata[1][k].answer = a[k].startval;
       clickboxdata[1][k].answers = a[0].answers;
       clickboxdata[1][k].k = k;
       clickboxdata[1][k].startval = a[k].startval;
       clickboxdata[1][k].minval = a[k].minval;
       clickboxdata[1][k].maxval = a[k].maxval;
       clickboxdata[1][k].type = INTARRAY;
       clickboxdata[1][k].wantpreview = 0;
       clickboxdata[1][k].widget[0] = text[k];
       clickboxdata[1][k].decimalpoints = 0;
       clickboxdata[1][k].f1 = null;
       clickboxdata[1][k].f2 = null;
       clickboxdata[1][k].f3 = f;
       clickboxdata[1][k].callbacktype = 0;
       XtAddCallback(slider[k], XmNvalueChangedCallback,
          (XtCallbackProc)slidercb, (XtPointer)&clickboxdata[1][k]);
       if(a[k].wantdragcb)
        XtAddCallback(slider[k], XmNdragCallback,
          (XtCallbackProc)slidercb, (XtPointer)&clickboxdata[1][k]);
       y+=ypct;
   }

   //--------Ok, Cancel, Help buttons & their callbacks---------------//


   addstandardbuttons(form, &w[wc], &helptopic, NULL);
   wc += 3;    // addstandardbuttons adds 3 widgets

   XtManageChild(form);
   event_loop(form);
   for(k=0;k<noofboxes;k++)
        XmScaleGetValue(slider[k], &a[k].answer);
   XtUnmanageChild(form);
   for(k=0;k<wc;k++) XtDestroyWidget(w[k]);
   delete[] w;
   return 0;
}




//--------------------------------------------------------------------//
// save_ok - returns YES if everything saved & ok to quit.            //
//           Returns NO if user clicks "no".                          //
//           Returns ABORT if user clicks "cancel"                    //
//--------------------------------------------------------------------//
int save_ok(int ino)
{  
    char name[FILENAMELENGTH];
    int answer=YES;
    char tempstring[2*FILENAMELENGTH];
    g.getout=0;
    if(ino<0) return(ERROR);                // should not happen
    sprintf(tempstring,"Save image #%d? \n(%s)",ino, name);

    switch(message(tempstring,YESNOQUESTION))
    {  case YES: answer=YES; 
                 if(g.getout)answer=ABORT;  // Clicked Cancel 
                 break;
       case NO:  answer=NO; break;          // No, dont save but ok to quit
       case CANCEL: answer=ABORT; break;    // Clicked Cancel in message box
    }
    return(answer);
}




 


//--------------------------------------------------------------------//
// cant_overwrite_file - returns 1 if file exists & dont want to      //
// erase it                                                           //
//--------------------------------------------------------------------//
int cant_overwrite_file(char* filename)
{
     char temp[2048];
     if(!access(filename,F_OK))
     {  sprintf(temp,"Overwrite existing file\n%s (y/n)?",filename);
        if(message(temp,QUESTION)!=YES)
        {  message((char*)"File transfer aborted",WARNING); 
           return(ABORT);
        }
     }
     return(OK);
}




//--------------------------------------------------------------------//
// Callbacks                                                          //
//--------------------------------------------------------------------//

//--------------------------------------------------------------------//
// clickboxcb - callback for getting a value in dialog box.  Called   //
// by a text widget. Reads the Widget's string, converts it to an     //
// integer, checks it against the values in the clickboxinfo struct,  //
// and sets the slider value in the 'c->widget' (which must be a      //
// XmScale).                                                          // 
//--------------------------------------------------------------------//
void clickboxcb(Widget w, XtPointer client_data, XmACB *call_data)
{
  call_data=call_data;  // Keep compiler quiet
  int answer=0;
  double fanswer=0;
  char *string = new char[256];

  clickboxinfo *c = (clickboxinfo*)client_data;
  string = XmTextGetString(w);
  if(string) 
  {   if(c->type==INTEGER) 
      {    answer = atoi(string);
           answer=max(min(answer,c->maxval),c->minval);
           XmScaleSetValue(c->widget[0], answer);
      }
      if(c->type==DOUBLE) 
      {    fanswer = atof(string) * 1.000001 * pow(10, g.signif);
           c->fanswer = fanswer;
           fanswer=max(min(fanswer,c->maxval),c->minval);
           XmScaleSetValue(c->widget[0], (int)(fanswer));
      }
  }
  delete[] string;
}


//--------------------------------------------------------------------//
// slidercb - Called by a slider widget, reads the slider's value,    //
// Widget's string, converts it to a string, and sets the text in the //
// 'c->widget' (which must be a XmText).                              //
//--------------------------------------------------------------------//
void slidercb(Widget w, XtPointer client_data, XmACB *call_data)
{
  call_data=call_data;  // Keep compiler quiet
  int answer=0;
  double fanswer=0.0;
  char *string = new char[256];
  strcpy(string," ");
  clickboxinfo *c = (clickboxinfo*)client_data;
  XmScaleGetValue(w, &answer);

  switch(c->type)
  {     case INTEGER: 
             c->f1(answer);  
             itoa(answer, string, 10); 
             break;
        case DOUBLE:
             fanswer = ((double)answer)/pow(10, g.signif);  
             c->fanswer = fanswer;
             c->f2(fanswer); 
             gcvt(fanswer, g.signif, string); 
             break;
        case INTARRAY:
             c->answers[c->k] = answer;
             c->answer = answer;
             c->f3(c->answers);
             itoa(c->answers[c->k], string, 10); 
             break;
  }
  XmTextSetString(c->widget[0], string);
  delete[] string;
}



//--------------------------------------------------------------------//
// timer_callback                                                     //
// Callback function for asynchronous or automatic events             //
//--------------------------------------------------------------------//
void timer_callback(XtPointer client_data, XtIntervalId *timer_id)
{  
    client_data=client_data; timer_id=timer_id;  // Keep compiler quiet
    // Restart timer once every 'DELAY' milliseconds
    int DELAY = 250;

    XtAppAddTimeOut(XtWidgetToApplicationContext(g.main_widget), DELAY, 
       timer_callback, NULL);
}
 


//--------------------------------------------------------------------//
// whatevent - prints name of event for debugging.                    //
//--------------------------------------------------------------------//
void whatevent(int eventtype)
{
  switch(eventtype)
  { case KeyPress: printf("keypress ");fflush(stdout); break;
    case KeyRelease: printf("keyrelease ");fflush(stdout); break;
    case Expose: printf("expose ");fflush(stdout); break;
    case MotionNotify: printf("motionnotify ");fflush(stdout); break;
    case ButtonPress: printf("buttonpress ");fflush(stdout); break;
    case ButtonRelease: printf("buttonrelease ");fflush(stdout); break;
    case EnterNotify: printf("enternotify ");fflush(stdout); break;
    case LeaveNotify: printf("leavenotify ");fflush(stdout); break;
    case FocusIn: printf("focusin ");fflush(stdout); break;
    case FocusOut: printf("focusout ");fflush(stdout); break;
    case KeymapNotify: printf("keymapnotify ");fflush(stdout); break;
    case GraphicsExpose: printf("graphicsexpose ");fflush(stdout); break;
    case NoExpose: printf("noexpose ");fflush(stdout); break;
    case CirculateNotify: printf("circulatenotify ");fflush(stdout); break;
    case ConfigureNotify : printf("configurenotify ");fflush(stdout); break;
    case MapNotify: printf("mapnotify ");fflush(stdout); break;
    case ResizeRequest: printf("resizerequest ");fflush(stdout); break;
    case MappingNotify: printf("mappingnotify ");fflush(stdout); break;
    case ColormapNotify: printf("colormapnotify ");fflush(stdout); break;
    case ReparentNotify: printf("reparentnotify ");fflush(stdout); break;
    default: printf("Unknown ");fflush(stdout); break;
  }
}



//--------------------------------------------------------------------//
// getfilename                                                        //
// Returns filename                                                   //
//--------------------------------------------------------------------//
char *getfilename(char *oldtitle)
{ 
   int k;
   clickboxinfo c;
   c.title = oldtitle;
   c.widget[1] = NULL;
   c.helptopic = 0;
   c.done      = 0;
   c.wc        = 0;
   c.w         = new Widget[20];
   c.answers   = new int[10];
   c.f1 = null;
   c.f2 = null;
   c.f3 = null;
   c.wc = 0;
   c.form = NULL;
   c.type = SINGLE;
   c.list = NULL;
   c.callbacktype = 0;
   filenamecb(g.main_widget, &c, NULL);   // This changes c.wc
   while(!c.done)
        XtAppProcessEvent (XtWidgetToApplicationContext(g.main_widget), XtIMAll);
   for(k=0;k<c.wc;k++) XtDestroyWidget(c.w[k]);
   delete[] c.w;
   delete[] c.answers;
   if(strlen(c.title)>0) return c.title;
   else return oldtitle;
}


//--------------------------------------------------------------------//
// getfilenames - allocates array `names', gets one or more filenames,//
// allocates space for each filename, and returns the no. of names    //
// selected by user. Caller must free the strings and the array.      //
// Returns list of filenames and puts their count in `count'.         //
//--------------------------------------------------------------------//
char **getfilenames(int *count)
{ 
   int k;
   clickboxinfo c;
   c.widget[1] = NULL;
   c.helptopic = 0;
   c.done      = 0;
   c.wc        = 0;
   c.w         = new Widget[20];
   c.f1 = null;
   c.f2 = null;
   c.f3 = null;
   c.wc = 0;
   c.form = NULL;
   c.type = MULTIPLE;
   c.count = 0;
   c.list = NULL;
   strcpy(c.title, "Multiple Filenames");
   filenamecb(g.main_widget, &c, NULL);    // This changes c.wc
   while(!c.done)
        XtAppProcessEvent (XtWidgetToApplicationContext(g.main_widget), XtIMAll);
   for(k=0;k<c.wc;k++) XtDestroyWidget(c.w[k]);
   delete[] c.w;
   *count = c.count;
   return c.list;
}


//--------------------------------------------------------------------//
// warningmessagecb - handle warning messages from Xt (mainly "unable //
// to allocate colormap entry for default background").               //
//--------------------------------------------------------------------//
void warningmessagecb(String aaa)
{ 
   aaa=aaa;
   printstatus(XTWARNING);
}


//--------------------------------------------------------------------//
// filenamecb                                                         //
// Called in response to mouse click on filename pushbutton.          //
// Calling routine must maintain a dynamic list of Widgets, then      //
//   destroy them when finished.                                      //
//--------------------------------------------------------------------//
void filenamecb(Widget widget, XtPointer client_data, XmACB *call_data)
{ 
   widget=widget;call_data=call_data;  // Keep compiler quiet
   Widget f,form,w,w3;
   XmString title, dir_mask, xms, fpath;
   int n;
   Arg args[100];
   clickboxinfo *c = (clickboxinfo *)client_data;
   int wc = c->wc;

   int rx,ry,wx,wy; 
   uint keys;
   Window rwin, cwin; 

   ////  Find current position of main window 
   XQueryPointer(g.display,g.mainwindow,&rwin,&cwin,&rx,&ry,&wx,&wy,&keys);

   n = 0;
   XtSetArg(args[n], XmNdialogStyle, XmDIALOG_MODELESS); n++;
   XtSetArg(args[n], XmNtitle, (char*)"file Selection"); n++;
   XtSetArg(args[n], XmNresizable, False); n++;
   XtSetArg(args[n], XmNautoUnmanage, False); n++;
   XtSetArg(args[n], XmNkeyboardFocusPolicy, XmEXPLICIT); n++;
   XtSetArg(args[n], XmNnavigationType, XmEXCLUSIVE_TAB_GROUP); n++;
   XtSetArg(args[n], XmNdialogTitle, xms = XmStringCreateSimple((char*)"Select File"));n++;
   form = c->w[wc++] = XmCreateFormDialog(g.main_widget, (char*)"DialogForm", args, n);
   XmStringFree(xms);


   title = XmStringLtoRCreate((char*)"Filter",XmSTRING_DEFAULT_CHARSET);   
   dir_mask = XmStringLtoRCreate((char*)"*",XmSTRING_DEFAULT_CHARSET); 


   n = 0;
   XtSetArg(args[n], XmNdirMask, dir_mask); n++;  
   XtSetArg(args[n], XmNresizable, True); n++;  
   XtSetArg(args[n], XmNfilterLabelString, title); n++; 
   XtSetArg(args[n], XmNdirectory, fpath = XmStringLtoRCreate(c->path,XmSTRING_DEFAULT_CHARSET)); n++;
  ////  On Solaris the File List is not drawn correctly 
  ////  so give it lots of extra space.
#ifdef SOLARIS 
   XtSetArg(args[n], XmNwidth, 800); n++;
#endif
   XtSetArg(args[n], XmNtopPosition, 0); n++;
   XtSetArg(args[n], XmNleftPosition, 0); n++;
   XtSetArg(args[n], XmNrightPosition, 100); n++;  
   XtSetArg(args[n], XmNfractionBase, 100); n++;   
   XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
   XtSetArg(args[n], XmNleftAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNrightAttachment, XmATTACH_POSITION); n++;
   XtSetArg(args[n], XmNlistVisibleItemCount, 50); n++;
   f = c->w[wc++] = XmCreateFileSelectionBox(form, (char*)"fileSelector", args, n);

   Widget child;
   if(c->type==MULTIPLE)
   {   child = c->listwidget = XmFileSelectionBoxGetChild(f,XmDIALOG_LIST);
       XtVaSetValues(child, XmNselectionPolicy, XmMULTIPLE_SELECT, NULL);
   }

   ////   The first pushbutton is automatically placed in the wrong
   ////   position, so an unused, unmanaged pushbutton must be added
   ////   to trick Motif.
   
   n = 0;
   c->w[wc++] = XmCreatePushButton(f, (char*)"crudola", args, n);

   n = 0;
   XtSetArg(args[n], XmNchildPlacement, XmPLACE_BELOW_SELECTION); n++;
   w3 = c->w[wc++] = XmCreatePushButton(f, (char*)"Chdir", args, n);
 

   ////  Set c->widget[0] to the fileselectionbox widget.
   ////  c->widget[1] is the 'filename' push button in dialog box.

   c->widget[0] = form; 
   XtAddCallback(f, XmNcancelCallback, fscancelcb, (XtPointer)c);
   XtAddCallback(f, XmNokCallback, fsokcb, (XtPointer)c); 
   XtAddCallback(w3, XmNactivateCallback, fscdcb, (XtPointer)f);
   XmStringFree(fpath);
   XmStringFree(title);
   XmStringFree(dir_mask);

   XtManageChild(w3);
   XtManageChild(f);
   XtManageChild(form);

   //// It is necessary to manually set the colors in each subwindow.
           
   w = XmFileSelectionBoxGetChild(f, XmDIALOG_LIST);
   XtVaSetValues(w, 
#ifdef SOLARIS 
            XmNwidth, 400,
#endif
            XmNscrollBarDisplayPolicy, XmSTATIC,  // Essential for Solaris
            NULL);

   c->wc = wc;
   return;
}


//--------------------------------------------------------------------//
// fsokcb - callback for "ok" button in file selection box.           //
// Caller passes a clickboxinfo struct in client_data.                //
// Widget2 must be set to a pushbutton Widget or to NULL.             // 
// This is where memory is allocated for multifilename selector.      //
// The list array must be freed with delete[].                        //
// The filename strings must be freed with XtFree().                  //
//--------------------------------------------------------------------//
void fsokcb(Widget w, XtPointer client_data, void *call_data)
{
   g.state = NORMAL;
   call_data=call_data;
   Widget selectiontext;  // XmText for user to type in a filename 
   Arg args[100];
   XmString xms;
   XmStringTable xmstrings;
   int k,n = 0;
   clickboxinfo *c = (clickboxinfo *)client_data;
   char *filename;    // Do not allocate space, this is done automatically
                      // by XmStringGetLtoR().


   c->count=0;
   ////  `filename' is malloc'd to match the free() below.
   filename = (char*) malloc(FILENAMELENGTH);
   
   if(c->type == MULTIPLE)
   {   XtVaGetValues(c->listwidget, XmNselectedItemCount, &c->count, NULL); 

       ////  This returns the XmList items themselves, not a copy. The 
       ////  XmStringTable is automatically created by FileSelectionBox.
       ////  The application must not free the XmStrings (see OSF p. 1-859).
       
       XtVaGetValues(c->listwidget, XmNselectedItems, &xmstrings, NULL);
       c->list = new char*[c->count+1];
       for(k=0;k<c->count;k++)
       {
            ////  This allocates memory for items in list, use XtFree to free it.
            XmStringGetLtoR(xmstrings[k], XmSTRING_DEFAULT_CHARSET, &c->list[k]);
       }
       if(c->count>0) 
       {    strcpy(c->title, c->list[0]);
            strcpy(filename,c->list[0]);
       }
   }

   ////  For single file mode, Motif puts filename in an editable text widget. 
   ////  If user didn't select a multiple file, also check the text widget.

   if(c->type == SINGLE || c->count==0)
   {    selectiontext = XmFileSelectionBoxGetChild(w, XmDIALOG_TEXT);
        XtVaGetValues(selectiontext, XmNvalue, &filename, NULL);
        if(c->type == MULTIPLE) 
        {    c->list[0] = XtMalloc(FILENAMELENGTH);  // Free this with XtFree()
             strcpy(c->list[0], filename);
        }
        strcpy(c->title, filename);
        c->count=1;
   }

   ////  Put the new filename on the pushbutton. 
   ////  The pushbutton's XmNresizable resource must be False, otherwise
   ////  Motif will use this as an excuse to resize the entire dialog box.
   ////  Unmanage c->calling widget, which is the bulletin board - not w,
   ////  which is the filenamewidget on top.

   ////  c->widget[1] is the 'filename' pushbutton in dialogbox.
   ////  If calling from somewhere else, this should be set to NULL.
     
   n = 0;
   XtSetArg(args[n], XmNwidth, 200); n++; 
   XtSetArg(args[n], XmNlabelString, xms=XmStringCreateSimple(basefilename(filename))); n++;
   if(c->widget[1] != NULL)
       XtSetValues(c->widget[1], args, n); // Change label in boxbutton Widget
   XmStringFree(xms);
   XtUnmanageChild(c->widget[0]);          // Unmanage the bulletin board Widget
   c->done = 1;                            // Set flag indicating done
   free(filename);

   ////  Focus goes to OK button when file is selected, so pressing Enter
   ////  twice will load the image.
   
   if(c->form != NULL) XmProcessTraversal(c->form, XmTRAVERSE_NEXT_TAB_GROUP); 

}


//--------------------------------------------------------------------//
// fscancelcb - callback for "cancel" button in file selection box.   //
//--------------------------------------------------------------------//
void fscancelcb(Widget w, XtPointer client_data, void *call_data)
{
   g.state = NORMAL;
   w=w; call_data=call_data;  // Keep compiler quiet
   clickboxinfo *c = (clickboxinfo *)client_data;
   XtUnmanageChild(c->widget[0]);
}


//--------------------------------------------------------------------//
// fscdcb - callback for "chdir" button in file selection box.        //
//--------------------------------------------------------------------//
void fscdcb(Widget w, XtPointer client_data, void *call_data)
{
   w=w; call_data=call_data;  // Keep compiler quiet
   Widget f = (Widget)client_data;
   char *dir;
   dir = new char[1024];
   XmString xms = XmStringCreateLtoR(dir, XmSTRING_DEFAULT_CHARSET);
   XtVaGetValues(f, XmNdirectory,  &xms, NULL);       
   XmStringGetLtoR(xms, XmSTRING_DEFAULT_CHARSET, &dir);
   chdir(dir);
   strcpy(g.currentdir,dir);
   XmStringFree(xms);
   delete[] dir;
}



//--------------------------------------------------------------------//
// listcb - callback for list box.                                    //
// Item text is placed in 'string', l->selection points to answer no. //
//--------------------------------------------------------------------//
void listcb(Widget w, XtPointer client_data, void *call_data)
{
   int k;
   listinfo *l = (listinfo *)client_data;
   char *string;
   XmString xms;
   XmListCallbackStruct *ptr;
   ptr = (XmListCallbackStruct*) call_data;
   XmStringGetLtoR(ptr->item, XmSTRING_DEFAULT_CHARSET, &string);
 
   if(l->selection!=NULL) *l->selection = ptr->item_position;
 
   *l->selection = *l->selection - 1;
   k =  *l->selection;
   if(l->allowedit && k >= l->editrow)
   {     message(l->edittitle, l->info[k] + l->editcol, PROMPT, l->maxstringsize);
         xms = XmStringCreateSimple(l->info[k]);
         XmListReplaceItemsPosUnselected(w, &xms, 1, k+1);
         XmStringFree(xms);
   }
}


//--------------------------------------------------------------------//
// executecb - callback for main menu items.                          //
//--------------------------------------------------------------------//
void executecb(Widget w, XtPointer client_data, void *call_data)
{
  if(g.busy) return;
  g.busy = 1;
  w=w; call_data=call_data;  // Keep compiler quiet
  char cmd[256];
  static int noofargs = 0;
  static char **arg;                // up to 20 arguments
  int k;
  arg = new char*[20];
  for(k=0; k<20; k++) {  arg[k] = new char[80]; arg[k][0] = 0; }
  strcpy(cmd, (char*)client_data);
  execute(cmd, arg, noofargs);
  for(k=0; k<20; k++) delete[] arg[k];
  delete[] arg;
  g.busy = 0;
}


//--------------------------------------------------------------------//
// disarmcb - callback for releasing a pushbutton.                    //
//--------------------------------------------------------------------//
void disarmcb(Widget w, XtPointer client_data, XmACB *call_data)
{  
    w=w; call_data=call_data; client_data=client_data;  // Keep compiler quiet
}


//--------------------------------------------------------------------//
// quitcb - callback for quitting.                                    //
//--------------------------------------------------------------------//
void quitcb(Widget w, XtPointer client_data, void *call_data)
{
  w=w; call_data=call_data; client_data=client_data;  // Keep compiler quiet
  int status=YES;
  graph_close();
  savesettings();
  if(g.touched) 
  {   status = message((char*)"Save current spreadsheet?", YESNOQUESTION);
      if(status==YES || status==OK) save(NULL, 0);
  }
  if(status!=CANCEL) exit(0);
}


//--------------------------------------------------------------------//
// entercb - callback for pressing a key on a pushbutton              //
//--------------------------------------------------------------------//
void entercb(Widget w, XtPointer client_data, XEvent *event)
{
   w=w;   
   if(event->type!=KeyPress) return;
   char buffer[20];
   XComposeStatus compose;
   XKeyEvent keyevent = event->xkey;
   KeySym keysym;
   int key=0;
    
   Widget widget = *(Widget *)client_data;
 
   //// All this is to allow Enter key to activate a button
   int bufsize=20;
   XLookupString(&keyevent,buffer,bufsize,&keysym,&compose);
   key = keysym & 0xffff;
   if(key==XK_Return) g.dialogdone = widget;
   if(key==XK_Escape) 
   {    g.getout = 1;
        g.dialogdone=XtParent(w);
        g.state = NORMAL;
   }
}


//--------------------------------------------------------------------//
// okcb - callback                                                    //
//--------------------------------------------------------------------//
void okcb(Widget w, XtPointer client_data, XmACB *call_data)
{
   w=w; call_data=call_data; // keep compiler quiet
   Widget widget = *(Widget *)client_data;
   g.dialogdone = widget;
}


//--------------------------------------------------------------------//
// savecb - callback for saving list data                             //
//--------------------------------------------------------------------//
void savecb(Widget w, XtPointer client_data, XmACB *call_data)
{
   w=w; call_data=call_data; // keep compiler quiet
   static char filename[FILENAMELENGTH]="";
   listinfo *l = (listinfo *)client_data;
   char temp[2*FILENAMELENGTH] = "Data";
   int k, count;
   FILE *fp;

   count = l->size;
   if(count==0){ message((char*)"No data to save",ERROR); return; }
   message((char*)"New filename:",filename,PROMPT,FILENAMELENGTH-1);
   if(filename[0]==0 || filename==NULL) return;
   if(cant_overwrite_file(filename)) return;
   if ((fp=fopen(filename,"wt")) == NULL)
        error_message(filename, errno);
   else
   {    fprintf(fp,"%s\n",l->title);
        for(k=0;k<count;k++)
            fprintf(fp,"%s\n",l->info[k]);
        sprintf(temp,"%s saved in %s",l->title,filename);
        message(temp);
        fclose(fp);
   }
}


//--------------------------------------------------------------------//
// okmsgcb - callback for message box ok button                       //
//--------------------------------------------------------------------//
void okmsgcb(Widget w, XtPointer client_data, void *call_data)
{
   call_data=call_data; // keep compiler quiet
   clickboxinfo *c = (clickboxinfo *)client_data;
   g.dialogdone=XtParent(w);
   if(c!=NULL) c->answer = YES;
   g.busy = 0;
}


//--------------------------------------------------------------------//
// cancelmsgcb -  callback for message box cancel button              //
//--------------------------------------------------------------------//
void cancelmsgcb(Widget w, XtPointer client_data, void *call_data)
{
   call_data=call_data; // keep compiler quiet
   g.dialogdone=XtParent(w);
   g.getout = 1;
   g.busy = 0;
}


//--------------------------------------------------------------------//
// nomsgcb -  callback for message box 'no' button                    //
//--------------------------------------------------------------------//
void nomsgcb(Widget w, XtPointer client_data, void *call_data)
{
   call_data=call_data;  // keep compiler quiet
   clickboxinfo *c = (clickboxinfo *)client_data;
   c->answer = NO;       // This is not zero!
   g.dialogdone=XtParent(w);
   g.busy = 0;
}


//--------------------------------------------------------------------//
// promptmsgcb - callback for message box ok button to get string     //
//--------------------------------------------------------------------//
void promptmsgcb(Widget w, XtPointer client_data, void *call_data)
{
   clickboxinfo *c = (clickboxinfo *)client_data;
   char *string;         
   XmSelectionBoxCallbackStruct *ptr;
   ptr = (XmSelectionBoxCallbackStruct *)call_data;
    ////  This function allocates space so strcpy it then free it.
   XmStringGetLtoR(ptr->value, XmSTRING_DEFAULT_CHARSET, &string);
    ////  Make sure it is short enough
   if(string!=NULL && string[0])
   {   strcpy(c->title,string);
       c->title[c->maxstringsize-1] = 0;
       XtFree(string);
   }
   g.dialogdone=XtParent(w);
   c->answer = INFO;
}
 

//--------------------------------------------------------------------//
// maincancelcb -  callback for cancel button on main screen          //
//--------------------------------------------------------------------//
void maincancelcb(Widget w, XtPointer client_data, XmACB *call_data)
{
   w=w; call_data=call_data; client_data=client_data; // keep compiler quiet
   g.getout = 1;
   g.state = NORMAL;
   printstatus(g.state);
   g.busy = 0;
}


//--------------------------------------------------------------------//
// helpcb - callback for getting help                                 //
// The address of int helptopic is in client_data.                    //
//--------------------------------------------------------------------//
void helpcb(Widget w, XtPointer client_data, XmACB *call_data)
{
    w=w; call_data=call_data;  // keep compiler quiet
    int *h = (int *)client_data; 
    help(*h);
    g.busy = 0;
} 



//--------------------------------------------------------------------//
// mainbuttoncb                                                       //
//--------------------------------------------------------------------//
void mainbuttoncb(Widget w, XtPointer client_data, void *call_data)
{
   w=w; call_data=call_data;  // keep compiler quiet
   int k,noofargs=0;
   char **arg;                // up to 20 arguments
   arg = new char*[20];
   for(k=0;k<20;k++) arg[k] = new char[80];
   pushbuttoncbstruct *p = (pushbuttoncbstruct *) client_data;
   parse_string(p->command, arg, NULL, noofargs, 0, 80, 99, ' ',0);
   execute(arg[0], arg, noofargs);
   for(k=0;k<20;k++) delete[] arg[k];
   delete[] arg;
}



//--------------------------------------------------------------------//
// getword - returns first word demarcated by a char. in `buffer'     //
// 'count' must be initialized to 0 by calling routine before first   //
// call (keeps track of buffer position).                             //
// If word is enclosed in "" it sets is_string to 1.                  //
//--------------------------------------------------------------------//
void getword(char *buffer, char *word, char delimiter, 
    int maxlen, int &count, int &is_string)
{
    char tempstring[maxlen+1];  
    int k=count, pos=0, instring=0,hit=0;
    if((int)strlen(buffer) >= maxlen) buffer[maxlen]=0;

    while(buffer[k])
    {   
         if(buffer[k]=='\"'){ instring=1-instring; k++; }
         if(instring)
         {   //// Eliminate leading spaces
             while(buffer[k]==' ' && !hit) { k++; continue; }            
             if(buffer[k]!='\"') tempstring[pos++]=buffer[k];
             is_string = 1;
             hit=1;
         }else
         {   //if(buffer[k]==',') { k++; continue; }
             if(buffer[k]!=delimiter && 
                buffer[k]!='\n' && buffer[k]!='\t' ) 
                tempstring[pos++]=buffer[k];
             else if(pos) break;
             is_string = 0;
         }
         tempstring[pos]=0;
         k++;
         if(buffer[k]=='\n') break;
    }
    count = k;    
    strcpy(word, tempstring);
}    


//--------------------------------------------------------------------//
// parse_string                                                       //
// input: a single command with arguments  e.g. "load test 1 1 2 3".  //
// Each word k is placed in buf[k][80]. buf[0][] is the command, the  //
// rest are arguments separated by spaces.                            //
// The no. of arguments, not counting buf[0], is placed in `noofargs'.//
// `startarg' is optional and is 0 by default.                        //
// returns 1 if it is a string, 0 if it is a number, -1 if error      //
//--------------------------------------------------------------------//
int parse_string(char *cmd, char **buf, int *stringflag, int &noofargs, 
     int startarg, int maxlen, int maxargs, char delimiter, int skip)
{   
    int count=0, is_string=0;
    char word[1024];
    noofargs = 0;
    do
    {    word[0]=0;
         while(skip > 0)
         {    getword(cmd, word, delimiter, 1024, count, is_string);
              skip--;
         }
         getword(cmd, word, delimiter, 1024, count, is_string);
         if(strlen(word)==0) break;
         if(stringflag != NULL) stringflag[noofargs] = is_string;
         if(startarg + noofargs < maxargs)
         {
            strncpy(buf[startarg + noofargs], word, maxlen);
            noofargs++;
         }
         if(noofargs > maxargs){ message((char*)"Error parsing string"); return -1; }
         if(count >= (int)strlen(cmd)) break;
    }while(1);

    return is_string;
}


//--------------------------------------------------------------------//
// parse_title                                                        //
// input: a string starting with '#'.                                 //
// puts each title demarcated by #..# in s.title array                //
//--------------------------------------------------------------------//
void parse_title(char *string, char *destination, int &noofargs, 
     int titlecol, int maxlen)
{   
    int first_char, count=0, is_string;
    char word[128];
    noofargs = 0;
    count = strchr(string, '#') - string;
    do
    {    word[0]=0;
         getword(string, word, '#', 80, count, is_string);
         if(strlen(word)==0) break;
         //// eliminate leading spaces
         first_char = 0;
         while(word[first_char]==' '){ first_char++; continue; }
         strncpy(destination, word+first_char, maxlen);
         if(count >= (int)strlen(string)) break;
    }while(1);
}


//--------------------------------------------------------------------//
// parse_header                                                       //
// input: a string containing text in quotes                          //
// puts each title in s.title array                                   //
// no. of titles is returned in titlecol                              //
//--------------------------------------------------------------------//
void parse_header(char *string, char **destination, int &noofargs, 
     int &count, int maxlen)
{   
     parse_string(string, destination, NULL, count, 1, maxlen, s.cols-1, ' ', 0);
}


//--------------------------------------------------------------------------//
//  set_widget_string                                                       //
//--------------------------------------------------------------------------//
void set_widget_string(Widget w, char* s)
{
  XmString xms;
  if(!s) return;
  if(!w) return;
  if(!XtIsManaged(w)) return;
  XtVaSetValues(w, XmNlabelString,  xms = XmStringCreateLocalized(s), NULL);  
  XtVaSetValues(w, XmNcursorPosition, strlen(s),NULL); 
  if(xms) XmStringFree(xms);                
}
