//--------------------------------------------------------------------------//
// xdata1.cc                                                                //
// Latest revision: 07-31-2023                                              //
//--------------------------------------------------------------------------//

#include "xdata.h"

extern Globals g;
extern Spreadsheet s;

//--------------------------------------------------------------------------//
// initialize_globals                                                       //
// initialize global data - except for `g.text' which is  initialized in    //
// create_spreadsheet.                                                      //
//--------------------------------------------------------------------------//
void initialize_globals(void)
{
    char *ptr=NULL;
    char tempstring[2*1024];
    g.busy = 0;
    g.dialogdone=0;
    strcpy(g.version, "1.0.0");
    g.main_xsize = 800;
    g.main_ysize = 1000;
    g.getout=0;                   // Flag for 'cancel'
    g.key_ascii=0;
    g.key_alt=0;
    g.state=NORMAL;
    g.key_ctrl=0;
    g.key_shift=0;
    g.signif = 8;                 // Significant digits to display (overridden by xdata.ini)
    g.font = new char[256];
    strcpy(g.font,"*helvetica-bold-r-normal--12*");
    strcpy(g.filename,"Untitled");
    strcpy(g.gvfilename,"/usr/bin/gv");
    g.want_messages = 1;
    strcpy(g.printer_command,"lpr -Plp");
    strcpy(g.printer_filename,"Spreadsheet.dat");
    g.rows = 10000;
    g.cols = 50;
    g.graph_active = 0;
    g.autoupdate=1;
    g.autoscale=1;
    g.touched=0;
    g.mask =  Button1MotionMask |    // Click & Drag
              PointerMotionMask |    // Mouse movement  
              ButtonPressMask   |
              ButtonReleaseMask;

    ////  $HOME/.xdata for xdataa.ini, formats in Bourne
    ////  $home/.xdata in tcsh
    ptr = getenv("HOME");
    if(ptr!=NULL) strcpy(g.homedir, ptr);    
    else 
    {   ptr = getenv("home");
        if(ptr!=NULL) strcpy(g.homedir, ptr);    
        else 
        {   strcpy(g.homedir, g.startdir);
            fprintf(stderr,"Unable to determine home directory in __FILE__ line __LINE__\n");
        }
    }
    strcat(g.waveletpath,"/usr/local/lib/wavelets");
    strcat(g.homedir,"/.xdata");
    printf("Home directory %s\n",g.homedir);
    sprintf(tempstring,"%s/xdata.ini", g.homedir);
    if(access(tempstring, F_OK))   // If directory does not exist, create it
        mkdir(g.homedir, 0755);
    readsettings();   

    if(g.rows>0) s.rows = g.rows;
    if(g.cols>0) s.cols = g.cols;
    g.commands = new char[EDITORSIZE];  
    g.commands[0]=0;
    printf("\nInitializing to %d rows x %d columns\n",g.rows,g.cols);
    initialize_evaluator();
    define_variable((char*)"index",  0);   
    define_variable((char*)"p",  3.14159265358979291704599745571613);
    g.string = 0;
    g.message_count = 0;
    g.number_type = DEC;
}


//-------------------------------------------------------------------//
//   define_variable                                                 //
//-------------------------------------------------------------------//
void define_variable(char *name, double value)
{
    add_variable(name, value);
    printf("variable %8s %8.8g defined\n",name,value);
}


//-------------------------------------------------------------------//
//   define_user_variable                                            //
//-------------------------------------------------------------------//
void define_user_variable(char **arg, int noofargs)
{
   arg=arg; noofargs=noofargs;
   printf("I need to add some code here\n");
}



//-------------------------------------------------------------------//
// execute                                                           //
// execute a command from menu or a macro. This method is better than//
// having a separate callback for each menu item because it makes it //
// to use macros.                                                    //
//-------------------------------------------------------------------//
int execute(char *command, char **arg, int noofargs)
{
  int j,k,status=OK;
  char o[2048];
  char tempstring[2048];
  g.getout=0;
  int e=0;   // flag if something was executed
//  leavespreadsheet(); // calculate maxrows + maxcols

  for(k=0,j=0;k<=(int)strlen(command);k++)
  {  switch(command[k]) // Clean up command if from menu
       {   case '.':    // If abbreviation
           case '_':    // If underscore
           case ' ': break;
           default : o[j++] = tolower(command[k]);
                     o[j]=0;
       }
  }

  ////  The commands (in alphabetical order) 
  ////  Cant put in struct due to compiler bugs.         

  ////  Data management
  if(strcmp(o, "abouttheprogram")==SAME){ e=1; abouttheprogram();}
  if(strcmp(o, "bin")==SAME){ e=1; bin(arg, noofargs, 1);}
  if(strcmp(o, "bintotal")==SAME){ e=1; bin(arg, noofargs, 0);}
  if(strcmp(o, "binaverage")==SAME){ e=1; bin(arg, noofargs, 1);}
  if(strcmp(o, "changerowtitle")==SAME){ e=1; changerowtitle(arg, noofargs);}
  if(strcmp(o, "copy")==SAME){ e=1; copy(arg, noofargs, COPY); }
  if(strcmp(o, "decimate")==SAME){ e=1; decimate(arg, noofargs);}
  if(strcmp(o, "define")==SAME){ e=1; define_user_variable(arg, noofargs);}
  if(strcmp(o, "delete")==SAME){ e=1; delete_selected_cells(arg, noofargs, 0);}
  if(strcmp(o, "deletecolumn")==SAME){ e=1; delete_column(arg, noofargs);}
  if(strcmp(o, "deleterow")==SAME){ e=1; delete_row(arg, noofargs);}
  if(strcmp(o, "deselect")==SAME){ e=1; deselectallcells(arg, noofargs, 0);}
  if(strcmp(o, "down")==SAME){ e=1; move_cursor(DOWN);}
  if(strcmp(o, "erase")==SAME){ e=1; erase();}
  if(strcmp(o, "errorallsets")==SAME){ e=1; errorallsets();}
  if(strcmp(o, "errorpropagation")==SAME){ e=1; error_propagation();}
  if(strcmp(o, "errorpropagationsetvalue")==SAME){ e=1; error_propagationsetvalue();}
  if(strcmp(o, "errorpropagationfrommean")==SAME){ e=1; errorpropagationfrommean();}
  if(strcmp(o, "errorpropagationratio")==SAME){ e=1; errorpropagationratio();}
  if(strcmp(o, "errorsets")==SAME){ e=1; errorsets();}
  if(strcmp(o, "export")==SAME){ e=1; export_data(arg, noofargs);}
  if(strcmp(o, "findintegerdifferences")==SAME){ e=1; findintegerdifferences(arg, noofargs);}
  if(strcmp(o, "help")==SAME){ e=1; help(0);}
  if(strcmp(o, "hex")==SAME){ e=1; hex(arg, noofargs);}
  if(strcmp(o, "insert")==SAME){ e=1; insert_cell(s.current_row,s.current_col);}
  if(strcmp(o, "insertrow")==SAME){ e=1; insert_row(s.current_row,s.current_col);}
  if(strcmp(o, "length")==SAME){ e=1; set_length(arg, noofargs, 0); }
  if(strcmp(o, "left")==SAME){ e=1; move_cursor(LEFT);}
  if(strcmp(o, "linkedsort")==SAME){ e=1; linkedsort(arg, noofargs);}
  if(strcmp(o, "load")==SAME){ e=1; read(arg, noofargs);}
  if(strcmp(o, "find")==SAME){ e=1; maximum(arg, noofargs, EQU);}
  if(strcmp(o, "manual")==SAME){ e=1; manual();}
  if(strcmp(o, "max")==SAME){ e=1; maximum(arg, noofargs, MAX);}
  if(strcmp(o, "min")==SAME){ e=1; maximum(arg, noofargs, MIN);}
  if(strcmp(o, "move")==SAME){ e=1; move(arg, noofargs); }
  if(strcmp(o, "msmstotic")==SAME){ e=1; msmstotic(arg, noofargs);}
  if(strcmp(o, "paste")==SAME){ e=1;  copy(arg, noofargs, PASTE);}
  if(strcmp(o, "pipe")==SAME){ e=1; pipe();}
  if(strcmp(o, "print")==SAME){ e=1; print_spreadsheet();}
  if(strcmp(o, "quit")==SAME){ e=1; quitcb(0,0,0);}
  if(strcmp(o, "read")==SAME){ e=1; read(arg, noofargs);}
  if(strcmp(o, "reload")==SAME){ e=1; reload(arg, noofargs);}
  if(strcmp(o, "reorder")==SAME){ e=1; reorder();}
  if(strcmp(o, "refresh")==SAME){ e=1; refreshspreadsheet();}
  if(strcmp(o, "removespikes")==SAME){ e=1; remove_spikes(arg, noofargs);}
  if(strcmp(o, "right")==SAME){ e=1; move_cursor(RIGHT);}
  if(strcmp(o, "save")==SAME){ e=1; save(arg, noofargs);}
  if(strcmp(o, "select")==SAME){ e=1; sselect(arg, noofargs);}
  if(strcmp(o, "selectcolumn")==SAME){ e=1; select_column(s.current_col,1);}
  if(strcmp(o, "selectrow")==SAME){ e=1; select_row(s.current_row,1);}
  if(strcmp(o, "set")==SAME){ e=1; set(arg, noofargs);}
  if(strcmp(o, "setformula")==SAME){ e=1; set_formula();}
  if(strcmp(o, "sethexmode")==SAME){ e=1; sethexmode();}
  if(strcmp(o, "settonumber")==SAME){ e=1; math(arg,noofargs,SETTONUMBER);}
  if(strcmp(o, "settostring")==SAME){ e=1; math(arg,noofargs,SETTOSTRING);}
  if(strcmp(o, "sort")==SAME){ e=1; sort(arg, noofargs);}
  if(strcmp(o, "string")==SAME) {  g.string = 1-g.string; }
  if(strcmp(o, "swap")==SAME){ e=1; swap_columns(arg, noofargs);}
  if(strcmp(o, "unselectcolumn")==SAME){ e=1; select_column(s.current_col,0);}
  if(strcmp(o, "unselectrow")==SAME){ e=1; select_row(s.current_row,0);}
  if(strcmp(o, "up")==SAME){ e=1; move_cursor(UP);}
  if(strcmp(o, "whatis")==SAME){ e=1; whatis(arg, noofargs);}

  if(strcmp(o, "c")==SAME){ e=1; copy(arg, noofargs, COPY); }
  if(strcmp(o, "m")==SAME){ e=1; move(arg, noofargs); }
  if(strcmp(o, "i")==SAME){ e=1; insert_cell(s.current_row,s.current_col);}
  if(strcmp(o, "l")==SAME){ e=1; set_length(arg, noofargs, 0); }
  if(strcmp(o, "q")==SAME){ e=1; quitcb(0,0,0);}
  if(strcmp(o, "r")==SAME){ e=1; read(arg, noofargs);}
  if(strcmp(o, "s")==SAME){ e=1; save(arg, noofargs);}
  if(strcmp(o, "x")==SAME){ e=1; delete_column(arg, noofargs);}

  //// Statistics
  if(strcmp(o, "binomialprobability")==SAME){ e=1; binomialprobability(arg, noofargs);}
  if(strcmp(o, "combinese")==SAME){ e=1; combinese();}
  if(strcmp(o, "combinese3")==SAME){ e=1; combinese3();}
  if(strcmp(o, "chisq1")==SAME){ e=1; chisq1();}
  if(strcmp(o, "chisq2")==SAME){ e=1; chisq2();}
  if(strcmp(o, "chisq3")==SAME){ e=1; chisq3();}
  if(strcmp(o, "ftest")==SAME){ e=1; ftest();}
  if(strcmp(o, "anova")==SAME){ e=1; anova();}
  if(strcmp(o, "mannwhitney")==SAME){ e=1; mannwhitney();}
  if(strcmp(o, "poisson")==SAME){ e=1; poisson(arg,noofargs);}
  if(strcmp(o, "poisson2")==SAME){ e=1; poisson2(arg,noofargs);}
  if(strcmp(o, "poisson3")==SAME){ e=1; poisson3(arg,noofargs);}
  if(strcmp(o, "pvaluefromr")==SAME){ e=1; pvaluefromr();}
  if(strcmp(o, "pvaluelinreg")==SAME){ e=1; pvaluelinreg();}
  if(strcmp(o, "rowstats")==SAME){ e=1; rowstats();}
  if(strcmp(o, "stat")==SAME){ e=1; statistics();}
  if(strcmp(o, "tptest")==SAME){ e=1; tptest();}
  if(strcmp(o, "ttest1")==SAME){ e=1; ttest1();}
  if(strcmp(o, "ttest2")==SAME){ e=1; ttest2();}
  if(strcmp(o, "ttestallvalues")==SAME){ e=1; ttestallvalues();}
  if(strcmp(o, "ttestunequal")==SAME){ e=1; ttest_unequal();}
  if(strcmp(o, "ttestfrommean1")==SAME){ e=1; ttestfrommean1();}
  if(strcmp(o, "ttestfrommeanse1")==SAME){ e=1; ttestfrommeanse1();}
  if(strcmp(o, "ttestfrommean2")==SAME){ e=1; ttestfrommean2();}
  if(strcmp(o, "ttestfrommeanse2")==SAME){ e=1; ttestfrommeanse2();}

  //// Math
  if(strcmp(o, "add")==SAME){ e=1; math(arg,noofargs,ADD);}
  if(strcmp(o, "areaundercurve")==SAME){ e=1; areaundercurve();}
  if(strcmp(o, "cbrt")==SAME){ e=1; math(arg,noofargs,CBRT);}
  if(strcmp(o, "comb")==SAME){ e=1; fact(arg,noofargs,COMB);}
  if(strcmp(o, "cos")==SAME){ e=1; math(arg,noofargs,COS);}
  if(strcmp(o, "differentiate")==SAME){ e=1; math(arg,noofargs,DIFFERENTIATE);}
  if(strcmp(o, "div")==SAME){ e=1; math(arg,noofargs,DIVIDE);}
  if(strcmp(o, "divide")==SAME){ e=1; math(arg,noofargs,DIVIDE);}
  if(strcmp(o, "evaluate")==SAME){ e=1; evaluate(arg,noofargs);} // evaluate nlr equation
  if(strcmp(o, "ex")==SAME){ e=1; shrink(arg, noofargs);}
  if(strcmp(o, "expand")==SAME){ e=1; shrink(arg, noofargs);}
  if(strcmp(o, "exp")==SAME){ e=1; math(arg,noofargs,EXPONENTIATE);}
  if(strcmp(o, "exponentiate")==SAME){ e=1; math(arg,noofargs,EXPONENTIATE);}
  if(strcmp(o, "exponential")==SAME){ e=1; math(arg,noofargs,EXPONENTIAL);}
  if(strcmp(o, "fact")==SAME){ e=1; fact(arg,noofargs,FACT);}
  if(strcmp(o, "findpeaks")==SAME){ e=1; findpeaks(arg,noofargs);}
  if(strcmp(o, "formula")==SAME)
  {  e=1; edit((char*)"Formula editor",(char*)"Formula editor (Shift-enter=execute line)",g.commands,22,70,65536,0); }
  if(strcmp(o, "fft")==SAME){ e=1; fft(arg,noofargs);}
  if(strcmp(o, "graph")==SAME){ e=1; graph(0,arg,noofargs);}
  if(strcmp(o, "integrate")==SAME){ e=1; math(arg,noofargs,INTEGRATE);}
  if(strcmp(o, "integraterange")==SAME){ e=1; integraterange(arg,noofargs);}
  if(strcmp(o, "int")==SAME){ e=1; math(arg,noofargs,INT);}
  if(strcmp(o, "int2")==SAME){ e=1; math(arg,noofargs,INT);}
  if(strcmp(o, "inverse")==SAME){ e=1; math(arg,noofargs,INVERSE);}
  if(strcmp(o, "linreg")==SAME){ e=1; linreg(arg,noofargs);}
  if(strcmp(o, "log10")==SAME){ e=1; math(arg,noofargs,LOG10);}
  if(strcmp(o, "ln")==SAME){ e=1; math(arg,noofargs,LN);}
  if(strcmp(o, "lr")==SAME){ e=1; linreg(arg,noofargs);}
  if(strcmp(o, "mul")==SAME){ e=1; math(arg,noofargs,MULTIPLY);}
  if(strcmp(o, "multiply")==SAME){ e=1; math(arg,noofargs,MULTIPLY);}
  if(strcmp(o, "n")==SAME){ e=1; math(arg,noofargs,NEGATIVETOZERO);}
  if(strcmp(o, "neg")==SAME){ e=1; math(arg,noofargs,NEGATIVETOZERO);}
  if(strcmp(o, "negatives")==SAME){ e=1; math(arg,noofargs,NEGATIVETOZERO);}
  if(strcmp(o, "nlr")==SAME){ e=1; marq(arg,noofargs);}
  if(strcmp(o, "normalize")==SAME){ e=1; normalize(arg, noofargs);}
  if(strcmp(o, "normalize2")==SAME){ e=1; normalize2(arg, noofargs);}
  if(strcmp(o, "normalizemax")==SAME){ e=1; normalizemax(arg, noofargs);}
  if(strcmp(o, "pageup")==SAME){ e=1; page_up();}
  if(strcmp(o, "pagedown")==SAME){ e=1; page_down();}
  if(strcmp(o, "perm")==SAME){ e=1; fact(arg,noofargs,PERM);}
  if(strcmp(o, "pkc")==SAME){ e=1; pkc();}
  if(strcmp(o, "ram")==SAME){ e=1; math(arg,noofargs,RAMP);}
  if(strcmp(o, "ramp")==SAME){ e=1; math(arg,noofargs,RAMP);}
  if(strcmp(o, "ran")==SAME){ e=1; math(arg,noofargs,RANDOM);}
  if(strcmp(o, "rand")==SAME){ e=1; math(arg,noofargs,RANDOM);}
  if(strcmp(o, "random")==SAME){ e=1; math(arg,noofargs,RANDOM);}
  if(strcmp(o, "reformatdif")==SAME){ e=1; reformat_dif_table(arg,noofargs);}
  if(strcmp(o, "reformatmassspec")==SAME){ e=1; reformatmassspec(arg,noofargs);}
  if(strcmp(o, "replacewithmean")==SAME){ e=1; replacewithmean(arg,noofargs,1);}
  if(strcmp(o, "replacewithmax")==SAME){ e=1; replacewithmean(arg,noofargs,2);}
  if(strcmp(o, "replacewithmin")==SAME){ e=1; replacewithmean(arg,noofargs,3);}
  if(strcmp(o, "reverse")==SAME){ e=1; reverse(arg,noofargs);}
  if(strcmp(o, "rowtitle")==SAME){ e=1; changerowtitle(arg, noofargs);}
  if(strcmp(o, "shr")==SAME){ e=1; shrink(arg, noofargs);}
  if(strcmp(o, "shrink")==SAME){ e=1; shrink(arg, noofargs);}
  if(strcmp(o, "sin")==SAME){ e=1; math(arg,noofargs,SIN);}
  if(strcmp(o, "smo")==SAME){ e=1; smooth(arg, noofargs);}
  if(strcmp(o, "smooth")==SAME){ e=1; smooth(arg, noofargs);}
  if(strcmp(o, "sqr")==SAME){ e=1; math(arg,noofargs,SQRT);}
  if(strcmp(o, "sqrt")==SAME){ e=1; math(arg,noofargs,SQRT);}
  if(strcmp(o, "sub")==SAME){ e=1; math(arg,noofargs,SUBTRACT);}
  if(strcmp(o, "subtract")==SAME){ e=1; math(arg,noofargs,SUBTRACT);}
  if(strcmp(o, "baseline")==SAME){ e=1; subtract_baseline(arg,noofargs);}
  if(strcmp(o, "subtractbaseline")==SAME){ e=1; subtract_baseline(arg,noofargs);}
  if(strcmp(o, "tan")==SAME){ e=1; math(arg,noofargs,TAN);}
  if(strcmp(o, "t")==SAME){ e=1; changetitle(arg, noofargs);}
  if(strcmp(o, "title")==SAME){ e=1; changetitle(arg, noofargs);}
  if(strcmp(o, "ungraph")==SAME){ e=1; graph_close();}
  if(strcmp(o, "wavelet")==SAME){ e=1; wavelet(arg,noofargs);}

  //// functions for integers and ordinary numbers
  if(strcmp(o, "+")==SAME){ e=1; math(arg,noofargs,ADD);}
  if(strcmp(o, "/")==SAME){ e=1; math(arg,noofargs,DIVIDE);}
  if(strcmp(o, "1/x")==SAME){ e=1; math(arg,noofargs,INVERSE);}
  if(strcmp(o, "*")==SAME){ e=1; math(arg,noofargs,MULTIPLY);}
  if(strcmp(o, "10^x")==SAME){ e=1; math(arg,noofargs,TENTOX);}
  if(strcmp(o, "-")==SAME){ e=1; math(arg,noofargs,SUBTRACT);}
  if(strcmp(o, "%")==SAME){ e=1; math(arg,noofargs,MOD);}

  //// functions for hex integers only
  if(strcmp(o, "logic")==SAME){ e=1; logic();}

  //// modular arithmetic only
  if(strcmp(o, "mod")==SAME){ e=1; math(arg,noofargs,MOD);}
  if(strcmp(o, "modinv")==SAME){ e=1; math(arg,noofargs,MODINVERSE);}
  if(strcmp(o, "modpwr")==SAME){ e=1; math(arg,noofargs,MODPWR);}


  if(!e)
  {    e=1; 
       strcpy(tempstring, command);
       for(k=1;k<noofargs;k++) strcat(tempstring, arg[k]);          // Always at least 1 arg
       if(!strncmp(o, "if(", 3)) evaluate_conditional(tempstring);  // Parse column conditional
       else if(o[0]== 'y') 
       {
            evaluate_formula(tempstring);                           // Parse result column
       }else 
       {   
            eval(tempstring);                                       // Evaluate formula
       }
  }

  //--------------------------------------------------------------//
  g.getout=0;
  if(e==0) status=UNKNOWN;
  printstatus(g.state);
 // refreshspreadsheet();

  return status;
}




//--------------------------------------------------------------------//
// itoa                                                               //
// Convert integer to string.                                         //
//--------------------------------------------------------------------//
char *itoa(int value, char *str, int radix)
{
   switch(radix)
   {  
//      case 2:  sprintf(str, "%b",value); break;
      case 8:  sprintf(str, "%o",value); break; 
      case 10: sprintf(str, "%d",value); break; 
      case 16: sprintf(str, "%lx", (long unsigned int) value); break; 
      default: perror("Error converting string"); exit(1);
   } 
   return(str);
}


//--------------------------------------------------------------------//
// ltoa                                                               //
// Convert long integer to string.                                    //
//--------------------------------------------------------------------//
char *ltoa(long int value, char *str, int radix)
{
   return itoa((int)value, str, radix); 
}




//-----------------------------------------------------------------------//
//  gcvt                                                                 //
//  convert double to a string with signif significant digits.           //
//  use our own gcvt because the linux one is totally screwed up.        //
//-----------------------------------------------------------------------//
#ifndef HAVE_GCVT 
char *gcvt(double val, int signif, char *buf)
{  
   buf=buf; val=val;signif=signif;
//// keep as high as possible to avoid sci not hex integers
g.signif=99;
   char gtemp[128];
   char format[32];
   sprintf(format,"%c.%dg",'%',g.signif);
   sprintf(buf, format, val);
   strcpy(gtemp, buf);
   return buf;
}
#endif

//-------------------------------------------------------------------//
// null - do nothing funcs for clickbox(), plot(), etc.              //
//-------------------------------------------------------------------//
void null(int a)
{ a=a; }
void null(int a[10])
{ a[0]=a[0]; }
void null(double a)
{ a=a; }
void null(void)
{}
void null(dialoginfo *a, int b, int c, int d, int e)
{ a=a; b=b; c=c; d=d; e=e; }




//-------------------------------------------------------------------//
// between    - returns 1 if a is between b and c                    //
//-------------------------------------------------------------------//
int between(int a,int b,int c)
{   if((a>=b)&&(a<=c))return(1); 
    else return(0);
}

//-------------------------------------------------------------------//
// between    - returns 1 if a is between b and c                    //
//-------------------------------------------------------------------//
int between(double a, double b, double c)
{   if((a>=b)&&(a<=c))return(1); 
    else return(0);
}



//-------------------------------------------------------------------//
// path  - strips path from filename and returns the path name.      //
//-------------------------------------------------------------------//
void path(char *pathname, char *path)
{  int k,hit=0;
   char tempstring[FILENAMELENGTH];
   strcpy(tempstring,pathname);
   for(k=strlen(tempstring);k>0;k--)
   {  if((tempstring[k]=='/')||(tempstring[k]=='\\')) 
      {  tempstring[k+1]=0;  
         hit=1; 
         break; 
      }
   }
   if(hit) strcpy(path,tempstring);
   return;
}

//-------------------------------------------------------------------//
// basefilename                                                      //
// strip path from filename & returns file name w/o path             //
//-------------------------------------------------------------------//
char* basefilename(char* pathname)
{  int k;
   int slashpos=0;
   char temp[FILENAMELENGTH];
   strcpy(temp,pathname);

   for(k=0;k<(int)strlen(temp);k++)          // find last slash or colon
       if((temp[k]=='/')||(temp[k]=='\\')||(temp[k]==':')) slashpos=k+1;   
   return(pathname+slashpos);
}


//-------------------------------------------------------------------//
// printstatus                                                       //
//-------------------------------------------------------------------//
void printstatus(int state)
{
    state=state;
}


//--------------------------------------------------------------------------//
// make_2d_alias                                                            //
// usage: uint** array;                                                     //
//        uint* array_1d;                                                   //
//        array_1d = (uint*)malloc(x*y*sizeof(uint));                       //
//        array = make_2d_alias(array_1d,x,y);   // Dont mult.by sizeof uint//
//         ...                                   // compiler knows its size //
//        free(array);                                                      //
//        free(array_1d);                                                   //
// Address array using  array[y][x]  or array_1d[index]                     //
// Last parameter is first index = slowest changing index                   //
// Create a 2d alias from an array so the same data can be accessed using   //
// either [][] or [] notation. This permits interfacing with libraries      //
// requiring either type.                                                   //
//--------------------------------------------------------------------------//
int **make_2d_alias(int *b, int x, int y)
{  int k;
   int **result;
   result = (int**)malloc(y*sizeof(int *));
   for(k=0;k<y;k++) result[k] = b + k*x;
   return result;
}
long long int **make_2d_alias(long long int *b, int x, int y)
{  int k;
   long long int **result;
   result = (long long int**)malloc(y*sizeof(long long int *));
   for(k=0;k<y;k++) result[k] = b + k*x;
   return result;
}
uint **make_2d_alias(uint *b, int x, int y)
{  int k;
   uint **result;
   result = (uint**)malloc(y*sizeof(uint *));
   for(k=0;k<y;k++) result[k] = b + k*x;
   return result;
} 
uchar **make_2d_alias(uchar *b, int x, int y)
{  int k;
   uchar **result;
   result = (uchar**)malloc(y*sizeof(uchar *));
   for(k=0;k<y;k++) result[k] = b + k*x;
   return result;
} 
double **make_2d_alias(double *b, int x, int y)
{  int k;
   double **result;
   result = (double**)malloc(y*sizeof(double *));
   for(k=0;k<y;k++) result[k] = b + k*x;
   return result;
} 
long double **make_2d_alias(long double *b, int x, int y)
{  int k;
   long double **result;
   result = (long double**)malloc(y*sizeof(long double *));
   for(k=0;k<y;k++) result[k] = b + k*x;
   return result;
} 



//--------------------------------------------------------------------------//
// strupr                                                                   //
//--------------------------------------------------------------------------//
char *strupr(char *s)
{
   int k;
   for(k=0;k<(int)strlen(s);k++) s[k]=toupper(s[k]);
   return s;
}


//--------------------------------------------------------------------------//
// strlwr                                                                   //
//--------------------------------------------------------------------------//
char *strlwr(char *s)
{
   int k;
   for(k=0;k<(int)strlen(s);k++) s[k]=tolower(s[k]);
   return s;
}

  
//--------------------------------------------------------------------------//
//  filesize                                                                //
//--------------------------------------------------------------------------//
int filesize(char *filename)
{
     struct stat dstatus;
     stat(filename, &dstatus);
     return dstatus.st_size;
}
        

//-------------------------------------------------------------------//
// savesettings                                                      //
//-------------------------------------------------------------------//
void savesettings(void)
{
  FILE* fp;
  int k;
  char tempstring[2*FILENAMELENGTH]="";
  sprintf(tempstring,"%s/xdata.ini",g.homedir);
  if((fp=fopen(tempstring, "wt"))!=NULL )
  { 
       fprintf(fp,"version %s\n",g.version);
       fprintf(fp,"rows %d\n",g.rows);
       fprintf(fp,"cols %d\n",g.cols);
       fprintf(fp,"xsize %d\n",g.main_xsize);
       fprintf(fp,"ysize %d\n",g.main_ysize);
       fprintf(fp,"print_command %s\n",g.printer_command);
       fprintf(fp,"print_file %s\n",g.printer_filename);
       fprintf(fp,"signif_digits %d\n",g.signif);
       fprintf(fp,"wavelet_path %s\n",g.waveletpath);
       fprintf(fp,"gvfilename %s\n",g.gvfilename);
       fprintf(fp,"nbuttons %d\n",g.nbuttons);
       for(k=0;k<g.nbuttons;k++)
       {
          fprintf(fp,"button %s %s\n", g.button[k]->label, g.button[k]->cmd);
       }
  }else fprintf(stderr,"Can't save settings\n");
  fclose(fp);
}
        

//-------------------------------------------------------------------//
// readsettings                                                      //
//-------------------------------------------------------------------//
void readsettings(void)
{ 
  FILE* fp;
  int gotconfig=1;
  int space=0, ret=0, but=0,k;
  char name[FILENAMELENGTH]="";
  char value[FILENAMELENGTH]="";
  char initversion[80]="";
  const int LABEL_LENGTH = 20;
  const int COMMAND_LENGTH = 128;
  char tempstring[2*1024];

  sprintf(tempstring, "%s/xdata.ini", g.homedir);
  if((fp=fopen(tempstring,"rb"))==NULL ) gotconfig=0;
  else fclose(fp);

  initversion[0]=0;
  if(gotconfig)
  {  fp=fopen(tempstring,"rt");
     while(!feof(fp))
     { 
       fgets(name, 80, fp);
       if(feof(fp))break;
       space = strcspn(name," ");                // Position of 1st space
       ret = (int)(index(name, '\n') - name); // Position of final \n
       name[space]=0;
       name[ret]=0;
       strcpy(value, 1+name+space);
       if(strcmp(name,"version")==SAME) strcpy(initversion,value); 
       if(strcmp(name,"button")==SAME) 
       {   if(but>=g.nbuttons)
               fprintf(stderr,"xdata.ini: Error: button count mismatch, ignoring user button\n");
           else
           {   space = strcspn(value," ");           // Position of 1st space
               space = min(space,LABEL_LENGTH-2);    // Max. of 19 chars
               strncpy(g.button[but]->label, value, space);
               g.button[but]->label[space]=0;
               strncpy(g.button[but]->cmd, value+space+1, COMMAND_LENGTH);
               but++;
           }
       }
       if(strcmp(name,"nbuttons")==SAME) 
       {   g.nbuttons = max(0,atoi(value));
           g.button = new ButtonStruct*[g.nbuttons];
           for(k=0;k<g.nbuttons;k++)
           {   g.button[k] = new ButtonStruct;
               g.button[k]->label = new char[LABEL_LENGTH];
               g.button[k]->cmd = new char[COMMAND_LENGTH];
           }
       }
       if(strcmp(name,"rows")==SAME) g.rows=atoi(value);
       if(strcmp(name,"cols")==SAME) g.cols=atoi(value);
       if(strcmp(name,"xsize")==SAME) g.main_xsize=atoi(value);
       if(strcmp(name,"ysize")==SAME) g.main_ysize=atoi(value);
       if(strcmp(name,"print_command")==SAME) strcpy(g.printer_command,value);
       if(strcmp(name,"print_file")==SAME) strcpy(g.printer_filename,value);
       if(strcmp(name,"signif_digits")==SAME) g.signif=atoi(value);
       if(strcmp(name,"wavelet_path")==SAME) strcpy(g.waveletpath, value);
       if(strcmp(name,"gvfilename")==SAME) strcpy(g.gvfilename, value);
     }
     fclose(fp);
  }
  if(!gotconfig)
  {
       const int NBUTTONS=48;
       char buttonname[NBUTTONS][20] ={ "Read",   "Save",  
                                        "Edit",   "New",
                                        "Pipe",   "Print",
                                        "Mean",   "TTest",
                                        "Add",    "Mult", 
                                        "FFT",    "L.R.", 
                                        "nlr",    "Ints",  
                                        "FTest",  "Cmnd", 
                                        "Copy",   "Paste",
                                        "Insert", "Delete",
                                        "Up",     "Down",
                                        "Left",   "Right",
                                        "Refrsh", "Erase",
                                        "Graph",  "Gr.off",
                                        "PgUp",   "PgDn",
                                        "String", "Reload",
                                        "Formula","Quit",
                                        "Hex",     ""
                                      };
       char buttoncmd[NBUTTONS][20]  ={ "read",   "save",  
                                        "formula","length",
                                        "pipe",   "print",
                                        "stat",   "ttest", 
                                        "add",    "mult",  
                                        "fft",    "linreg", 
                                        "nlr",    "ramp",  
                                        "ftest",  "commands",
                                        "copy",   "paste",
                                        "insert", "delete",
                                        "up",     "down",
                                        "left",   "right",
                                        "refresh","erase",
                                        "graph",  "ungraph",
                                        "page_up","page_down",
                                        "string", "reload",
                                        "setformula","quit",
                                        "sethexmode",  ""
                                      };
       g.nbuttons = NBUTTONS;
       g.button = new ButtonStruct*[NBUTTONS];
       for(k=0;k<NBUTTONS;k++) 
       {   g.button[k] = new ButtonStruct;
           g.button[k]->label = new char[20];
           g.button[k]->cmd = new char[20];
           strcpy(g.button[k]->label, buttonname[k]);
           strcpy(g.button[k]->cmd, buttoncmd[k]);
       }
  }
}        


//-------------------------------------------------------------------//
// abouttheprogram                                                   //
//-------------------------------------------------------------------//
void abouttheprogram(void)
{
   char *tempstring = new char[1024];
   sprintf(tempstring, "xdata version %s\nCompiled on %s %s",
      g.version, __DATE__, __TIME__);
   message(tempstring);
   delete[] tempstring;
}

//-------------------------------------------------------------------//
// help                                                              //
//-------------------------------------------------------------------//
void help(int helptopic)
{
   char tempstring[16384];
   sprintf(tempstring,"To resize a column: Shift + Center mouse button\n\n\
\n\
Data file format\n\n\
1. Numbers are in plain text.\n\
2. Empty cells are specified by a dash ( - ) surrounded by spaces.\n\
3. To keep compatibility with the old format, lines starting with # are \n\
   column titles.\n\
4. If the first line in the file starts with a quotation mark, xdata uses\n\
   the first line as a list of column titles. This is the same behavior as\n\
   in R, which calls it a header. If the old-style column titles are present,\n\
   they are preserved.\n\
5. If the second line in the file starts with a quotation mark, xdata uses\n\
   the second line as a list of row titles.\n\
6. Column titles can be changed in menu or by hitting Ctrl-T.\n\
   Row labels can be changed in menu or by hitting Ctrl-R.\n\
7. Lines starting with ## are equations or comments and are put in the formula editor.\n\
8. Entries starting with #= are formulas.\n\
\n\
   Example: \n\n\
   %cdogs%c %ccats%c %chorses%c   = if first line, sets the column titles.\n\
   %cdogs%c %ccats%c %chorses%c   = if second line,sets the row titles.\n\
   %cdogs%c %ccats%c %chorses%c   = if subsequent lines = strings in spreadsheet.\n\
                              The strings must be in double quotes.\n\
   ##dogs cats horses       = Comment. If it is a valid command,\n\
                              it can be executed by hitting Shift-Enter.\n\
   #dogs                    = sets title column 1 as dogs.\n\
   #cats                    = sets title column 2 as cats.\n\
   #horses                  = sets title column 3 as horses.\n\
   #= 6 1 x[2][2]+x[3][3]   = puts the formula in row 6, column 1.\n\
   1.0   4.0  5             = first line of data \n\
   1.0   -    5             = line of data with missing number \n\
   1.0  %cwoof%c  5           = line of data containing a string\n\
\n\
If the file is intended to be a collection of strings, then the\n\
first line must start with a # to prevent the program from interpreting\n\
it as an R-style header.\n\
\n\
Graphs\n\n\
Graph drawing uses gnuplot, a crude graph drawing program. Xdata\n\
feeds data to gnuplot through a Unix pipe (popen). Gnuplot must be\n\
closed manually when finished.\n\n\
\n\
Formula examples\n\n\
Cells are specified in row, column order. For example:\n\
   x[row][col] \n\
   sum(x[row1][col1]:x[row2][col2]) \n\
   sum(row1 col1 : row2 col2) \n\n\
Math syntax: y1=y2+y3  where 1,2, and 3 are column numbers\n\
  You can define variables, e.g. myvariable=1.414\n\
  To execute current line in the Editor, hit Shift-Enter.\n\
  To execute all lines in the Editor, click Execute.\n\n\
\n\
Selecting and Copying data\n\n\
\n\
Click and drag left mouse button on a group of cells to select data.\n\
Click middle mouse button to paste the selected cells into current position.\n\
Some operations require selecting a column. To do this, click on the column title.\n\n\
\n\
Evaluate equation - Advanced Math\n\n\
1. Put x values in one column, parameters a0 . . . an in second\n\
   column, and specify a third column for the results.\n\
2. Select an equation from the list, or click on a blank space in the\n\
   list and enter a new equation in the Information box where it \n\
   says Enter value.\n\
3. Set the number of parameters to use. \n\
   e.g., if you set it to 2, Xdata will only use the first two\n\
   rows in the parameters column.\n\
4. It will calculate a new set of data using the selected equation.\n\n\
\n\
Poisson  - Statistics\n\n\
Poisson calculations are done with the Gnu gmp high precision library.\n\
Results are shown in a text window, as these numbers don't fit in the\n\
editor. They are also printed on stdout.\n\n\
\n\
Anova  - Statistics\n\n\
1. Put three variables in three columns\n\
   Example: dogs.dat\n\
   %cdogs%c %ccats%c %c horses%c \n\
    1.0   4.0  5\n\
    1.2   4.2  7\n\
    1.1   4.1  6\n\
    1.0   4.2  6\n\
    1.1   4.2  5\n\
    1.2   4.2  4\n\
2. Select all three colums by clicking on title \n\
3. Select One-way ANOVA\n\
4. Result should be 8.38438e-09\n\
Compare with result from R: \n\n\
summary(aov(weight~animal))\n\
            Df Sum Sq Mean Sq F value   Pr(>F)    \n\
animal       2  60.97  30.485   82.02 8.38e-09 ***\n\
Residuals   15   5.57   0.372\n\n\
F-test\n\n\
1. Select two columns\n\
2. F-test\n\
F ratio is the ratio of the variances.\nProbability is the \n\
same regardless of whether you use f or 1/f.",0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22,0x22);
if(strlen(tempstring) >=16384) { message((char*)"Help screen too big"); 
return;}
   edit((char*)"Help",(char*)"Help",tempstring,21,30,1024,0);
}

//-------------------------------------------------------------------//
// manual                                                            //
//-------------------------------------------------------------------//
void manual(void)
{
   system("nedit /usr/local/bin/xdata-manual.txt");
}


//-------------------------------------------------------------------//
// setcoltitle                                                       //
// sets one column title, must call setcoltitles to complete         //
//-------------------------------------------------------------------//
void setcoltitle(int col, char *title)
{
   if(!between(col,0,s.cols))return;
   strcpy(s.title[col], title); 
   s.titlecols = max(s.titlecols, col);
}

//-------------------------------------------------------------------//
// setallcoltitles                                                   //
// sets all column titles, no way to set just one title              //
// string array s.title must be filled before calling                //
// time consuming. will make screen blink if called repeatedly       //
//-------------------------------------------------------------------//
void setallcoltitles(char **title)
{
   XtVaSetValues(s.spreadsheet_widget, XmNcolumnLabels, s.title, NULL); 
}

//-------------------------------------------------------------------//
// setrowtitle                                                       //
//-------------------------------------------------------------------//
void setrowtitle(int row, char *title)
{
   if(!between(row, 0, s.rows))return;
   strcpy(s.rowlabel[row], title); 
   XtVaSetValues(s.spreadsheet_widget, XmNrowLabels, s.rowlabel, NULL); 
   s.titlerows = max(s.titlerows, row);
}

//-------------------------------------------------------------------//
// changerowtitle                                                    //
//-------------------------------------------------------------------//
void changerowtitle(char **arg, int noofargs)
{
   int k;
   char title[128];
   char tempstring[128];
   int row = s.current_row;
   sprintf(tempstring, "%d", row+1);
   if(noofargs<2) 
   {   //  message((char*)"Column no.", tempstring, PROMPT, 1023);
       //  col = atoi(tempstring)-1;
       row = s.current_row;
   }else row = atoi(arg[1]);
   if(!between(row, 0, s.rows))return;
   strcpy(title, s.rowlabel[row]);
   if(noofargs<3) 
   {     message((char*)"Row Label", title, PROMPT, 1023); 
   }else strcpy(title, arg[2]);

   for(k=3; k<noofargs; k++) 
   {  strcat(title, " "); 
      strcat(title, arg[k]);
   }
   setrowtitle(row, title);
}

//-------------------------------------------------------------------//
// changetitle                                                       //
// user changes column title                                         //
//-------------------------------------------------------------------//
void changetitle(char **arg, int noofargs)
{
   int k;
   char title[128];
   char tempstring[128];
   int col = s.current_col;
   sprintf(tempstring, "%d", col+1);
   if(noofargs<2) 
   {   //  message((char*)"Column no.", tempstring, PROMPT, 1023);
       //  col = atoi(tempstring)-1;
       col = s.current_col;
   }else col = atoi(arg[1]);
   if(!between(col,0,s.cols))return;
   strcpy(title, s.title[col]);
   if(noofargs<3) 
   {     message((char*)"Title", title, PROMPT, 1023); 
   }else strcpy(title, arg[2]);

   for(k=3; k<noofargs; k++) 
   {  strcat(title, " "); 
      strcat(title, arg[k]);
   }
   setcoltitle(col, title);
   setallcoltitles(s.title);
   s.header_exists = 1; 
}




//-------------------------------------------------------------------//
// set_formula                                                       //
//-------------------------------------------------------------------//
void set_formula(void)
{
   double answer;
   static char formula[1024]="";
   int row=s.current_row, col=s.current_col;
   if(s.form[col][row][0])strcpy(formula, s.form[col][row]);
   message((char*)"Formula", formula, PROMPT, 1023); 
   strcpy(s.form[col][row], formula);
   g.message_count = 0;
   answer = eval(formula);
   if(g.message_count) 
   {
         message((char*)"Invalid formula", ERROR);
         s.form[col][row][0] = 0;  // no good
         g.message_count=0; 
         g.getout = 1;
   }else
   {
         set_cell_value(row,col,answer);  //// formula must be valid at this point or crash
         s.formula[col][row] = 1; 
         s.string[col][row]  = 0; 
         s.number[col][row]  = 0; 
         s.hex[col][row]  = 0; 
   }
}


//-------------------------------------------------------------------//
// sethexmode  toggle hex mode                                       //
//-------------------------------------------------------------------//
void sethexmode(void)
{
   g.number_type = 1 - g.number_type;
   refreshspreadsheet();
}
