//--------------------------------------------------------------------------//
// xdata10.cc                                                               //
// Latest revision: 08-03-2023                                              //
// Wavelets                                                                 //
//--------------------------------------------------------------------------//

#include "xdata.h"

extern Globals g;
extern Spreadsheet s;
char logic_operand[1024] = "ffff";


//--------------------------------------------------------------------------//
// findpeaks                                                                //
//--------------------------------------------------------------------------//
void findpeaks(char **arg, int noofargs)
{
  int j,k,kk;
  static int mx=1;               // column for data   
  static int my=s.maxcol;        // column for data   
  static int mp=s.maxcol;        // column for positions
  static int ma=s.maxcol+1;      // column for areas  
  static int smoothing=5;        // amount to smooth to eliminate noise
  static char minpeakareastring[128] = "0.00";  // minimum peak area

  int count, end, osign=0, peak=0, sign=0, peakcount=0, opeakcount;
  double *tdata, *dy, *d2y, *area, *yy, *maxima;
  double minpeakarea;
  int *xafter, *xx;
  char tempstring[100]; 

  ////------------------------////
  //// Start of menu section  ////

  Dialog dialog;
  dialog.needrepeat = 0;  

  strcpy(dialog.title,"Find Peaks");
  strcpy(dialog.boxes[1],"Column for x (0=integers)");
  strcpy(dialog.boxes[2],"Column for y   ");
  strcpy(dialog.boxes[3],"Column for peak positions");
  strcpy(dialog.boxes[4],"Column for peak areas");
  strcpy(dialog.boxes[5],"Minimum peak area");
  strcpy(dialog.boxes[6],"Smoothing");

  dialog.boxtype[0]=LABEL;
  strcpy(dialog.boxes[0],"Parameters");

  dialog.boxtype[1]=INTCLICKBOX;
  sprintf(dialog.answer[1],"1");    //// x
  dialog.boxmin[1]=1; 
  dialog.boxmax[1]=s.cols;

  dialog.boxtype[2]=INTCLICKBOX;    //// y
  sprintf(dialog.answer[2],"%d",s.current_col);
  dialog.boxmin[2]=1; 
  dialog.boxmax[2]=s.cols;

  dialog.boxtype[3]=INTCLICKBOX;    ////  peak x location
  sprintf(dialog.answer[3],"%d",mp);
  dialog.boxmin[3]=1; 
  dialog.boxmax[3]=s.cols;

  dialog.boxtype[4]=INTCLICKBOX;    ////  peak area
  sprintf(dialog.answer[4],"%d",ma);
  dialog.boxmin[4]=1; 
  dialog.boxmax[4]=s.cols;

  dialog.boxtype[5]=STRING;         //// min area for peak 
  sprintf(dialog.answer[5], minpeakareastring);

  dialog.boxtype[6]=INTCLICKBOX;    ////  peak area
  sprintf(dialog.answer[6],"%d",smoothing);
  dialog.boxmin[6]=1; 
  dialog.boxmax[6]=21;

  for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
  for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
  for(j=0;j<20;j++) dialog.boxcomeback[j]=0;

  dialog.noofradios=0;
  dialog.noofboxes=7;
  dialog.helptopic=0;
  dialog.keep=0;
  dialog.needrepeat=0;
  dialog.want_changecicb = 0;
  dialog.f1 = null;
  dialog.transient = 1;
  strcpy(dialog.path, ".");


  ///////---///////
  dialogbox(dialog);
  ///////---///////


  if(g.getout){ g.getout=0; return; }

  mx = atoi(dialog.answer[1]);
  my = atoi(dialog.answer[2]);
  mp = atoi(dialog.answer[3]);
  ma = atoi(dialog.answer[4]);
  strcpy(minpeakareastring, dialog.answer[5]);
  smoothing = atoi(dialog.answer[6]);
  minpeakarea = atof(minpeakareastring);


  ////  End of menu section   ////
  ////------------------------////
  ////  findpeaks             ////


   count = s.maxrow[my];
   tdata = new double[count+10];              // Temporary smoothed data
   dy  = new double[count+10];                // 1st deriv               
   d2y = new double[count+10];                // 2nd deriv
   xx  = new int[count+10];                   // x pos of peak k
   yy  = new double[count+10];                //
   maxima = new double[count+10];

   xx[0] = 0;
   yy[0] = s.x[my][0];
   
   for(k=0; k<count; k++) tdata[k] = s.x[my][k];
   smooth_baseline(tdata, count, 3);

   ////  Count the number of peaks
                                             // Take deriv.
   for(k=1; k<=count; k++) dy[k] = tdata[k] - tdata[k-1];
   for(k=1; k<=count; k++) d2y[k] = dy[k] - dy[k-1];
   if(smoothing>=3)
   {   smooth_baseline(dy, count, smoothing);
       smooth_baseline(d2y, count, smoothing);
   }
   for(k=2; k<count-2; k++)
   {                                         // If a maximum
       if(sgn(dy[k+1])== 1) sign=1;  
       if(sgn(dy[k+1])==-1) sign=-1;  
       if( sign!=osign && sgn(d2y[k])==-1 && k>3 )
       {   maxima[peak] = k;
           peak++;
           if(peak>=count){ message((char*)"Too many peaks",ERROR); break; }
           xx[peak] = k;
           yy[peak] = s.x[my][k];    // Use original data for y at max.
           k++;
       }    
       osign = sign;
   }

   ////  Find start, end, maximum, and area of each peak
 
   area = new double[peak+10];               // Peak areas
   xafter  = new int[peak+10];               // x position of end of peak


   xafter[0] = 0;
   for(k=1; k<=peak; k++)
   {   area[k-1]=0;
       osign=1;

       ////  Peaks are numbered from 0 to peak-1
       ////  Measure left half of peak, stop at end of previous peak

       for(j=xx[k]; j>xx[k-1]; j--)
       {   sign = sgn(dy[j]);  
//           dx = fabs( s.x[my][j+1] -  s.x[my][j]);
           area[k-1] += s.x[my][j];
           if(xx[k]==0) break;    
           if(sign!=osign && j<xx[k]-3) break;
       }                                     // At this point start of peak=j    
       osign=-1;
       if(k<peak) end=xx[k+1]; else end=count;

       ////  Measure right half of peak, stop at start of next peak

       for(j=xx[k]+1; j<end; j++)
       {   sign = sgn(dy[j]); 
//           dx = fabs( s.x[my][j+1] -  s.x[my][j]);
           area[k-1] += s.x[my][j];
           xafter[k] = j;
           if(xx[k]==0) break;    
           if(sign!=osign && j>xx[k]+3) break;
       }                                    // At this point end of peak=xafter[k]
   }   


   kk=0;
   peakcount = 0;
   opeakcount = s.maxrow[mp];
   for(k=0; k<peak; k++)
   {    if(area[k] < minpeakarea) continue;        
        peakcount++;
        kk++;
        s.x[ma][kk] = area[k];
        if(mx==-1)
            s.x[mp][kk] = maxima[k];
        else
            s.x[mp][kk] = s.x[mx][cint(maxima[k])];
//printf("%d %g %g\n",kk, s.x[mx][cint(maxima[k])], area[k]);
        s.number[ma][kk] = 1;
        s.string[ma][kk] = 0;
        s.hex[ma][kk] = 0;
        s.number[mp][kk] = 1;
        s.string[mp][kk] = 0;
        s.hex[mp][kk] = 0;
   }
   for(k=peakcount+1; k<=opeakcount; k++)
   {   
         set_cell(k, ma, (char*)"");
         set_cell(k, mp, (char*)"");
   }

   sprintf(tempstring, "%d peaks detected", kk);
   message(tempstring);

   s.maxcol = max(s.maxcol, 1+max(ma, mp));
   s.maxrow[ma] = peakcount;
   s.maxrow[mp] = peakcount;
   setcoltitle(ma, (char*)"Area");
   setcoltitle(mp, (char*)"Peak");

   delete[] area;
   delete[] xafter;   
   delete[] maxima;
   delete[] yy; 
   delete[] xx;
   delete[] d2y;
   delete[] dy;
   refreshspreadsheet();
}




//--------------------------------------------------------------------------//
// logic                                                                    //
//--------------------------------------------------------------------------//
void logic(void)  
{  
     static Dialog dialog;
     static int y1=1,y2=2,y3=3;
     int j,k;
 
     ////------------------------////
     //// Start of menu section  ////
     ////------------------------////
     dialog.needrepeat = -1;  
     strcpy(dialog.title,"Logic operations");
     strcpy(dialog.radio[0][0],"Convert to");
     strcpy(dialog.radio[0][1],"Base 10");
     strcpy(dialog.radio[0][2],"Base 16");

     strcpy(dialog.radio[1][0],"Operands");
     strcpy(dialog.radio[1][1],"Set + constant");
     strcpy(dialog.radio[1][2],"Two sets");

     strcpy(dialog.radio[2][0],"Operator");
     strcpy(dialog.radio[2][1],"Add +");
     strcpy(dialog.radio[2][2],"Subtract -");
     strcpy(dialog.radio[2][3],"Multiply *");
     strcpy(dialog.radio[2][4],"Divide   /");
     strcpy(dialog.radio[2][5],"And &");
     strcpy(dialog.radio[2][6],"Or |");
     strcpy(dialog.radio[2][7],"Xor ^");
     strcpy(dialog.radio[2][8],"Not ~");
     strcpy(dialog.radio[2][9],"Mod %");

     dialog.radiono[0] = 3;
     dialog.radiono[1] = 3;
     dialog.radiono[2] = 10;
     dialog.radioset[0] = -1;
     dialog.radioset[1] = -1;
     dialog.radioset[2] = -1;
     for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;

     dialog.boxtype[0]=LABEL;
     strcpy(dialog.boxes[0],"Data sets");

     dialog.boxtype[1]=INTCLICKBOX;
     strcpy(dialog.boxes[1],"Set #1");
     sprintf(dialog.answer[1],"%d", y1);
     dialog.boxmin[1]=1; 
     dialog.boxmax[1]=s.cols;

     dialog.boxtype[2]=INTCLICKBOX;
     strcpy(dialog.boxes[2],"Set #2");
     sprintf(dialog.answer[2],"%d",y2);
     dialog.boxmin[2]=1; 
     dialog.boxmax[2]=s.cols;

     dialog.boxtype[3]=INTCLICKBOX;
     strcpy(dialog.boxes[3],"Result");
     sprintf(dialog.answer[3],"%d",y3);
     dialog.boxmin[3]=1; 
     dialog.boxmax[3]=s.cols;

     dialog.boxtype[4]=STRING;
     strcpy(dialog.boxes[4],"Constant");
     sprintf(dialog.answer[4],"%s",logic_operand);

     for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
     for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
     for(j=0;j<20;j++) dialog.boxcomeback[j]=0;

     dialog.noofradios=3;
     dialog.noofboxes=5;
     dialog.helptopic=0;
     dialog.keep=0; 
     dialog.needrepeat=0;
     dialog.want_changecicb = 0;
     dialog.f1 = logic_check;
     dialog.transient = 1;
     strcpy(dialog.path, ".");

     ///////---///////
     dialogbox(dialog);
     ///////---///////
     
}

//--------------------------------------------------------------------------//
// logic_check                                                              //
//--------------------------------------------------------------------------//
void logic_check(Dialog *a, int radio, int radiobutton, int box, int boxbutton)
{
   int y1, y2, y3,  kk;
   static int noofoperands=-1, op=-1;
   char tempstring[1024];
   long long int operand;   

   switch (radio)
   {
        case -1: return;
        case  0: noofoperands=1; op=radiobutton; break;
        case  1: noofoperands=radiobutton; op=-1; return;
        case  2: op=radiobutton; break;
   }

   y1   = atoi(a->answer[1]);
   y2   = atoi(a->answer[2]);
   y3   = atoi(a->answer[3]);  //// result   
   strcpy(logic_operand, a->answer[4]);
   operand = strtoll(logic_operand, NULL, 16);

   if(radio == 0) //// convert set
   {  
       for(kk=1; kk<=s.maxrow[y1]; kk++)
       {   
           if(s.hex[y1][kk])
           {
               switch (op)
               { 
                  case 1: //// convert to dec result is bigger, so must pass a copy
                          hextointeger(s.text[y1][kk], tempstring);
                          strcpy(s.text[y1][kk], tempstring);
                          strcpy(s.disp[y1][kk], s.text[y1][kk]);
                          break;
                  case 2: //// convert to hex 
                          integertohex(s.text[y1][kk], tempstring);
                          strcpy(s.text[y1][kk], tempstring);
                          strcpy(s.disp[y1][kk], s.text[y1][kk]);
                          break;
               }
           }
       }
   }
   if(radio == 1) //// set one or 2 operands
       noofoperands = radiobutton;   
   if(radio == 2) //// do some math 
       long_integer_math(y1, y2, y3, noofoperands, operand, op);
   refreshspreadsheet();
}


//--------------------------------------------------------------------------//
// long_integer_math                                                        //
// for two columns          for one column   for both                       //
// y1=source column         y1=column        iii3=answer                    //
// y2=operand column        y2=operand                                      //
// y3=result                y3 not used                                     //
// op=operator              op=operator                                     //
//--------------------------------------------------------------------------//
void long_integer_math(int y1, int y2, int y3, int noofoperands, long long int operand, int op)
{
   int kk, destcol = 1;
   long long int iii1=0, iii2=0, iii3=0;
   enum { NUL, ADD, SUB, MUL, DIV, AND, OR, XOR, NOT, MOD }; //// must match menu above

   for(kk=1; kk<=s.maxrow[y1]; kk++)
   {   
       if(noofoperands == 1)
       {     iii2 = operand;
             destcol = y1;
       }
       else
       {     iii2 = strtoll(s.text[y2][kk], NULL, 16);
             destcol = y3;
       }
       if(s.hex[y1][kk])
       {
             iii1 = strtoll(s.text[y1][kk], NULL, 16);
             switch (op) 
             {  
                 case ADD:  iii3 = iii1 + iii2; break;
                 case SUB:  iii3 = iii1 - iii2; break;
                 case MUL:  iii3 = iii1 * iii2; break;
                 case DIV:  if(iii2 !=0) iii3 = iii1 / iii2; break;
                 case AND:  iii3 = iii1 & iii2; break;
                 case OR:   iii3 = iii1 | iii2; break;
                 case XOR:  iii3 = iii1 ^ iii2; break;
                 case NOT:  iii3 = iii1 &~iii2; break;
                 case MOD:  iii3 = iii1 % iii2; break;
             }
       }
       sprintf(s.text[destcol][kk], "%llx", iii3);
   }
   for(kk=1; kk<=s.maxrow[y1]; kk++)
   {   s.maxrow[destcol] = s.maxrow[y1];
       s.hex[destcol][kk]    = 1;
       s.string[destcol][kk] = 0;
       s.number[destcol][kk] = 0;
       s.maxcol = max(s.maxcol, 1 + destcol);
       XbaeMatrixSetCellBackground(s.spreadsheet_widget, kk, destcol, g.hexcolor);
   }
}
