//--------------------------------------------------------------------------//
// xdata11.cc                                                               //
// Latest revision: 08-02-2023                                              //
// Spreadsheet creation and handling                                        //
//--------------------------------------------------------------------------//

#include "xdata.h"

extern Globals g;
extern Spreadsheet s;
extern int selecting;
extern int mouse_button;
extern int inhibit_select;
extern int format;


//--------------------------------------------------------------------//
//  modifycb                                                          //
//--------------------------------------------------------------------//
void modifycb(Widget w, XtPointer client_data, XBMMVCB *call_data)
{
   int backspace = 0;
   w=w; client_data=client_data; 
   XBMMVCB *ptr = (XBMMVCB *)call_data;
   XmTextBlock block = ptr->verify->text;
   int insertsize = block->length;
   XmTextPosition insertpos = ptr->verify->newInsert;
   int row = ptr->row, col = ptr->column, olen, pos;
   static char text[100];
   char *newtext = block->ptr;
   //   char *oldtext = (char*)ptr->prev_text;
   char *oldtext = (char *)XbaeMatrixGetCell(w, row, col);
   insertpos = strlen(oldtext);
   ptr->verify->newInsert = insertpos+1;
   text[0] = 0;
 
   ////  There is no way to specifically detect a backspace in Xbae's
   ////  widget, so if newtext is null assume it was a backspace.
   ////  A lot of futzing around is needed to get the text while typing 
   ////  instead of after it leaves the cell.   Otherwise, if a menu is 
   ////  called while editing a cell, the most recent character will be lost.  

   ////  DO NOT USE xbae 4.6.60, backspace is messed up. Must use old
   ////  4.6.2 version in /usr/X11R6/lib

   ////  I have absolutely no idea why these work
   ////  Only works like this, not in a switch

   //// There is no way to set the correct background color
   //// until you leave the cell

   if(newtext != NULL && newtext[0] == 5)                 // Ctrl-E
   {   
          execute((char*)"formula", NULL, 0);
          newtext[0] = 32;
          insertsize = 0;
          return;
   }
   if(newtext != NULL && newtext[0] == 17)                 // Ctrl-Q
   {   
          execute((char*)"quit", NULL, 0);
          newtext[0] = 32;
          insertsize = 0;
          return;
   }
   if(newtext != NULL && newtext[0] == 18)                 // Ctrl-R
   {   
          execute((char*)"changerowtitle", NULL, 0);
          newtext[0] = 32;
          insertsize = 0;
          return;
   }
   if(newtext != NULL && newtext[0] == 19)                 // Ctrl-S
   {   
          execute((char*)"save", NULL, 0);
          newtext[0] = 32;
          insertsize = 0;
          return;
   }
   if(newtext != NULL && newtext[0] == 20)                 // Ctrl-T
   {   
          execute((char*)"title", NULL, 0);
          newtext[0] = 32;
          insertsize = 0;
          return;
   }


   if(oldtext==NULL) olen=0; else olen=(int)strlen(oldtext);

   if((int)strlen(oldtext) >= CELLBYTES) return;           // more than 256 chars

   if(insertpos==0)                                        // At start
   {      if(insertsize && newtext != NULL) strcpy(text, newtext);
          if(olen) strcat(text, oldtext);
   }else  if(insertpos < olen)                              // Insert in middle
   {      pos = insertpos + insertsize - 1;
          if(olen){ strcpy(text, oldtext); text[pos]=0; }
          if(insertsize && newtext != NULL) strcat(text,newtext);
          if(olen) strcat(text, oldtext + insertpos);
   }
   else                                                   // Append
   {      if(olen) strcpy(text, oldtext);
          if(insertsize && newtext != NULL) strcat(text, newtext);  // add a character
          else 
          {   text[max(0, (int)strlen(text))-1]=0; 
              backspace = 1;
//XmTextPosition pos = XmTextGetCursorPosition(w);
//XmTextSetInsertionPosition(s.spreadsheet_widget,(XmTextPosition)4);
//printf("%d\n",pos);
//return;
          } // else backspace
   }

   if(g.string)
   {      s.number[col][row] = 0;
          s.string[col][row] = 1;
          s.hex[col][row] = 0;
   }else if(g.number_type == DEC)
   {      s.number[col][row] = 1;
          s.string[col][row] = 0;
          s.hex[col][row] = 0;
   }else if(g.number_type == HEX)
   {      s.number[col][row] = 0;
          s.string[col][row] = 0;
          s.hex[col][row] = 1;
   }

   if(insertsize && newtext!=NULL) newtext[0] = 0;
   strcpy(s.text[col][row], text);

   //// put number in spreadsheet in case they select column
   //// but don't evaluate it
   double value=0;
   value = hextonumber(s.text[col][row]);   //// atof
   if(s.number[col][row]) set_cell_value(row, col, value);
   if(s.hex[col][row]) strcpy(s.disp[col][row], s.text[col][row]);

   s.rows = max(row, s.rows);
   s.cols = max(col, s.cols);
   s.maxrow[col] = max(s.maxrow[col], row);
   s.maxcol = max(s.maxcol, col);
   ptr->verify->newInsert = strlen(text);

   XbaeMatrixSetCell(s.spreadsheet_widget, row, col, text);
   XtVaSetValues(s.spreadsheet_widget, XmNcursorPosition, 4, NULL); 

   //// get around bug in xbae where backspacing puts cursor at left.
   if(backspace)
   {
      XbaeMatrixEditCell(s.spreadsheet_widget,row+1,col);
      XbaeMatrixEditCell(s.spreadsheet_widget,row,col);
   }
   g.touched = 1;
//printf("aaa %g\n",s.x[col][row]); 
}


//--------------------------------------------------------------------//
//  leavecellcb                                                       //
//--------------------------------------------------------------------//
void leavecellcb(Widget w, XtPointer client_data, XBMLCCB *call_data)
{
   selecting = 0;
   w=w; client_data=client_data; 
   double value=0;
   int oselected;
   int row = s.current_row;
   int col = s.current_col;   
   g.getout = 0;
   char tempstring[1024];

   if(call_data->value!=NULL && strlen(call_data->value))
   {    row = call_data->row;
        col = call_data->column;
        oselected = s.selected[col][row];
        select_cell(row,col,False);

        if(s.hex[col][row])
        {
//              value = hextonumber(s.text[col][row]);
        }else
        {
              strcpy(tempstring,s.text[col][row]);
              value = eval(tempstring);  //// evaluate in dec mode only
              format_number(value, s.text[col][row]);    //// number to string
        }

        strcpy(s.disp[col][row], s.text[col][row]);

        if(g.getout) 
             g.getout=0; 
        else 
        {    if(oselected) select_cell(row,col,True);
             if(s.number[col][row] && row>0 && col>0) 
             {  
                 set_cell_value(row, col, value);
             }else if(strcmp(" ", s.text[col][row])) // if not skip
             {
                 set_cell_text(row,col,s.text[col][row]);
             }
             set_background(s.spreadsheet_widget, row, col);
        }
   }
   if(strlen(call_data->value)==0){ s.number[col][row]=0; s.text[col][row][0]=0; }

}


//--------------------------------------------------------------------//
//  set_background                                                    //
//--------------------------------------------------------------------//
void set_background(Widget w, int row, int col)
{
   if(s.form[col][row][0])
       XbaeMatrixSetCellBackground(s.spreadsheet_widget, row, col, g.ecolor);
   else if(s.string[col][row])
       XbaeMatrixSetCellBackground(s.spreadsheet_widget, row, col, g.scolor);
   else if(s.hex[col][row])
   {
       XbaeMatrixSetCellBackground(s.spreadsheet_widget, row, col, g.hexcolor);
   }
   else
       XbaeMatrixSetCellBackground(s.spreadsheet_widget, row, col, g.bcolor);
   //// set row, col labels to purple
   if(row==0 || col==0)
       XbaeMatrixSetCellBackground(s.spreadsheet_widget, row, col, g.main_fixedcolor);
   XFlush(g.display);
}

//--------------------------------------------------------------------//
//  format_number                                                     //
//  convert numbers into strings depending on base to display         //
//  gives an error for everything, reason unknown                     //
//--------------------------------------------------------------------//
void format_number(double xx, char *result)
{
     switch(g.number_type)
     {
         case DEC:
//              sprintf(result, "%-13.13g", xx);
//            sprintf(result, "%'-13.13g", xx); will add commas
              gcvt(xx, 99, result);
              break;
         case OCT:
              sprintf(result, "%o", (int)xx);
              break;
         case HEX:
              sprintf(result, "%lx", (long unsigned int)xx);
              break;
         default:
              break;
     }
}

//--------------------------------------------------------------------//
//  integertohex                                                      //
//  convert hex strings into numbers, only if in hex mode             //
//--------------------------------------------------------------------//
void integertohex(char *decstring, char *hexstring)
{
   long long int kkk;
   kkk = strtoll(decstring, NULL, 10);
   sprintf(hexstring, "%llx", kkk);
}

//--------------------------------------------------------------------//
//  hextointeger                                                      //
//  convert hex strings into base 10 numbers, only if in hex mode     //
//--------------------------------------------------------------------//
void hextointeger(char *hexstring, char *decstring)
{
   static char tempstring[1000];
   long long int kkk;
   sprintf(tempstring, "0x%s", hexstring);
   kkk = strtoll(tempstring, NULL, 16);
   sprintf(decstring, "%lld", kkk);
}


//--------------------------------------------------------------------//
//  hextonumber                                                       //
//  convert hex strings into numbers, only if in hex mode             //
//  replacement for atof                                              //
//--------------------------------------------------------------------//
double hextonumber(char *hexstring)
{
   static char tempstring[1000];
   double value;
   long long int kkk;

   if(g.number_type == HEX)
   {     sprintf(tempstring, "0x%s", hexstring);
         kkk = strtoll(tempstring, NULL, 16);
         value = (double)kkk;
//printf("hextonumber %g %d %lld %s\n",value, (int)value, (long long int)value, tempstring);
         if( value >= 9.22337e+18 )
         {   message((char*)"Value too large");
             printf("Value exceeds range of data type\n");
         }
   }else
   {     value = atof(hexstring);
   }
   return value;
}

//--------------------------------------------------------------------//
//  add_prefix  add prefix to string. string must be big enough       //
//--------------------------------------------------------------------//
void add_prefix(char *prefix, char *string)
{
   static char tempstring[1000];
   strcat(tempstring, string);
   sprintf(string, "%s%s", prefix, tempstring);
}

//--------------------------------------------------------------------//
//  remove_prefix  remove prefix of specified length from string.     //
//--------------------------------------------------------------------//
void remove_prefix(int length, char *string)
{
   int kk;
   for(kk=0; kk<length; kk++) string[kk] = string[kk+length];
   string[(int)strlen(string) - length] = 0;
printf("string >>%s<<\n", string);
}
