//--------------------------------------------------------------------------//
// xdata2.cc                                                                //
// Latest revision: 07-31-2023                                              //
// Spreadsheet creation and handling                                        //
//--------------------------------------------------------------------------//

#include "xdata.h"

extern Globals g;
extern Spreadsheet s;
char filename[FILENAMELENGTH]="Unknown.dat";
int selecting=0;
int mouse_button=0;
int inhibit_select=0;
int format=1;
static int want_print_to_file = 0;
int mouse_mask = KeyPressMask |         // Keyboard
                 StructureNotifyMask |
                 SubstructureNotifyMask |
                 ColormapChangeMask |
                 EnterWindowMask |
                 LeaveWindowMask |   
                 Button1MotionMask |    // Click & Drag
                 PointerMotionMask;

//--------------------------------------------------------------------//
// create_spreadsheet                                                 //
//--------------------------------------------------------------------//
void create_spreadsheet(Widget parent)
{ 
  static char text_translations[] = "#override\
        <Key>osfUp: EditCell(Up)\n\
        <Key>osfRight: EditCell(Right)\n\
        <Key>osfLeft: EditCell(Left)\n\
        <Key>osfDown: EditCell(Down)\n\
        <Key>osfPageUp: PageUp()\n\
        <Key>osfPageDown: PageDown()\n\
        <Key>Enter: EditCell(Down)\n\
        <Key>Return: EditCell(Down)\n\
        <Key>KP_Up: EditCell(Up)\n\
        <Key>KP_Down: EditCell(Down)\n\
        <Key>KP_Left: EditCell(Left)\n\
        <Key>KP_Right: EditCell(Right)\n\
        <Key>Down: EditCell(Down)\n";
  int i,j,k,n,cols,rows;
  Arg args[100];
  Widget form;  
  Widget frame, button_drawing_area;
  Dimension mainmenuheight;
  int y1;
  rows=s.rows; cols=s.cols;
  XtTranslations translations;
  translations = XtParseTranslationTable(text_translations);
 
  XtVaGetValues(g.menubar, XmNheight, &mainmenuheight, NULL);       
  y1 = (int)mainmenuheight;
  n=0;
  XtSetArg(args[n], XmNtopAttachment, XmATTACH_WIDGET); n++;
  XtSetArg(args[n], XmNtopWidget, g.menubar); n++;
  XtSetArg(args[n], XmNwidth, g.main_xsize); n++;
  XtSetArg(args[n], XmNheight, g.main_ysize+10); n++;
  XtSetArg(args[n], XmNx, 0); n++;
  XtSetArg(args[n], XmNy, y1); n++;
  XtSetArg(args[n], XmNresizePolicy, XmRESIZE_ANY); n++;
  form = XmCreateForm(parent, (char*)"SpreadsheetForm", args, n);

  //// Make an extra row to park cursor when refreshing spreadsheet
  s.spreadsheet_widget = XtVaCreateManagedWidget("Spreadsheet",
                                xbaeMatrixWidgetClass,  form,
                                XmNtitle,               "Spreadsheet",
                                XmNleftAttachment,      XmATTACH_POSITION,
                                XmNleftOffset,          120,
                                XmNcolumns,             s.cols,
                                XmNrows,                s.rows,
                                XmNrightAttachment,     XmATTACH_FORM,
                                XmNtopAttachment,       XmATTACH_FORM,
                                XmNbottomAttachment,    XmATTACH_FORM,
                                XmNtopOffset,           20,
                                XmNbottomOffset,        30,
                                XmNfixedRows,           1, //1st row immovable
                                XmNfixedColumns,        1, //1st col immovable
                                XmNallowColumnResize,   True,
                                XmNcellMarginHeight,    0,
                                XmNcellMarginWidth,     0,
                                XmNtraverseFixedCells,  True,
                                XmNselectScrollVisible, False,
                                XmNgridType,            XmGRID_SHADOW_IN,
                                XmNhorizontalScrollBarDisplayPolicy, 
                                                        XmDISPLAY_STATIC,
                                XmNverticalScrollBarDisplayPolicy, 
                                                        XmDISPLAY_STATIC,                               
                                XmNtextTranslations,    translations,
                                XmNspace,               2,
                                XmNbuttonLabels,        True,
                                XmNcalcCursorPosition,  True,
                                NULL);
  
  //// For mainbuttoncb, the command must be a valid command for execute().
  //// Arbitrary no. of user-defined buttons from .ini

 //--------------------frame for buttons-----------------------------//

  n=0;
  XtSetArg(args[n], XmNshadowType, XmSHADOW_ETCHED_IN); n++;
  XtSetArg(args[n], XmNx, 1); n++;
  XtSetArg(args[n], XmNy, 2); n++;
  XtSetArg(args[n], XmNwidth, 117); n++;
  XtSetArg(args[n], XmNheight, 20*(g.nbuttons/2)); n++;
  frame = XmCreateFrame(form, (char*)"frame", args, n);

  //--------------info area (inside frame)---------------------------//

  n=0;
  XtSetArg(args[n], XmNtitle, "ButtonArea"); n++;
  XtSetArg(args[n], XmNresizePolicy, XmRESIZE_NONE); n++;
  XtSetArg(args[n], XmNx, 0); n++;
  XtSetArg(args[n], XmNwidth, 113); n++;
  XtSetArg(args[n], XmNheight, 20*(g.nbuttons/2)); n++;
  button_drawing_area = XmCreateDrawingArea(frame, (char*)"button_drawing_area", args, n);

  for(k=0;k<g.nbuttons;k++) 
      addmainbutton(button_drawing_area, g.button[k]->label, g.button[k]->cmd, 
          mainbuttoncb);

  XtManageChild(form);
  XtMapWidget(parent);
  XtManageChild(s.spreadsheet_widget);
  XtRealizeWidget(s.spreadsheet_widget);
  XtManageChild(frame);
  XtManageChild(button_drawing_area);

  s.spreadsheet_form_window = XtWindow(form);
  XtAddCallback(s.spreadsheet_widget,XmNmodifyVerifyCallback,(XtCallbackProc)modifycb, NULL);
  XtAddCallback(s.spreadsheet_widget,XmNmotionVerifyCallback,(XtCallbackProc)motioncb, NULL);
  XtAddCallback(s.spreadsheet_widget,XmNselectCellCallback,(XtCallbackProc)selectcb, NULL);
  XtAddCallback(s.spreadsheet_widget,XmNenterCellCallback,(XtCallbackProc)entercellcb, NULL);
  XtAddCallback(s.spreadsheet_widget,XmNleaveCellCallback,(XtCallbackProc)leavecellcb, NULL);
  XtAddCallback(s.spreadsheet_widget,XmNselectCellCallback,(XtCallbackProc)selectcb, NULL);
  XtAddCallback(s.spreadsheet_widget,XmNtraverseCellCallback,(XtCallbackProc)traversecellcb, NULL);
  XtAddCallback(s.spreadsheet_widget,XmNlabelActivateCallback,
       (XtCallbackProc)columnselectcb, NULL);
  XtAddEventHandler(s.spreadsheet_widget, g.mask, False, (XtEH)mousecb, (XtP)NULL);

  short int col_widths[cols+2];
  for(k=0;k<cols;k++) col_widths[k]=CELLWIDTH;
  col_widths[0] = 4;   //// no of digits for row numbers in 1st col
  XtVaSetValues(s.spreadsheet_widget, XmNcolumnWidths, col_widths, NULL);

  short int col_maxlengths[cols+2];
  for(k=0;k<cols;k++) col_maxlengths[k]=1024;
  XtVaSetValues(s.spreadsheet_widget, XmNcolumnMaxLengths, col_maxlengths, NULL);

  ////  Have to use 'beginning' because cell on cursor is always aligned left
  ////  regardless of the setting.
  uchar col_alignments[cols];
  for(k=0;k<cols;k++) col_alignments[k]=XmALIGNMENT_BEGINNING;
  XtVaSetValues(s.spreadsheet_widget, XmNcolumnAlignments, col_alignments, NULL);

  Boolean col_button_labels[cols];
  for(k=0;k<cols;k++)col_button_labels[k]=True;
  XtVaSetValues(s.spreadsheet_widget, XmNcolumnButtonLabels, col_button_labels, NULL);

  Boolean row_button_labels[rows];
  for(k=0;k<rows;k++)row_button_labels[k]=True;
  XtVaSetValues(s.spreadsheet_widget, XmNrowButtonLabels, row_button_labels, NULL);


  ////  s.text is a char[col][row][10]
  s.text = new char **[cols+2];
  s.disp = new char **[cols+2];
  s.form = new char **[cols+2];
  s.title = new char *[cols+2];
  s.otitle = new char *[cols+2];
  s.collabel = new char *[cols+2];
  s.rowlabel = new char *[rows+2];
  s.title = new char *[cols+2];
  s.otitle = new char *[cols+2];
  for(j=0;j<cols;j++)
  {    s.text[j] = new char *[rows+2];
       s.disp[j] = new char *[rows+2];
       s.form[j] = new char *[rows+2];
       s.title[j] = new char[FILENAMELENGTH];
       s.otitle[j] = new char[FILENAMELENGTH];
       s.collabel[j] = new char[FILENAMELENGTH];
       s.title[j][0] = 0;
       s.otitle[j][0] = 0;
       s.collabel[j][0] = 0;
       for(i=0;i<rows;i++)
       {
            s.text[j][i] = new char[CELLBYTES+1];
            memset(s.text[j][i], 0, CELLBYTES);
            s.disp[j][i] = new char[CELLBYTES+1];
            memset(s.disp[j][i], 0, CELLBYTES);
            s.form[j][i] = new char[CELLBYTES+1];
            memset(s.form[j][i], 0, CELLBYTES);
       }
  }
  for(j=0;j<rows;j++)
  {
       s.rowlabel[j] = new char[FILENAMELENGTH];
       s.rowlabel[j][0] = 0;
  }

  for(k=1;k<cols;k++) sprintf(s.title[k], "Col. %d",k);
  sprintf(s.title[0], "Row");

  for(k=1;k<cols;k++) sprintf(s.collabel[k], "%d",k);
  XtVaSetValues(s.spreadsheet_widget, XmNcolumnLabels, s.title, NULL);

  for(k=1;k<rows;k++) sprintf(s.rowlabel[k], "%d",k);
  XtVaSetValues(s.spreadsheet_widget, XmNrowLabels, s.rowlabel, NULL);

  ////  Flag if cell is selected
  s.selected_1d = new uchar[rows*cols];
  s.selected = make_2d_alias(s.selected_1d, rows, cols);       
  for(i=0;i<cols;i++)
  for(j=0;j<rows;j++) s.selected[i][j]=0;

  ////  Flag if cell contains a number
  s.number_1d = new uchar[rows*cols];
  s.number = make_2d_alias(s.number_1d, rows, cols);       

  ////  Flag if cell contains a string
  s.string_1d = new uchar[rows*cols];
  s.string = make_2d_alias(s.string_1d, rows, cols);       

  ////  Flag if cell contains a formula
  s.formula_1d = new uchar[rows*cols];
  s.formula = make_2d_alias(s.formula_1d, rows, cols);       

  ////  Flag if cell contains a hex number
  s.hex_1d = new uchar[rows*cols];
  s.hex = make_2d_alias(s.hex_1d, rows, cols);       

  ////  The number
  s.x_1d = new double [rows*cols];
  s.x = make_2d_alias(s.x_1d, rows, cols);       


  ////  Length of each column
  s.maxrow = new int[cols];

  ////  Flag if column is selected
  s.column_selected = new uchar[cols];
  for(i=0;i<cols;i++) s.column_selected[i] = 0;

  ////  Flag if row is selected
  s.row_selected = new uchar[rows];
  for(i=0;i<rows;i++) s.row_selected[i] = 0;

  XbaeMatrixSetCell(s.spreadsheet_widget,0,0,(char*)"");  // Initialize memory
  
  //// initialize fixed row and col
  for(j=0;j<rows;j++) set_cell_value(j, 0, j);
  for(j=0;j<rows;j++)
      XbaeMatrixSetCellBackground(s.spreadsheet_widget, 0, j, 
      g.main_fixedcolor);
  for(j=0;j<cols;j++) set_cell_value(0, j, j);
  for(j=0;j<rows;j++)
      XbaeMatrixSetCellBackground(s.spreadsheet_widget, j, 0, 
      g.main_fixedcolor);

  for(i=0;i<cols;i++) s.maxrow[i] = 0;
  s.maxcol = 0;        // No. of nonblank columns

  s.header_exists = 0;
  s.current_row = 1;
  s.current_col = 1;
}  


//--------------------------------------------------------------------//
//  clickcb                                                           //
//--------------------------------------------------------------------//
void clickcb(Widget w, XtPointer client_data, XBMTCCB *call_data)
{
    int row, col;
    XBMECB *ptr = (XBMECB *) call_data;
    row = ptr->row;
    col = ptr->column;

   //// if it's in column 0, select the row
   if(col == 0) 
   {   s.row_selected[row] = 1 - s.row_selected[row];
       select_row(row, s.row_selected[row]);
   }
} 


//--------------------------------------------------------------------//
//  entercellcb                                                       //
//--------------------------------------------------------------------//
void entercellcb(Widget w, XtPointer client_data, XBMECB *call_data)
{
    selecting = 0;
    w=w; client_data=client_data;
    XBMECB *ptr = (XBMECB *) call_data;
    show_coordinates(ptr->row, ptr->column);
    s.current_row = ptr->row;
    s.current_col = ptr->column;
    call_data->select_text = s.selected[ptr->column][ptr->row];

    int j,k;
    char command[1024];
    double answer;

    //// execute any formulas in the spreadsheet

    for(j=1;j<s.maxcol;j++) 
    for(k=1;k<s.maxrow[j];k++) 
    {
       if(!s.string[j][k] && s.form[j][k][0]) 
       {
            strcpy(command, s.form[j][k]);
            g.message_count = 0;  
            answer = eval(command);
            if(g.message_count) 
            {    printf("error\n");
                 s.form[j][k][0] = 0;
                 g.message_count = 0;  
            }else
                 set_cell_value(k, j, answer);  //// formula must be valid at this point or crash
       }
    }
}


//--------------------------------------------------------------------//
//  show_coordinates                                                  //
//--------------------------------------------------------------------//
void show_coordinates(int row, int col)
{
  char tempstring[2*FILENAMELENGTH];
  int len, noofcols;
  sprintf(tempstring,"Row %d  Col %d     ",row, col);
  len = strlen(tempstring);

  XDrawImageString(g.display, s.spreadsheet_form_window,g.gc,118,14,tempstring,len);
  sprintf(tempstring, "%s        ",g.filename);
  len = strlen(tempstring);
//  XDrawImageString(g.display, s.spreadsheet_form_window,g.gc,210,14,tempstring,len);


  //// print current number / string / formula in left panel

  if(s.formula[col][row]) strcpy(tempstring,"Formula ");  
  else if(s.number[col][row]) strcpy(tempstring,"Value     ");
  else if(s.string[col][row]) strcpy(tempstring,"String    ");  
  else if(s.hex[col][row]) strcpy(tempstring,"Hex       ");  
  else strcpy(tempstring, "            ");
  len = (int)strlen(tempstring);
  XDrawImageString(g.display, s.spreadsheet_form_window,g.gc,4,545,tempstring,len);

  if(s.number[col][row]) 
       sprintf(tempstring,"%g                    ",s.x[col][row]);
  else if(s.string[col][row]) 
       sprintf(tempstring,"%s                    ",s.text[col][row]); 
  else if(s.hex[col][row]) 
       sprintf(tempstring,"%s                    ",s.text[col][row]); 
  else if(s.formula[col][row]) 
       sprintf(tempstring,"%s                    ",s.form[col][row]); 
  else 
       strcpy(tempstring, "                         ");
  
  len = (int)strlen(tempstring);
  XDrawImageString(g.display, s.spreadsheet_form_window,g.gc,4,560,tempstring,len);

  //// Show average of all selected data  
  int j;
  double x=0.0, mean=0.0, sum=0.0, se=0.0, sd=0.0,
        p=0.0, xx=0.0, variance=0.0;
  double lowest=0.0, highest=0.0;
  SelectStruct data;
  data.s = NULL;
  SelectCountStruct sel;
  sel.start = NULL;
  sel.end = NULL;
  int count=0;
  count = count_selected_data(&sel, 0);
  if(count>0)
  {    data.x = new double [count];
       data.isnumber = new uchar[count];
       get_selected_data(&data, 0, count, noofcols);
       sum = 0.0;
       for(j=0;j<count;j++) sum += data.x[j];
       mean = sum / count;
       for(j=0;j<count;j++) 
       {     x = data.x[j];
             xx = x-mean;          
             variance += (p=xx*xx);
       }
       variance /= (count-1);
       sd = sqrt(variance);
       se = sd / sqrt(count);
  }
//sprintf(tempstring,"Mean=%10.10g SE=%10.10g N=%d  Total=%10.10g      SD=%10.10g                                           ",mean,se,count,sum,sd);
//  sprintf(tempstring,"N=%d  Total=%10.10g      Mean=%10.10g   SE=%10.10g   SD=%10.10g  ",count,sum,mean,se,sd);
  sprintf(tempstring,"N=%d           ",count );
  len = (int)strlen(tempstring);
  XDrawImageString(g.display, s.spreadsheet_form_window,g.gc,258,14,tempstring,len);

  sprintf(tempstring,"Total=%10.10g         ",sum   );
  len = (int)strlen(tempstring);
  XDrawImageString(g.display, s.spreadsheet_form_window,g.gc,328,14,tempstring,len);

  sprintf(tempstring,"Mean=%10.10g          ",mean  );
  len = (int)strlen(tempstring);
  XDrawImageString(g.display, s.spreadsheet_form_window,g.gc,478,14,tempstring,len);

  sprintf(tempstring,"SE=%10.10g            ",se    );
  len = (int)strlen(tempstring);
  XDrawImageString(g.display, s.spreadsheet_form_window,g.gc,598,14,tempstring,len);

  sprintf(tempstring,"SD=%10.10g            ",sd    );
  len = (int)strlen(tempstring);
  XDrawImageString(g.display, s.spreadsheet_form_window,g.gc,718,14,tempstring,len);

  range(col, lowest, highest);

  strcpy(tempstring,"Lowest");
  len = (int)strlen(tempstring);
  XDrawImageString(g.display, s.spreadsheet_form_window,g.gc,4,480,tempstring,len);

  sprintf(tempstring,"%g              ", lowest);
  len = (int)strlen(tempstring);
  XDrawImageString(g.display, s.spreadsheet_form_window,g.gc,4,496,tempstring,len);

  strcpy(tempstring,"Highest");
  len = (int)strlen(tempstring);
  XDrawImageString(g.display, s.spreadsheet_form_window,g.gc,4,512,tempstring,len);

  sprintf(tempstring,"%g              ", highest);
  len = (int)strlen(tempstring);
  XDrawImageString(g.display, s.spreadsheet_form_window,g.gc,4,528,tempstring,len);

}


//--------------------------------------------------------------------//
//  range                                                             //
//--------------------------------------------------------------------//
void range(int col, double &lowest, double &highest)
{
  //// calculate highest and lowest of current col
  int k;
  for(k=1; k<=s.maxrow[col]; k++) 
  {  if(s.number[col][k])
     {   lowest = highest = s.x[col][k];  //// get a starting value
         break;
     }  
  }
  for(k=1; k<=s.maxrow[col]; k++) 
  { 
     if(s.number[col][k])
     {   if(s.x[col][k] > highest) highest = s.x[col][k];
         if(s.x[col][k] < lowest) lowest = s.x[col][k];
     }  
  }
}




//--------------------------------------------------------------------//
//  motioncb                                                          //
//--------------------------------------------------------------------//
void motioncb(Widget w, XtPointer client_data, XBMMVCB *call_data)
{
  printf("motioncb\n");
}


//--------------------------------------------------------------------//
//  inputcb                                                           //
//--------------------------------------------------------------------//
void inputcb(Widget w, XtPointer client_data, XBMLCCB *call_data)
{
    w=w; client_data=client_data; call_data=call_data; 
printf("inputcb\n");
}


//--------------------------------------------------------------------//
//  drawcellcb                                                        //
//--------------------------------------------------------------------//
void drawcellcb(Widget w, XtPointer client_data, XBMDCCB *call_data)
{
    w=w; client_data=client_data; call_data=call_data; 
}


//--------------------------------------------------------------------//
//  changecb                                                          //
//--------------------------------------------------------------------//
void changecb(Widget w, XtPointer client_data, XmACB *call_data)
{
    w=w; client_data=client_data; call_data=call_data; 
printf("changecb\n");
}


//--------------------------------------------------------------------//
//  writecellcb                                                       //
//--------------------------------------------------------------------//
void writecellcb(Widget w, XtPointer client_data, XBMWCCB *call_data)
{
    w=w; client_data=client_data;call_data=call_data;  
printf("writecb\n");
}

//--------------------------------------------------------------------//
//  dragcb                                                            //
//--------------------------------------------------------------------//
void dragcb(Widget w, XtPointer client_data, XBMECB *call_data)
{
    w=w; client_data=client_data;call_data=call_data;
printf("dragcb\n");
}


//--------------------------------------------------------------------//
//  selectcb                                                          //
//--------------------------------------------------------------------//
void selectcb(Widget w, XtPointer client_data, XBMSCCB *call_data)
{
    w=w; client_data =client_data; call_data=call_data;
    int row,col;
    row = call_data->row;
    col = call_data->column;
    toggle_cell(row, col);
    s.column_selected[col]=0; 
}


//--------------------------------------------------------------------//
//  traversecellcb                                                    //
//--------------------------------------------------------------------//
void traversecellcb(Widget w, XtPointer client_data, XBMTCCB *call_data)
{
         w=w; client_data=client_data; call_data=call_data;
         int nextrow=call_data->next_row;
         int nextcol=call_data->next_column;
         int row=call_data->row;
         int col=call_data->column;

         ////  Prevent wrapping around
         if(nextrow==s.rows-1 && row==0) call_data->next_row = 0;
         if(nextcol==s.cols-1 && col==0) { call_data->next_column = 0; call_data->next_row++;} 
         if(nextrow==0 && row==s.rows-1) call_data->next_row = s.rows-1;
         if(nextcol==0 && col==s.cols-1) { call_data->next_column = s.cols-1; call_data->next_row--;}
}


//--------------------------------------------------------------------//
//  leavespreadsheet                                                  //
//--------------------------------------------------------------------//
void leavespreadsheet(void)
{
    ////  Edit a random cell to force Xbae to commit the changes by
    ////  calling leavecellcb.
    ////  If user doesnt move off the cell, s.text could be inconsistent
    ////  if user deleted characters after translating the Delete key to
    ////  delete-next-character().

    XbaeMatrixEditCell(s.spreadsheet_widget,s.current_row,s.current_col);
    XbaeMatrixCommitEdit(s.spreadsheet_widget, False);
    selecting = 0;

    ////  Recalculate no. of non-blank columns
    int k;
    s.maxcol = 0;
    for(k=1; k<s.cols; k++) if(s.maxrow[k]) s.maxcol = k;
    return;
}


//--------------------------------------------------------------------//
//  refreshspreadsheet                                                //
//--------------------------------------------------------------------//
void refreshspreadsheet(void)
{
   g.getout = 0;
   int i,j,r,c;
//   char tempstring[1024];
   r = s.current_row;
   c = s.current_col;
//   if(s.number[c][r]) gcvt(s.x[c][r], g.signif, tempstring);
   g.getout=0;
   for(j=1;j<=s.maxcol;j++)
   {     
         for(i=1;i<=s.maxrow[j];i++)
         {    
             if(s.number[j][i]) 
             {
                  format_number(s.x[j][i], s.disp[j][i]);
                  strcpy(s.text[j][i], s.disp[j][i]);
             }else
             { 
                   strncpy(s.disp[j][i], s.text[j][i], CELLBYTES);
             }
             XbaeMatrixSetCell(s.spreadsheet_widget,i,j,s.disp[j][i]);
             set_background(s.spreadsheet_widget, r, c);
         }
   }
   XbaeMatrixRefresh(s.spreadsheet_widget);

   ////  Move off of current cell then move back, so display is updated
   XbaeMatrixEditCell(s.spreadsheet_widget,r+1,c);
   XbaeMatrixEditCell(s.spreadsheet_widget,r,c);
   ////  Current cell must be set separately or it will revert to previous value

   //XbaeMatrixCommitEdit(s.spreadsheet_widget, False);
   if(s.number[c][r]) set_cell_value(r, c, s.x[c][r]);
   if(g.autoupdate && g.graph_active) graph(1, NULL, 0);
   setallcoltitles(s.title);
   g.touched = 1;
   return;
}


//--------------------------------------------------------------------//
//  columnselectcb                                                    //
//--------------------------------------------------------------------//
void columnselectcb(Widget w, XtPointer client_data, XBMLACB *call_data)
{
    w=w; client_data=client_data;call_data=call_data;  
    inhibit_select = 1;  // set to 0 on next button release in mousecb
    XbaeMatrixLabelActivateCallbackStruct *cbs =
	( XbaeMatrixLabelActivateCallbackStruct * )call_data;
    if( cbs->row_label )
        select_row(call_data->row, 1-s.row_selected[call_data->row]);
    else
        select_column(call_data->column, 1-s.column_selected[call_data->column]);
}


//--------------------------------------------------------------------//
//  select_column                                                     //
//--------------------------------------------------------------------//
void select_column(int col, int value)
{
    if(col<=0 || s.maxrow[col]<=0) return;
    s.column_selected[col] = value; 
    int k;
    for(k=1;k<=s.maxrow[col];k++) if(s.number[col][k]) select_cell(k,col,value);  
    show_coordinates(1, col);

}

//--------------------------------------------------------------------//
//  select_row                                                        //
//--------------------------------------------------------------------//
void select_row(int row, int value)
{
    if(row<0) return;
    int k;
    s.row_selected[row] = value;
    for(k=1; k<=s.maxcol; k++) if(s.number[k][row]) select_cell(row,k,value);  
    show_coordinates(row, 1);
}


//--------------------------------------------------------------------//
//  select_cell                                                       //
//--------------------------------------------------------------------//
void select_cell(int row, int col, int value)
{
    if(!between(row,0,s.rows) || !between(col,0,s.cols)) return;
    if(value==1)
        XbaeMatrixSelectCell(s.spreadsheet_widget, row, col);
    else         
        XbaeMatrixDeselectCell(s.spreadsheet_widget, row, col);
    s.selected[col][row] = value;

}


//--------------------------------------------------------------------//
//  toggle_cell                                                       //
//--------------------------------------------------------------------//
void toggle_cell(int row, int col)
{
printf("toggle\n");
    static int orow=-1, ocol=-1;
    if(selecting==0) deselectallcells(0,0,0);
    if (row!=orow || col!=ocol)
    {    if(s.selected[col][row]==True)
         {   XbaeMatrixDeselectCell(s.spreadsheet_widget, row, col);
             s.selected[col][row] = False;
         }else
         {   XbaeMatrixSelectCell(s.spreadsheet_widget, row, col);
             s.selected[col][row] = True;
         }
    }
    orow=row;
    ocol=col;
}

//--------------------------------------------------------------------//
//  deselectallcells                                                  //
//--------------------------------------------------------------------//
void deselectallcells(char **arg, int noofargs, int m)
{ 
    arg=arg; noofargs=noofargs; m=m;
    int i,j;
    for(j=0;j<s.cols;j++)    
    {    s.column_selected[j] = 0;
         for(i=0;i<s.rows;i++)
         {   if(s.selected[j][i]==True)
             {   XbaeMatrixDeselectCell(s.spreadsheet_widget, i, j);
                 s.selected[j][i] = False;
             }
         }
    }
}

//--------------------------------------------------------------------//
//  save                                                              //
//  save all or some of the data                                      //
//--------------------------------------------------------------------//
void save(char **arg, int noofargs)
{ 
  int width[s.cols];      // width of each field
  int gotstring[s.cols];  // flag if string in a col
  Dialog dialog;

  int endcol=max(0, s.maxcol), endrow=0, startcol=1, startrow=1;
  FILE *fp;
  char tempstring[1024];
  char tempstring2[1024];
  int j,k,row,col,position,length;

  for(k=0;k<s.cols;k++) endrow = max(endrow, s.maxrow[k]); 

  if(noofargs>=2) 
  {
        strcpy(filename, arg[1]);
  }else
  {    
        strcpy(filename, g.filename);
        dialog.needrepeat = 0;  
        strcpy(dialog.title,"Save data");

        strcpy(dialog.radio[0][0],"Header");    
        strcpy(dialog.radio[0][1],"Complete");   //// both R and # headers
        strcpy(dialog.radio[0][2],"Hash only");  //// # headers
        strcpy(dialog.radio[0][3],"R only");     //// R headers
        strcpy(dialog.radio[0][4],"None");       //// no headers 
        dialog.radiono[0]=5;
        dialog.radioset[0] = format; 

        strcpy(dialog.boxes[0]," ");
        strcpy(dialog.boxes[1],"Filename");
        strcpy(dialog.boxes[2],"Starting column");
        strcpy(dialog.boxes[3],"Ending column");
        strcpy(dialog.boxes[4],"Starting row");
        strcpy(dialog.boxes[5],"Ending row");

        dialog.boxtype[0]=LABEL;
        dialog.boxtype[1]=STRING;
           strcpy(dialog.answer[1],basefilename(filename));
        dialog.boxtype[2]=INTCLICKBOX; 
           sprintf(dialog.answer[2],"%d",startcol);
           dialog.boxmin[2]=1; 
           dialog.boxmax[2]=s.cols;
        dialog.boxtype[3]=INTCLICKBOX; 
           sprintf(dialog.answer[3],"%d",endcol);
           dialog.boxmin[3]=1; 
           dialog.boxmax[3]=s.cols;
        dialog.boxtype[4]=INTCLICKBOX; 
           sprintf(dialog.answer[4],"%d",startrow);
           dialog.boxmin[4]=1; 
           dialog.boxmax[4]=s.rows;
        dialog.boxtype[5]=INTCLICKBOX; 
           sprintf(dialog.answer[5],"%d",endrow);
           dialog.boxmin[5]=1; 
           dialog.boxmax[5]=s.rows;
        for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
        for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
        for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
        dialog.noofradios=1;
        dialog.noofboxes=6;
        dialog.helptopic=0;
        dialog.keep=0;
        dialog.needrepeat=0;
        dialog.want_changecicb = 0;
        dialog.f1 = null;
        dialog.transient = 1;
        strcpy(dialog.path, ".");

        ///////---///////
        dialogbox(dialog);
        ///////---///////
 
        if(g.getout){ g.getout=0; return; }
        format  = dialog.radioset[0]; 
        strcpy(filename,dialog.answer[1]);
        startcol = atoi(dialog.answer[2]);
        endcol   = atoi(dialog.answer[3]);
        startrow = atoi(dialog.answer[4]);
        endrow   = atoi(dialog.answer[5]);
  }


  if(cant_overwrite_file(filename)) return;
  if((fp=fopen(filename,"wt")) == NULL)
      error_message(filename, errno);
  else
  {   
      //// s.text is a char[col][row][10]
      //// don't save a header unless user changed a title or
      //// file originally contained one
      printf("Header format %d\n",format);
      if(format != 4) s.header_exists =1;
      if(!s.header_exists) printf("No header\n");
      if(format != 4 && s.header_exists) 
      {
         //// Save col labels in quotes, header=1 or 3

         if((format==1 || format==3))
         {   for(col=startcol; col<=endcol; col++)
                if(strlen(s.title[col])) fprintf(fp,"%c%s%c ",0x22, s.title[col], 0x22);
             fprintf(fp,"\n");
         }

         //// Save row labels in quotes, header=1 only 
         
         if((format==1) && s.titlerows)
         {   for(row=startrow; row<=endrow; row++)
                if(strlen(s.rowlabel[row])) fprintf(fp,"%c%s%c ",0x22, s.rowlabel[row], 0x22);
             fprintf(fp,"\n");
         }

         ////  save titles = #title, header=1 or 2

         if((format==1) || (format==2))
         {   for(col=startcol; col<=endcol; col++)
                fprintf(fp,"# %s\n", s.title[col]);
         }

         ////  save commands = ##command
         if(format==1 || format==2)
         {
             position = 0;
             while(parselines(g.commands, tempstring, position))
             {
                 fprintf(fp,"##%s\n", tempstring);
             }
             for(j=1; j<s.maxcol; j++)
             for(k=1; k<s.maxrow[j]; k++)
                 if(s.form[j][k][0])
                     fprintf(fp,"#= %d %d %s\n", k,j,s.form[j][k]);
         }
      }
      printf("Saving rows %d to %d, column %d to %d\n",startrow,endrow,startcol,endcol);

      //// save data
      strcpy(tempstring, "     ");
    
      for(col=startcol; col<=endcol; col++)
      {   width[col] = 1;
          gotstring[col] = 0;
          for(row=startrow; row<=endrow; row++)
          {
              sprintf(tempstring, "%-13.13g",s.x[col][row]); // %-13s to left-justify doesn't work
              remove_trailing_space(tempstring);
              strcat(tempstring, " ");
              if(s.number[col][row])
                 width[col] = max(width[col], (int)strlen(tempstring));
              if(s.string[col][row] || s.hex[col][row])
              {  gotstring[col]=1;
                 width[col] = max(width[col], (int)strlen(s.text[col][row]));
              }
          }
      }
      for(col=startcol; col<=endcol; col++) 
          if(gotstring[col]) width[col]+=3; else width[col]++;
      for(row=startrow; row<=endrow; row++)
      {   for(col=startcol; col<=endcol; col++)
          {    
               tempstring[0] = 0;
               tempstring2[0] = 0;
               if(s.string[col][row] || s.hex[col][row])
               {
                    sprintf(tempstring, "\"%s\" ",s.text[col][row]);  //// a string in quotes
               }else if ((int)strlen(s.text[col][row])==0)
               { 
                   sprintf(tempstring," - ");          //// skip, missing data
               }else if(!strcmp(s.text[col][row],"-")) //// skip a number if user deleted it
               {
                   sprintf(tempstring," - ");          //// skip a number if already was deleted
               }else                                   //// print the number
               {
                   sprintf(tempstring, "%-13.13g ",s.x[col][row]); // %-13s to left-justify doesn't work
               }
               remove_trailing_space(tempstring);
               strcat(tempstring, " ");
               fprintf(fp, tempstring);

               //// pad with spaces
               length = strlen(tempstring);    
               for(k=0; k<width[col]; k++) tempstring2[k] = ' ';
               tempstring2[max(0, width[col] - length)]=0;
               fprintf(fp, tempstring2);
          }
          fprintf(fp, "\n");
      }
  }
  fclose(fp);
  strcpy(g.filename, filename);
  g.touched=0;
}


//--------------------------------------------------------------------//
//  export_data                                                       //
//  save one column of data                                           //
//--------------------------------------------------------------------//
void export_data(char **arg, int noofargs)
{ 
  const int WIDTH=18;  // width of each field
  Dialog dialog;
  FILE *fp;
  char tempstring[1024];
  int j,k,row,col=s.current_col;

  if(noofargs>=2) 
        strcpy(filename, arg[1]);
  else
  {     dialog.needrepeat = 0;  
        strcpy(dialog.title,"Export data");

        strcpy(dialog.radio[0][0],"Format"); 
        strcpy(dialog.radio[0][1],"Text column");
        strcpy(dialog.radio[0][2],"Text row");
        strcpy(dialog.radio[0][3],"R format row");             
        dialog.radiono[0]=4;
        dialog.radioset[0] = format + 1; 

        strcpy(dialog.boxes[0]," ");
        strcpy(dialog.boxes[1],"Filename");
        strcpy(dialog.boxes[2],"Column");

        dialog.boxtype[0]=LABEL;
        dialog.boxtype[1]=STRING;
           strcpy(dialog.answer[1],basefilename(filename));
        dialog.boxtype[2]=INTCLICKBOX; 
           sprintf(dialog.answer[2],"%d",col);
           dialog.boxmin[2]=1; 
           dialog.boxmax[2]=s.cols;
        for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
        for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
        for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
        dialog.noofradios=1;
        dialog.noofboxes=3;
        dialog.helptopic=0;
        dialog.keep=0;
        dialog.needrepeat=0;
        dialog.want_changecicb = 0;
        dialog.f1 = null;
        dialog.transient = 1;
        strcpy(dialog.path, ".");

        ///////---///////
        dialogbox(dialog);
        ///////---///////
 
        if(g.getout){ g.getout=0; return; }
        format  = dialog.radioset[0] -1; 
        strcpy(filename,dialog.answer[1]);
        col = atoi(dialog.answer[2]);
  }

  if(cant_overwrite_file(filename)) return;
  if((fp=fopen(filename,"wt")) == NULL)
      error_message(filename, errno);
  else
  {   
      ////  s.text is a char[col][row][10]
      fix_up_spreadsheet();

      //// save data
      for(k=0; k<WIDTH; k++) tempstring[k] = ' ';

      if(format==2) fprintf(fp, "("); 
      for(row=1; row<=s.maxrow[col]; row++)
      {      
               if(s.string[col][row] && format==1)
                   strcpy(tempstring,"0 ");
               else
                   sprintf(tempstring, " %g",s.x[col][row]); // %-13s to left-justify doesn't work
               fprintf(fp, tempstring);
               switch(format)
               { case 0: fprintf(fp, "\n"); break;
                 case 1: fprintf(fp, " "); break;
                 case 2: if(row<s.maxrow[col]) fprintf(fp, ", "); break;
          }
      }
      if(format==2) fprintf(fp, ")\n"); 
  }
  fclose(fp);
  g.touched=0;
}


//--------------------------------------------------------------------//
// parselines                                                         //
// Input: a multiline string possibly containing \n's.                //
// Starts at position 'position' in string, returns next part of the  //
// string up to the next \n. Increments position (passed by reference)//
// Returns 1 if not at the end of the string.                         //
// 'position' must be initialized to 0 before calling.                //
//--------------------------------------------------------------------//
int parselines(char *string, char *s, int &position)
{   
   char *start;
   char *end;
   if(position >= (int)strlen(string)) return 0;
   start = string+position;
   if((end = strchr(start, '\n')) == NULL)
   {   strcpy(s, start);
       position = strlen(string);
   }else
   {   strncpy(s, start, end-start); 
       s[end-start] = 0;      // strncpy doesn't add terminating 0
       position += end-start+1;
   }
   return 1;
}


//--------------------------------------------------------------------//
// fix_up_spreadsheet                                                 //
//--------------------------------------------------------------------//
void fix_up_spreadsheet(void)
{
  int col, row, highestrow;
  //// Fill in any non-rectangular areas with 0s
  highestrow = 0;
  for(col=0; col<s.maxcol; col++) 
      highestrow = max(highestrow, s.maxrow[col]);
  for(col=0; col<s.maxcol; col++) 
      s.maxrow[col] = highestrow;
  for(col=0; col<s.maxcol; col++) 
  for(row=0; row<s.maxrow[col]; row++) 
      if(strlen(s.text[col][row])==0)
         set_cell_value(row, col, 0.0);
  XtVaSetValues(s.spreadsheet_widget, XmNcolumnLabels, s.title, NULL);
}


//--------------------------------------------------------------------//
//  pipe                                                              //
//  print all the data to a named pipe                                //
//--------------------------------------------------------------------//
void pipe(void)
{ 
  int highest=0,i,j;
  FILE *fp;
  int status=0;
  mode_t fifo_mode = S_IRUSR | S_IWUSR;

  for(j=0;j<s.maxcol;j++)
        highest = max(highest, s.maxrow[j]);

  if(access("fifi",F_OK))   // If fifo does not exist, create it
  {   if((status = mkfifo("fifi", fifo_mode)==-1) && errno !=EEXIST)
           message((char*)"Could not create fifi!");
  }
  if(!status)                 // status==0 = Ok
  {    
       if((fp = fopen("fifi","wt"))==NULL)
              message((char*)"Cannot open fifi!");  
       else
       {    for(i=0;i<highest;i++)
            {    for(j=0;j<s.maxcol;j++)
                      fprintf(fp,"%g ",s.x[j][i]);
                 fprintf(fp,"\n");
            }
       }
       fclose(fp);
  }
}


//--------------------------------------------------------------------//
//  reload                                                            //
//  re-read last loaded file                                          //
//--------------------------------------------------------------------//
void reload(char **arg, int noofargs)
{ 
   noofargs = 2;
   strcpy(arg[1], g.filename);
   s.current_row = 1;
   s.current_col = 1;
   read(arg, noofargs);
}

//--------------------------------------------------------------------//
//  read                                                              //
//  read all the data, overwriting whatever is present                //
//--------------------------------------------------------------------//
void read(char **arg, int noofargs)
{ 
  Dialog dialog;
  static int skiplines=0;
  static char delimiter = ' ';
  static int delimiter_choice = 1;
  static int read_what = 1;      //1=entire file 2=one column
  static int read_header = 0;    //0=ignore header
  static int hit = 0;
  static char stripstring[1024] = "$,";
  static int getcol   = 1;  // what column to start reading from
  int startrow = s.current_row;  // destination row
  int startcol = s.current_col;  // destination col
  int j,k;
  char **textrow;
  char tempstring[1024];
  char extension[FILENAMELENGTH]="X";
  char *ptr;

  textrow = new char *[s.cols];
  for(k=0;k<s.cols;k++) textrow[k] = new char[CELLBYTES+1];
  if(noofargs>=2) 
        strcpy(filename, arg[1]);
  else
  {     dialog.needrepeat = 0;  

        strcpy(dialog.title,"Read data");

        strcpy(dialog.radio[0][0],"Delimiter"); 
        strcpy(dialog.radio[0][1],"Autodetect");
        strcpy(dialog.radio[0][2],"Spaces");
        strcpy(dialog.radio[0][3],"Commas");
        dialog.radiono[0]=4;
        dialog.radioset[0] = delimiter_choice; 

        strcpy(dialog.radio[1][0],"Read what"); 
        strcpy(dialog.radio[1][1],"Entire file");
        strcpy(dialog.radio[1][2],"Single column");
        dialog.radioset[1] = read_what; 
        dialog.radiono[1]=3;

        strcpy(dialog.boxes[0],"File parameters");
        strcpy(dialog.boxes[1],"Filename              ");
        strcpy(dialog.boxes[2],"Where to put it"); 
        strcpy(dialog.boxes[3],"Column");
        strcpy(dialog.boxes[4],"Row");
        strcpy(dialog.boxes[5],"What to read"); 
        strcpy(dialog.boxes[6],"Starting Column ");
        strcpy(dialog.boxes[7],"Characters to strip");
        strcpy(dialog.boxes[8],"Lines to skip");

        dialog.boxtype[0]=LABEL;
        dialog.boxtype[1]=FILENAME;
           strcpy(dialog.answer[1],basefilename(filename));

        dialog.boxtype[2]=LABEL;
        
        dialog.boxtype[3]=INTCLICKBOX; 
           sprintf(dialog.answer[3],"%d",startcol);
           dialog.boxmin[3]=1; 
           dialog.boxmax[3]=s.cols;
        dialog.boxtype[4]=INTCLICKBOX; 
           sprintf(dialog.answer[4],"%d",startrow);
           dialog.boxmin[4]=1; 
           dialog.boxmax[4]=s.rows;

        dialog.boxtype[5]=LABEL;

        dialog.boxtype[6]=INTCLICKBOX; 
           sprintf(dialog.answer[6],"%d",getcol);
           dialog.boxmin[6]=1; 
           dialog.boxmax[6]=50;
        dialog.boxtype[7]=STRING;
           strcpy(dialog.answer[7], stripstring);
        dialog.boxtype[8]=INTCLICKBOX; 
           sprintf(dialog.answer[8],"%d",skiplines);
           dialog.boxmin[8]=0; 
           dialog.boxmax[8]=s.rows-1;

//// dont do this, it detects header automatically
//      dialog.boxtype[9]=TOGGLE;
//         dialog.boxset[9] = read_header;


        for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
        for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
        for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
        dialog.noofradios=2;
        dialog.noofboxes=9;
        dialog.helptopic=0;
        dialog.keep=0;
        
        dialog.needrepeat=0;
        dialog.want_changecicb = 0;
        dialog.f1 = null;
        dialog.transient = 1;
        strcpy(dialog.path, ".");

        ///////---///////
        dialogbox(dialog);
        ///////---///////
 
        if(g.getout)return;

        delimiter_choice = dialog.radioset[0];
        read_what = dialog.radioset[1];

        strcpy(filename,dialog.answer[1]);       
        startcol = atoi(dialog.answer[3]);
        startrow = atoi(dialog.answer[4]);
        getcol = atoi(dialog.answer[6]);
        strcpy(stripstring, dialog.answer[7]);
        skiplines = atoi(dialog.answer[8]);
//      read_header = dialog.boxset[9];

  }

  if(!hit)
  {   hit=1;
      strcpy(g.filename, filename);
  }

  ////  Extract extension from filename if format is identified by extension

  switch(delimiter_choice)
  {
      case 1: delimiter = delimiter_guess(filename); break;
      case 2: delimiter = ' '; break;
      case 3: delimiter = ','; break;
  }
  strcpy(tempstring, filename);
  ptr = strrchr(tempstring,'.');
  if(ptr!=NULL) strcpy(extension,ptr+1);    
  extension[63] = 0;
  strlwr(extension);


  if(!strcmp(extension,"gcd"))
      readgcdfile(filename, textrow, startrow, startcol);
  else if(!strcmp(extension,"dif"))
      readdiffile(filename, textrow);
  else
      readplainfile(filename, textrow, startrow, startcol, delimiter, stripstring, read_what, read_header, getcol-1);

  for(k=0;k<s.cols;k++) delete[] textrow[k];
  delete[] textrow;

}


//--------------------------------------------------------------------//
//  erase  - erase all cells                                          //
//--------------------------------------------------------------------//
void erase(void)
{
   int row,col, status=NO;
   status = message((char*)"Delete all data from spreadsheet?", YESNOQUESTION);
   if(status != YES) return;
   for(col=1;col<s.cols;col++) 
   {    s.column_selected[col]=0;
        for(row=0;row<s.rows;row++)
        {    s.text[col][row][0] = 0;
             s.form[col][row][0] = 0;
             s.x[col][row] = 0.0;              
             s.number[col][row] = 0;              
        }
   }
   refreshspreadsheet();
   for(col=0;col<s.cols;col++) s.maxrow[col]=0;
   s.maxcol = 0;
   return;
}



//--------------------------------------------------------------------//
//  set_cell_value                                                    //
//  sets a cell to be a number for display                            //
//--------------------------------------------------------------------//
void set_cell_value(int row, int col, double value)
{
   char tempstring[1024];
   if(row > s.rows)
   {
       message((char*)"Too many data points!",ERROR); 
       return; 
   }
   s.maxcol = max(s.maxcol, col);
   s.maxrow[col] = max(s.maxrow[col], row);
   s.x[col][row] = value;
   if(g.number_type==HEX)
   {
       s.number[col][row] = 0;
       s.hex[col][row] = 1;
   }else
   {
       s.number[col][row] = 1;
       s.hex[col][row] = 0;
   }
   s.string[col][row] = 0;

   if((s.text[col][row][0]== '-') && (strlen((char*)s.text[col][row])==1)) //skip blank cell
   {  
      s.text[col][row][0]= 0;    //skip
   }else if(s.text[col][row])
   {   
      format_number(value, s.text[col][row]);
   }
   strcpy(s.disp[col][row], s.text[col][row]);
   strcpy(tempstring, s.text[col][row]);
   XbaeMatrixSetCell(s.spreadsheet_widget, row, col, tempstring);
   XbaeMatrixRefreshCell(s.spreadsheet_widget, row, col);
}

//--------------------------------------------------------------------//
//  set_cell_text                                                     //
//--------------------------------------------------------------------//
void set_cell_text(int row, int col, char *string)
{
   if(row >= s.rows){ message((char*)"Too many data points!",ERROR); return; }
   s.maxcol = max(s.maxcol, col);
   s.maxrow[col] = max(s.maxrow[col], row);
   s.x[col][row] = 0.0;
   strcpy(s.text[col][row], string);
   strcpy(s.disp[col][row], string);
   XbaeMatrixSetCell(s.spreadsheet_widget, row, col, string);
   XbaeMatrixRefreshCell(s.spreadsheet_widget, row, col);  
}



//--------------------------------------------------------------------//
//  set_cell                                                          //
//--------------------------------------------------------------------//
void set_cell(int row, int col, char *text)
{
   int k;
   if(strlen(text))
   {    s.x[col][row] = eval(text, col, row);
        strcpy(s.text[col][row], text);
        s.number[col][row]=1;
        s.string[col][row]=0;
        g.getout = 0;
        XbaeMatrixSetCell(s.spreadsheet_widget,row,col,text);
   }else
   { 
        s.x[col][row] = 0.0;
        s.text[col][row][0]=0;
        s.number[col][row]=0;
        s.string[col][row]=0;
        XbaeMatrixSetCell(s.spreadsheet_widget,row,col,(char*)"");
   }
   s.maxrow[col]=0;
   for(k=0;k<s.rows;k++) if(strlen(s.text[col][k])) s.maxrow[col]=k;
   XbaeMatrixRefreshCell(s.spreadsheet_widget,row,col);
   XbaeMatrixRefresh(s.spreadsheet_widget);  
   g.touched=1;
   return;
}



//--------------------------------------------------------------------//
//  insert_cell                                                       //
//--------------------------------------------------------------------//
void insert_cell(int row, int col)
{
   int j=col, k;
   for(k=s.maxrow[j]+1; k>row; k--)
   {     s.x[j][k] = s.x[j][k-1];
         s.number[j][k] = s.number[j][k-1];
         s.string[j][k] = s.string[j][k-1];
         s.hex[j][k] = s.hex[j][k-1];
         strcpy(s.text[j][k], s.text[j][k-1]);
         strcpy(s.disp[j][k], s.disp[j][k-1]);
         strcpy(s.form[j][k], s.form[j][k-1]);
         XbaeMatrixSetCell(s.spreadsheet_widget,k,j,s.disp[j][k]);
   }
   refreshspreadsheet();       
   set_cell(row,col,(char*)"");
   return;
}


//--------------------------------------------------------------------//
//  insert_row                                                        //
//--------------------------------------------------------------------//
void insert_row(int row, int col)
{
   int j, k;
   for(j=1; j<=s.maxcol; j++)
   {  for(k=s.maxrow[j]+1; k>row; k--)
      {     s.x[j][k] = s.x[j][k-1];
            s.number[j][k] = s.number[j][k-1];
            s.string[j][k] = s.string[j][k-1];
            s.hex[j][k] = s.hex[j][k-1];
            strcpy(s.text[j][k], s.text[j][k-1]);
            strcpy(s.disp[j][k], s.disp[j][k-1]);
            strcpy(s.form[j][k], s.form[j][k-1]);
            XbaeMatrixSetCell(s.spreadsheet_widget,k,j,s.disp[j][k]);
      }
   }
   for(j=1; j<=s.maxcol; j++)
   {        s.x[j][row] = 0;
            strcpy(s.text[j][row], "xxx");
            strcpy(s.disp[j][row], "xxx");
            XbaeMatrixSetCell(s.spreadsheet_widget,row,j,s.disp[j][row]);
            s.maxrow[j]++;  
   }
   refreshspreadsheet();       
   return;
}


//--------------------------------------------------------------------//
//  delete_selected_cells                                             //
//  If cells are selected, deletes all selected cells. Otherwise,     //
//  deletes the current cell. Not efficient for multiple cells.       //
//--------------------------------------------------------------------//
void delete_selected_cells(char **arg, int noofargs, int n)
{
   arg=arg; noofargs=noofargs; n=n;
   SelectCountStruct sel;
   sel.start = NULL;
   sel.end = NULL;
   int i,j,k;
   int selected_total = count_selected_data(&sel, 0);
   if(selected_total==0)
   {  if(!s.number[s.current_col][s.current_row])  return;
      else s.selected[s.current_col][s.current_row] = 1;
   }
   for(j=1; j<s.cols; j++)
   {
        for(i=min(s.maxrow[j], s.rows); i>0; i--)
        {  
             if(s.selected[j][i]) 
             {   
                 for(k=i; k<s.maxrow[j]; k++)
                 {           
                     s.x[j][k] = s.x[j][k+1];
                     strcpy(s.text[j][k], s.text[j][k+1]);
                     strcpy(s.form[j][k], s.form[j][k+1]);
                     strcpy(s.disp[j][k], s.disp[j][k+1]);
                     XbaeMatrixSetCell(s.spreadsheet_widget,k,j,s.disp[j][k]);
                 }
                 s.x[j][s.maxrow[j]] = 0.0;
                 XbaeMatrixSetCell(s.spreadsheet_widget, s.maxrow[j], j, (char*)"");
                 s.number[j][s.maxrow[j]] = 0;
                 s.selected[j][s.maxrow[j]] = 0;
                 s.maxrow[j]--;
                 select_cell(i, j, False);
             }
        }
        select_column(j, 0);
   }     
   select_cell(1, 1, 0);
   refreshspreadsheet();       
   return;
}

//--------------------------------------------------------------------//
//  erase_cell                                                        //
//--------------------------------------------------------------------//
void erase_cell(int row, int col)
{
   for(int k=row;k<s.maxrow[col];k++)
        set_cell(k,col,s.text[col][k+1]);
   set_cell(s.maxrow[col],col,(char*)"");
   select_cell(row, col, 0);
   g.touched=1;
}


//--------------------------------------------------------------------//
//  edit_cell                                                         //
//--------------------------------------------------------------------//
void edit_cell(int row, int col)
{
    XbaeMatrixEditCell(s.spreadsheet_widget, min(s.rows, max(0, row)),
                                             min(s.cols, max(0, col)));
}                                             

//--------------------------------------------------------------------//
//  move_cursor                                                       //
//--------------------------------------------------------------------//
void move_cursor(int direction)
{
printf("hi %d\n",direction);
// it does not seem to ever get here
   switch(direction)
   {   case UP:    edit_cell(s.current_row-1, s.current_col); break;
       case DOWN:  edit_cell(s.current_row+1, s.current_col); break;
       case LEFT:  edit_cell(s.current_row,   s.current_col-1); break;
       case RIGHT: edit_cell(s.current_row,   s.current_col+1); break;
   }
   return;
}


//--------------------------------------------------------------------//
//  count_one_dataset                                                 //
//--------------------------------------------------------------------//
int count_one_dataset(int set[], int n[])
{
  int i,j;
  int total_cols=0;         // Total no. of columns 

  ////  Count selected cells
  ////  Only numbers can be selected so don't need to check if a number

  for(j=1;j<s.cols;j++) if(column_has_selected_numbers(j)) set[total_cols++] = j;  
  if(total_cols != 1){ message((char*)"Must select only 1 column", ERROR); return ERROR; }
  n[0] = 0; 
  for(i=0;i<s.rows;i++) if(s.selected[set[0]][i]) n[0]++;
  return OK;
}



//--------------------------------------------------------------------//
//  count_two_datasets                                                //
//--------------------------------------------------------------------//
int count_two_datasets(int set[], int n[])
{
  int i,j;
  int total_cols=0;         // Total no. of columns 

  ////  Count selected cells
  ////  Only numbers can be selected so don't need to check if a number
  ////  Only count if user hasn't supplied them


  if(set[0]==0 && set[1]==0)
  {
      for(j=1;j<s.cols;j++) if(column_has_selected_numbers(j)) set[total_cols++] = j;  
      if(total_cols != 2){ message((char*)"Must select only 2 columns", ERROR); return ERROR; }
  }
  n[0] = 0;  n[1] = 0;  
  for(i=0;i<s.rows;i++) if(s.selected[set[0]][i]) n[0]++;
  for(i=0;i<s.rows;i++) if(s.selected[set[1]][i]) n[1]++;
  return OK;
}


//--------------------------------------------------------------------//
//  get_one_dataset                                                   //
//--------------------------------------------------------------------//
int get_one_dataset(double *data[1], int col[1])
{
  int i,n;
  ////  Only numbers can be selected so don't need to check if a number
  n = 0;                    // Total no. of data points in set 1
  for(i=1;i<s.rows;i++) if(s.selected[col[0]][i]) data[0][n++] = s.x[col[0]][i];
  return OK;
}


//--------------------------------------------------------------------//
//  get_two_datasets                                                  //
//--------------------------------------------------------------------//
int get_two_datasets(double *data[2], int col[2])
{
  int i,n;
  ////  Only numbers can be selected so don't need to check if a number
  n = 0;                    // Total no. of data points in set 1
  for(i=1; i<s.rows; i++) if(s.selected[col[0]][i]) data[0][n++] = s.x[col[0]][i];
  n = 0;                    // Total no. of data points in set 2
  for(i=1; i<s.rows; i++) if(s.selected[col[1]][i]) data[1][n++] = s.x[col[1]][i];
  return OK;
}


//--------------------------------------------------------------------//
//  column_has_selected_numbers                                       //
//--------------------------------------------------------------------//
int column_has_selected_numbers(int col)
{
    int row;
    for(row=1; row<s.rows; row++)
       if(s.selected[col][row]) return 1;
    return 0;
}

//--------------------------------------------------------------------//
//  swap_columns                                                      //
//--------------------------------------------------------------------//
void swap_columns(char **arg, int noofargs)
{
  static int col1=1, col2=2;
  int k,n;
  char tempstring[128];

  itoa(s.current_col, tempstring, 10);
  if(noofargs<2) 
  {     message((char*)"1st column no.", tempstring, PROMPT, 128);
        col1 = atoi(tempstring);
  }else col1 = atoi(arg[1]);
  if(g.getout){ g.getout=0; return; }

  itoa(col2, tempstring, 10);
  if(noofargs<2) 
  {     message((char*)"2nd column no.", tempstring, PROMPT, 128);
        col2 = atoi(tempstring);
  }else col2 = atoi(arg[2]);
  if(g.getout){ g.getout=0; return; }

  n = max(s.maxrow[col1],s.maxrow[col2]);
  for(k=1;k<=n;k++) dswap(s.x[col1][k], s.x[col2][k]);

  //// swap titles
  strcpy(tempstring, s.title[col1]);
  setcoltitle(col1, s.title[col2]);
  setcoltitle(col2, tempstring);

  refreshspreadsheet();
}


//--------------------------------------------------------------------//
//  delete_column                                                     //
//--------------------------------------------------------------------//
void delete_column(char **arg, int noofargs)
{
  static int col;
  int k,n;
  char tempstring[128];
  itoa(s.current_col, tempstring, 10);
  if(noofargs<1) 
  {     message((char*)"Column to delete", tempstring, PROMPT, 128);
        col = atoi(tempstring);
  }else col = atoi(arg[1]);
  if(!between(col,1,s.cols)) { message((char*)"Nonexistent column"); return; }

  if(g.getout){ g.getout=0; return; }

  n = s.maxrow[col];
  for(k=1; k<=n; k++) 
  {   s.x[col][k] = 0.0;
      s.text[col][k][0] = 0;
      s.disp[col][k][0] = 0;
      s.form[col][k][0] = 0;
      s.number[col][k] = 0;
  }
  refreshspreadsheet();
  if(col == s.maxcol) s.maxcol--;
  s.maxrow[col] = 0;
}

//--------------------------------------------------------------------//
//  delete_row                                                        //
//--------------------------------------------------------------------//
void delete_row(char **arg, int noofargs)
{
  static int row;
  int col, k, n;
  char tempstring[128];
  itoa(s.current_row, tempstring, 10);
  if(noofargs<1) 
  {     message((char*)"Row to delete", tempstring, PROMPT, 128);
        row = atoi(tempstring);
  }else row = atoi(arg[1]);
  if(!between(row, 1, s.rows)) { message((char*)"Nonexistent row"); return; }

  if(g.getout){ g.getout=0; return; }

  for(col=1; col<=s.maxcol; col++) 
  {   
      n = s.maxrow[col];  
      if(n<1) break;
      for(k=max(1, row); k<=n; k++) 
      {  
          s.x[col][k] = s.x[col][k+1];
          strcpy(s.text[col][k], s.text[col][k+1]);
          strcpy(s.disp[col][k], s.disp[col][k+1]);
          strcpy(s.form[col][k], s.form[col][k+1]);
          s.number[col][k] = s.number[col][k+1];
      }
      s.x[col][n] = 0;
      s.text[col][n][0] = 0;
      s.disp[col][n][0] = 0;
      s.form[col][n][0] = 0;
      s.number[col][n] = 0;
      XbaeMatrixSetCell(s.spreadsheet_widget,n,col,(char*)s.disp[col][n]);
      s.maxrow[col] = max(s.maxrow[col]-1, 0);
      n = max(n-1, 0);
      if(n<1) break;
  }
  refreshspreadsheet();
}



//--------------------------------------------------------------------//
//  maximum                                                           //
//--------------------------------------------------------------------//
void maximum(char **arg, int noofargs, int mode)
{
  static int col;
  static int startrow=1;
  static double findval=0.0;
  int k,n,row=-1;
  double testval=0.0, val;
  char tempstring[128];
  itoa(s.current_col, tempstring, 10);

  if(noofargs<1) 
  {     message((char*)"Column", tempstring, PROMPT, 128);
        col = atoi(tempstring);
  }else col = atoi(arg[1]);
  if(g.getout){ g.getout=0; return; }

  if(mode==EQU)
  {  sprintf(tempstring, "%g", findval);
     if(noofargs<2) 
     {     message((char*)"Value to find", tempstring, PROMPT, 128);
           findval = atof(tempstring);
     }
     if(g.getout){ g.getout=0; return; }
  }else startrow = 1;

  n = s.maxrow[col];
  for(k=1; k<=n; k++) if(s.number[col][k]){ testval = s.x[col][k]; row=k; break; }
  if(row==-1){ message((char*)"No numbers in column"); return; }

  for(k=startrow; k<=n; k++) 
  {    if(!s.number[col][k]) continue;
       val = s.x[col][k];
       if(mode==EQU) if(val == findval){ row = k; startrow = row+1; break; }
       if(mode==MAX) if(val > testval){ testval = val; row = k; }
       if(mode==MIN) if(val < testval){ testval = val; row = k; }
  }
  edit_cell(row, col); 
}


//--------------------------------------------------------------------//
//  mousecb                                                           //
//  note: buttonrelease can occur before buttonpress. Hence the hack  //
//  using inhibit_select.                                             //
//--------------------------------------------------------------------//
void mousecb(Widget w, XtP client_data, XEvent *event)
{
   static int selecting = 0;
   static int row1, col1, row2, col2;
   static int or1=0,oc1=0,or2=0,oc2=0;
   int r1,c1,r2,c2;
   w=w;client_data=client_data;
   int i,j;
   if(event->type == MotionNotify) 
   {    if(mouse_button && !inhibit_select) selecting = 1; else selecting=0;
   }
   if(event->type == ButtonPress)
   {    XbaeMatrixGetEventRowColumn(w, event, &row1, &col1);

        if(row1>0 && col1>0)
        {   deselectallcells(NULL, 0, 0);
        }
        mouse_button = (int)event->xbutton.button;
        if(mouse_button==2) 
        {    edit_cell(row1, col1);
             copy(NULL, 0, PASTE);
        }
   }
   if(event->type == ButtonRelease) 
   {    
        XbaeMatrixGetEventRowColumn(w, event, &row2, &col2);
        mouse_button = (int)event->xbutton.button;
        if(mouse_button==1 && selecting) copy(NULL, 0, COPY);
        if(row2>=0 && col2>0) selecting = 0;   
        mouse_button = 0;     
        selecting = 0;
        inhibit_select = 0;
   }
   if(selecting)
   {    XbaeMatrixGetEventRowColumn(w, event, &row2, &col2);
        if(row2>=0 && col2>=0) 
        show_coordinates(row2, col2);
        {  r1 = max(0,min(row1, row2));
           c1 = max(0,min(col1, col2));
           r2 = min(s.rows,max(row1, row2));
           c2 = min(s.rows,max(col1, col2));
           if(r1!=r2 || c1!=c2)
           {   for(i=r1; i<=r2; i++)
               for(j=c1; j<=c2; j++)
                   select_cell(i,j,1);
           }
           for(i=or1; i<=or2; i++)
           for(j=oc1; j<=oc2; j++)
           if(!between(i,r1,r2) || !between(j,c1,c2))
               select_cell(i,j,0);
           or1=r1; oc1=c1;
           or2=r2; oc2=c2;
        }
   }
}

//--------------------------------------------------------------------//
//  copy - copy to or paste cells from clipboard                      //
//  If noofargs is 0, copy selected cells to clipboard. Otherwise     //
//    copy an entire column from col1 to col2 a la npdata.            // 
//  mode can be COPY or PASTE.                                        //
//--------------------------------------------------------------------//
void copy(char **arg, int noofargs, int mode)
{
   static SelectStruct clipboard;
   static int ocount=0, hit=0;
   static SelectCountStruct sel;
   arg=arg; noofargs=noofargs; 
   int h, k, rowspercol=0, row=0, col=0, count=0, col1, col2, rr, cc;
   static int noofcols;   

   if(noofargs>=3)                    // npdata mode
   {     col1 = atoi(arg[1]);
         col2 = atoi(arg[2]);
         if(!between(col1+1, 0, s.cols) || !between(col2+1, 0, s.cols))
         {   message((char*)"Nonexistent column");
             return;
         }
         s.maxcol = max(s.maxcol, col2);
         s.maxrow[col2] = s.maxrow[col1];
         for(k=1; k<=s.maxrow[col1]; k++)
         {   s.number[col2][k] = s.number[col1][k]; 
             s.x[col2][k] = s.x[col1][k];
             strcpy(s.text[col2][k],s.text[col1][k]);
             strcpy(s.disp[col2][k],s.disp[col1][k]);
             strcpy(s.form[col2][k],s.form[col1][k]);
         }
         strcpy(s.title[col2], s.title[col1]);
         refreshspreadsheet();
      
         //// refresh titles
         XtVaSetValues(s.spreadsheet_widget, XmNcolumnLabels, s.title, NULL); 
         return;
   }
   if(mode==COPY)                    // copy to clipboard
   {     
         if(ocount)                  // must be 0 1st time through
         {     delete_clipboard(&clipboard, sel.count); 
               sel.count = 0;
               ocount = 0; 
         }
         if(!hit)
         {     sel.start = new int[s.cols];
               sel.end   = new int[s.cols];              
               for(k=1; k<s.cols; k++){ sel.start[k] = 0; sel.end[k] = 0;}
               hit = 1;
         }
         count = count_selected_data(&sel, 1);  
         if(count) 
         {     clipboard.x = new double [count];
               if(clipboard.x==NULL) return;  
               clipboard.s = new char *[count];
               if(clipboard.s==NULL){ delete_clipboard(&clipboard, count); return; }
               clipboard.isnumber = new uchar[count];
               if(clipboard.isnumber==NULL) { delete_clipboard(&clipboard, count); return; }
               get_selected_data(&clipboard, 1, count, noofcols);
               ocount = count;
         }
   }else                             // paste from clipboard
   {     cc = 0;                     // cc = column  rr = row in current column
         col = s.current_col;
         if(noofcols>0)
             rowspercol = ocount / noofcols;
         else
             rowspercol = 0;
         for(h=0,rr=0; h<ocount; h++,rr++)
         {    
              while(rr>=rowspercol){ rr=0; cc++; rowspercol=1+sel.end[cc]-sel.start[cc]; }
              row = s.current_row + rr + sel.start[cc];
              col = s.current_col + cc;
              if(row>0 && col >0) 
              {     
                    s.maxcol = max(s.maxcol, col);
                    if(clipboard.isnumber[h])
                    {
                         set_cell_value(row, col, clipboard.x[h]);
                    }else 
                    {
                         set_cell_text(row, col, clipboard.s[h]);
                    }
              }
              s.maxcol = max(s.maxcol, col);
              s.maxrow[col] = max(s.maxrow[col], row);
         }
         refreshspreadsheet();
   }
   return;
}


//--------------------------------------------------------------------//
//  move - move column                                                //
//--------------------------------------------------------------------//
void move(char **arg, int noofargs)
{
   int k, col1=0, col2=0, helptopic=0;
   if(noofargs>=1) col1 = atoi(arg[1]); 
   else col1 = clickbox((char*)"Source column", s.current_row, 0, s.rows, null, helptopic);
   if(noofargs>=2) col2 = atoi(arg[2]); 
   else col2 = clickbox((char*)"Destination column", s.current_row, 0, s.rows, null, helptopic);
   if(!between(col1,1,s.cols) || !between(col2,1,s.cols))
   {   message((char*)"Nonexistent column");
       return;
   }
   s.maxcol = max(s.maxcol, col2);
   s.maxrow[col2] = s.maxrow[col1];
   for(k=0;k<s.maxrow[col1]; k++)
   {   s.number[col2][k] = s.number[col1][k]; 
       if(s.number[col1][k]) 
            s.x[col2][k] = s.x[col1][k];
       else
            strcpy(s.text[col2][k],s.text[col1][k]);
       if(s.form[col2][k][0]) strcpy(s.form[col2][k],s.form[col1][k]);
   }
   sprintf(arg[1], "%d", col1);
   delete_column(arg, 1);
   refreshspreadsheet();
   return;
}


//--------------------------------------------------------------------//
//  count_selected_data                                               //
//  Set arrays to NULL if you don't want start & end points.          //
//  Set wantstrings to 0 if you only want to count numbers.           //
//--------------------------------------------------------------------//
int count_selected_data(SelectCountStruct *sel, int wantstrings)
{
  int i,j,istart=0,jstart=0,hit=0,oldj,ok=1;
  int total=0;              // Total no. of data points
  if(sel->start==NULL || sel->end ==NULL) ok=0;
  for(j=1; j<s.cols; j++)
  {  oldj=-1;
     if(ok){  sel->start[j]=0; sel->end[j]=0; }
     for(i=1; i<s.rows; i++)
     {   if(s.selected[j][i]) 
         {    //if(!s.number[j][i] && !wantstrings) continue;
              if(!hit){ istart=i; jstart=j; hit=1; }     // 1st hit=upper left
              total++;                                   
              if(ok)                                     // start & end are relative to u.l.
              {  if(j==oldj) sel->end[j-jstart] = i-istart; 
                 else      sel->start[j-jstart] = i-istart;
              }
              oldj = j;
         }
     }
  }
  sel->count = total;
  return total;
}


//--------------------------------------------------------------------//
//  get_selected_data                                                 //
//  Set wantstrings to 0 if you only want numbers.                    //
//--------------------------------------------------------------------//
int get_selected_data(SelectStruct *c, int wantstrings, int maxn, int &noofcols)
{
  int i,j,len,n;
  noofcols = 0;
  n=0;
  for(j=1; j<s.cols; j++)
  {  
       ////  count the number of columns
       for(i=1; i<s.rows; i++)
       {    if(s.selected[j][i]) { noofcols++; break; }
       }
       for(i=1; i<s.rows; i++)
       {    if(s.selected[j][i])
            {    c->isnumber[n] = 1-s.string[j][i];
                 if(c->isnumber[n])
                 { 
                     c->x[n] = s.x[j][i];
                 }else 
                 {   
                     if(!wantstrings) continue;
                     len = strlen(s.text[j][i]);
                     c->x[n] = 0.0;
                     c->s[n] = new char[len+1];
                     strcpy(c->s[n], s.text[j][i]);
                 }           
                 n++;
            }
            if(n>=maxn) return n-1;
       }
  }
  return n-1;
}


//--------------------------------------------------------------------//
//  get_column                                                        //
//  get all numbers in a specified column whether selected or not     //
//--------------------------------------------------------------------//
int get_column(SelectStruct *c, int column)
{
  int j,n;
  int col = column;
  n=0;
  for(j=0; j<s.maxrow[col]; j++)
  {    c->isnumber[n] = 1 - s.string[col][j];  // is it a number
       if(c->isnumber[n]) c->x[n] = s.x[col][j];
       n++;
  }
  return n-1;
}


//--------------------------------------------------------------------//
//  delete_clipboard - should really be a destructor                  //
//--------------------------------------------------------------------//
void delete_clipboard(SelectStruct *c, int n)
{
  int i;
  for(i=0; i<n; i++) if(!c->isnumber[i]) delete[] c->s[i];
  if(c->s != NULL) delete[] c->s;
  if(c->x != NULL) delete[] c->x;
  if(c->isnumber != NULL) delete[] c->isnumber;
}




//--------------------------------------------------------------------//
// select_down                                                        //
//--------------------------------------------------------------------//
void select_down(Widget w, XEvent *e, String *s, Cardinal *n)
{
   w=w; e=e; s=s; n=n;
   printf("select down\n");
}


//--------------------------------------------------------------------//
// select_up                                                          //
//--------------------------------------------------------------------//
void select_up(Widget w, XEvent *e, String *s, Cardinal *n)
{
   w=w; e=e; s=s; n=n;
   printf("select up\n");
}

//--------------------------------------------------------------------//
// page_down                                                          //
//--------------------------------------------------------------------//
void page_down()
{
   edit_cell(s.current_row+20, s.current_col);
}

//--------------------------------------------------------------------//
// page_up                                                            //
//--------------------------------------------------------------------//
void page_up()
{
   edit_cell(s.current_row-20, s.current_col);
}



//--------------------------------------------------------------------//
//  readplainfile                                                     //
//--------------------------------------------------------------------//
void readplainfile(char *filename, char **textrow, int startrow, 
int startcol, char delimiter, char *stripstring, int read_what, 
int read_header, int getcol)
{
  FILE *fp;
  static int skiplines=0;
  int rtitlecols = 0;        //// no. of R-style column titles found      
  s.titlecols = 0;           //// number of column titles found
  s.hashtitlecols = 0;       //// index for hash column titles      
  s.titlerows = 0;           //// number of column rows found
  int rtitles_processed = 0; //// no. of R-style titles handled so far
  int htitles_processed = 0; //// no. of #-style titles handled so far
  int dcol = 0;              //// destination column
  int status = 0;
  char *pos;
  int size = 200 * s.cols;
  int line=0, row=0,  rowcount=0, cols=0, rows=0, rr, cc, kdest, readcount=0;
  char buffer[2 * size];
  char tempstring[1024];
  int is_string[2 * s.cols];
  int i,j,k, hit_row_header=0, read_row_header=0, row_titles_processed=0;
  int hit_header = 0;
  for(k=0; k<s.cols; k++) is_string[k] = 0;
  for(k=0; k<s.cols; k++) strcpy(s.otitle[k], s.title[k]);
  if((fp=fopen(filename,"rt")) == NULL)
      error_message(filename, errno);
  else 
  {   row = startrow;
      do
      {   
            fgets(buffer, size, fp);
            readcount++;
            
            if(feof(fp)) break;
            remove_trailing_space(buffer);
            if(!strlen(buffer)) continue;

            //// assume header if quote on line 1
            if(buffer[0]==0x22 && !hit_header && readcount==1) read_header=1; 
            if(buffer[0]==0x22 && !hit_row_header && read_header && readcount==2) read_row_header=1; 

            //// get rid of high ascii characters
            for(i=0; i<(int)strlen(buffer)-1; i++) if((uchar)buffer[i]>127) buffer[i]=' ';

            //// If there's a header put it in column column labels
            //// column labels overwrite hash column titles to maintain
            //// consistency with R
 
            if(read_header && !hit_header && buffer[0]!='#' && readcount==1)
            {    
                 //// R-style header in quotes, must be first line
                 s.header_exists = 1; 
                 hit_header++;
                 rtitles_processed++;
                 if(getcol >= rtitles_processed) continue;

                 //// title takes precedence if header is present
                 parse_header(buffer, &s.title[startcol-1], cols, rtitlecols, FILENAMELENGTH);
                 continue;
            }else if(read_row_header && !hit_row_header)
            {
                 //// Row titles on left if present
                 s.header_exists = 1; 
                 hit_row_header++;
                 row_titles_processed++;
                 if(row_titles_processed >= s.rows) continue;
                 parse_header(buffer, &s.rowlabel[startrow-1], rows, s.titlerows, FILENAMELENGTH);
                 continue;            
            }else if((buffer[0] == '#') && (buffer[1] != '#') && (buffer[1] != '='))
            {    
                 //// 1 hash = a column title (preserved but not shown if a header is present)
                 s.header_exists = 1; 
                 hit_header++;
                 htitles_processed++;
                 if(htitles_processed >= s.cols) continue;
                 if(htitles_processed <= getcol) continue;
                 s.hashtitlecols++;
                 dcol = s.hashtitlecols+startcol-1;
                 parse_title(buffer, s.title[dcol], cols, 
                      s.hashtitlecols, FILENAMELENGTH);
                 remove_trailing_characters(s.title[dcol]);
                 continue;
            }else if(!hit_header)
            {    
                 //// no header found, so make something up. needed if reloading
                 //// set the header_exists flag to 0
                 ////sprintf(tempstring, "Col. %d",readcount);
                 ////setcoltitle(readcount, tempstring); 
            }
            //// 2 hashes = a command
            if(buffer[0] == '#' && buffer[1]=='#')
            {                    
                 //// Don't overwrite existing commands if reading a column
                 //// Read entire file = read_what is 1
                 //// Read single column = read_what is 2
       
                 if(read_what>1) continue;   
                 remove_all_cr(buffer);
                 if(!line)
                 {
                     strcpy(g.commands, buffer+2);
                 }else
                 {   g.commands[strlen(g.commands)+1] = 0;
                     g.commands[strlen(g.commands)+2] = 0;
                     strcat(g.commands, buffer+2);
                 }
                 line++;
                 hit_header++;
                 continue;
            }

            //// end of reading header and commands
 
           //// 1 hash + 1 equals = formula
            if(buffer[0] == '#' && buffer[1]=='=')
            {    ////  get row, col, and a string from #= 5 0 sum(1 1:4 1)
                 sscanf(buffer+3, "%d %d %s", &rr, &cc, tempstring);
                 pos = strchr(buffer, (int)' ');  // first space
                 pos = strchr(pos+1, (int)' ');   // 2nd space
                 pos = strchr(pos+1, (int)' ');   // 3rd space
                 strcpy(s.form[cc][rr], pos);
                 remove_trailing_characters(s.form[cc][rr]);
                 s.formula[cc][rr] = 1;
                 s.number[cc][rr] = 0;
                 s.string[cc][rr] = 0;
                 s.hex[cc][rr] = 0;
                 continue;
            }


            //// Eliminate unwanted characters like $, , etc       
            if(stripstring != NULL)
            {    for(i=0; i<(int)strlen(buffer)-1; i++)
                 {    if(strchr(stripstring, buffer[i])) 
                      {   for(j=i; j<(int)strlen(buffer)-1; j++) 
                                buffer[j] = buffer[j+1]; 
                          i--;
                      }
                 }
            }

            //// this is needed for some files, otherwise it adds an extra 
            //// column eg in chromeleon text files
            
            if(startcol+2 >= s.cols)
            {    message((char*)"Too many columns in data, edit xdata.ini to increase\nAborting read", ERROR);
                 return;
            }
            remove_all_cr(buffer);
            status = parse_string(buffer, textrow, is_string, cols, startcol, 
                 CELLBYTES, s.cols, delimiter, getcol);

            if(read_what == 2) cols=1; //// read only a single col

            if(status == -1)
            {    message((char*)"Too many columns in file\nIncrease columns in .xdata/xdata.ini\nand re-start xdata");
                 fclose(fp);
                 return;
            }
            if(!cols) continue;

            for(k=startcol+getcol; k<startcol+getcol+cols; k++) 
            {
                 kdest = k - getcol;
                 if(strlen(textrow[kdest])) strcpy(s.text[kdest][row], textrow[k-getcol]);
            }

            s.maxcol = max(s.maxcol, cols+startcol-1);
            for(k=startcol+getcol; k<startcol+getcol+cols; k++)
            {    
                 kdest = k - getcol;
                 strcpy(tempstring, s.text[kdest][row]);
                 if(!strcmp(strlwr(tempstring),"microsoft")) printf("yuck\n");
                 if(is_string[kdest-startcol]) 
                 {   
                     s.number[kdest][row] = 0;
                     s.string[kdest][row] = 1;
                     strncpy(s.disp[kdest][row], s.text[kdest][row], CELLBYTES-1);
                     XbaeMatrixSetCellBackground(s.spreadsheet_widget, row, kdest, g.scolor);
                     s.maxrow[kdest] = row;
                 }else
                 {   
                     //// if it is a skipped value, leave cell blank
                     if(!strcmp(s.text[kdest][row],"-")) //// skip
                     {   
                         s.text[kdest][row][0] = 0;
                         s.number[kdest][row] = 0;
                         s.string[kdest][row] = 0;
                         s.hex[kdest][row] = 0;
                         continue;
                     }else
                     {   s.number[kdest][row] = 1;
                         s.string[kdest][row] = 0;
                         s.hex[kdest][row] = 0;
                         g.message_count = 0;  // hack to see if eval has an error
                         s.x[kdest][row] = eval(s.text[kdest][row], kdest, row);
        
                         if(g.message_count) { fclose(fp); g.message_count=0; return; }  // don't get in endless loop of info messages
                         gcvt(s.x[kdest][row] , g.signif, s.disp[kdest][row]);
                         if(s.form[kdest][row][0])
                             XbaeMatrixSetCellBackground(s.spreadsheet_widget, row, kdest, g.ecolor);
                         s.maxrow[kdest] = row;
                     }
                 }
                 XbaeMatrixSetCell(s.spreadsheet_widget, row, kdest, s.disp[kdest][row]);
            }

            if(++rowcount > skiplines) row++;
            if(row >= s.rows){ message((char*)"Too many data points\nData truncated!",ERROR); break; }

      }while(!feof(fp));
      fclose(fp);
  }



  if(read_what == 1) //// only change filename on top bar if reading entire file
       XtVaSetValues(g.main_widget, XmNtitle, filename, NULL); 
  if(read_what == 2) ////
       rtitlecols = 0;

  //// refresh titles

  if(read_what==1)  //// read entire file
  {
      for(k=startcol; k<startcol+rtitlecols; k++)  setcoltitle(k, s.title[k]);     
  }else              //// reading one col, so put originals back
  { 
      for(k=0; k<s.cols; k++) 
      {   
           if(k!=startcol)
           {  strcpy(s.title[k], s.otitle[k]);
              setcoltitle(k, s.otitle[k]);     
           }
      }
  }
  setallcoltitles(s.title);

  if(hit_header) s.header_exists = 1;
  XtVaSetValues(s.spreadsheet_widget, XmNrowLabels, s.rowlabel, NULL); 
  g.getout = 0;

}



//--------------------------------------------------------------------//
//  readdiffile                                                       //
//--------------------------------------------------------------------//
void readdiffile(char *filename, char **textrow)
{
  enum{ NONE, TUPLE, X, Y, END };
  filename=filename; textrow=textrow;
  FILE *fp;
  static int startcol=1, startrow=0;
  int size = 200 * s.cols;
  int row=0, col=0;
  char buffer[size];
  int item_type=0;
  double value1, value2;
  int datamode = 0;

  if((fp=fopen(filename,"rt")) == NULL)
      error_message(filename, errno);
  else 
  {   row = startrow;
      do
      {     fgets(buffer, size, fp);
            if(feof(fp)) break;
            if(!strlen(buffer)) continue;
            remove_trailing_space(buffer);
            item_type = token_type(buffer, &value1, &value2);
            switch(item_type)
            {
                case TABLE:           break;
                case VECTORS:         break;
                case TUPLES:          break;
                case DATA:            break;
                case COMMENT:         break;
                case LABEL  :         break;
                case SIZE   :         break;
                case PERIODICITY:     break; // period in a time series
                case MINORSTART:      break;
                case TRUELENGTH:      break; // portion of a vector that contains significant values
                case DISPLAYUNITS:    break;
                case BOT: datamode=TUPLE;    // x values follow; start of new tuple
                   break; 
                case EOD: datamode=END;   break;
                case V:   datamode=Y;       // y values follow
                   break; 
                case NA:              break;
                case ERROR:           break;
                case TRUE_VALUE:      break;
                case FALSE_VALUE:     break;                   
                case NONE:                   // a tuple
                   if(datamode==TUPLE && value1==0) 
                   {   row++;
                       datamode=X;        
                   }
                   if(value1==-1) datamode=NONE;
                   if(datamode==X)
                   {   col = startcol;
                       add_value(value2, row, col);  
                   }
                   if(datamode==Y)
                   {   col++;
                       add_value(value2, row, col);  
                   }
                   break;
                
            }
      }while(!feof(fp));
      fclose(fp);
  }
  g.getout = 0;

}



//--------------------------------------------------------------------//
//  readgcdfile                                                       //
//--------------------------------------------------------------------//
void readgcdfile(char *filename, char **textrow, int startrow, int startcol)
{
  FILE *fp;
  uchar buffer[256];
  uchar hexbuffer[256];
  hexbuffer[16] = 0;
  buffer[8] = 0;
  double value=0;
  int hit1=0, hit2=0, hit3=0, hit4=0;
  int count = 0;
  int row, col;
  row = startrow;
  col = startcol;
  
  refreshspreadsheet();
  if((fp=fopen(filename,"rb")) == NULL)
      error_message(filename, errno);
  else 
  { 
      do
      {     
          if(!hit3)
          {
            fread(&buffer, 8,1,fp);  
            converttohex(buffer, hexbuffer);
            if(!strcmp((char*)hexbuffer,"8080808080808080")) hit1=1;
            if(hit1 && (!strcmp((char*)hexbuffer,"ffffffffffffffff"))) hit2=1;
            if(hit2 && ( strcmp((char*)hexbuffer,"ffffffffffffffff"))) hit3=1;
          }else
          {
            fread(&value, 8,1,fp);  
            if(value==0) hit4++;
            row = count + startrow;
            count++;

            s.number[col-startcol][row] = 1;
            s.string[col-startcol][row] = 0;
            s.hex[col-startcol][row] = 0;
            s.x[col-startcol][row] = value;
            sprintf(s.disp[col-startcol][row], "%g", value);

            if(hit4>50)
            {    count-=51; 
                 row-=51; 
                 break; 
            }
            if(feof(fp)) break;
          }
      }while(!feof(fp));
      fclose(fp);
      s.maxrow[col] = count+1;
      s.maxcol = max(col, s.maxcol);
      if(row >= s.rows){ message((char*)"Too many data points\nData truncated!",ERROR);  }
  }
  XtVaSetValues(g.main_widget, XmNtitle, filename, NULL); 
  g.getout = 0;
  for(row=0;row<count;row++)
      set_cell_value(row, col, s.x[col-startcol][row]);
}

//--------------------------------------------------------------------//
//  converttohex - do not use, only does 32 bit ints                  //
//  for gcd format files only                                         //
//--------------------------------------------------------------------//
void converttohex(uchar *buffer, uchar *hexbuffer)
{
   int k,c;
   char tempstring[256];
   tempstring[0] = 0;
   tempstring[1] = 0;
   tempstring[2] = 0;
   for(k=0;k<8;k++)
   {   c = buffer[k];
       sprintf(tempstring, "%x", c);
       hexbuffer[k*2] = tempstring[0];
       hexbuffer[k*2+1] = tempstring[1];
   }
}


//--------------------------------------------------------------------//
//  token_type                                                        //
//--------------------------------------------------------------------//
int token_type(char *buffer, double *value1, double *value2)
{
    char tempstring[64];
    int commapos=0;
    char *ptr;
    ptr = strchr(buffer,',');
    if(ptr) 
    {     commapos = (int)(ptr - buffer);
          strcpy(tempstring, buffer);
          tempstring[commapos] = 0;
          *value1 = atof(tempstring);
          strcpy(tempstring, buffer+commapos+1);
          *value2 = atof(tempstring);        
    }
    if(!strcmp(buffer,"TABLE")) return TABLE;
    if(!strcmp(buffer,"VECTORS")) return VECTORS;
    if(!strcmp(buffer,"TUPLES")) return TUPLES;
    if(!strcmp(buffer,"DATA")) return DATA;
    if(!strcmp(buffer,"COMMENT")) return COMMENT;
    if(!strcmp(buffer,"LABEL")) return LABEL;
    if(!strcmp(buffer,"SIZE")) return SIZE;
    if(!strcmp(buffer,"PERIODICITY")) return PERIODICITY;
    if(!strcmp(buffer,"MINORSTART")) return MINORSTART;
    if(!strcmp(buffer,"TRUELENGTH")) return TRUELENGTH;
    if(!strcmp(buffer,"DISPLAYUNITS")) return DISPLAYUNITS;
    if(!strcmp(buffer,"BOT")) return BOT;
    if(!strcmp(buffer,"EOD")) return EOD;
    if(!strcmp(buffer,"V")) return V;
    if(!strcmp(buffer,"NA")) return NA;
    if(!strcmp(buffer,"ERROR")) return ERROR;
    if(!strcmp(buffer,"TRUE")) return TRUE_VALUE;
    if(!strcmp(buffer,"FALSE")) return FALSE_VALUE;
    return NONE;
}


//--------------------------------------------------------------------//
// add_value                                                          //
//--------------------------------------------------------------------//
void add_value(double value, int row, int col)
{
   set_cell_value(row, col, value);
}


//--------------------------------------------------------------------//
// reformat_dif_table                                                 //
// repair data from Axxiom DIF file where data points are erroneously //
// split into 9 vectors.                                              //
//--------------------------------------------------------------------//
void reformat_dif_table(char **arg, int noofargs)
{
/*
    int row,j2,col,n;
    double *array;
    noofargs = 3;
    n = s.maxrow[1]; 
    array = new double [n*8+8];
    sprintf(arg[1], "1");
    sprintf(arg[2], "%d", n * 8);    
    shrink(arg, noofargs);
    j2 = 0;
    for(row=1; row<n; row++)
    for(col=1; col<9; col++)              // read across columns first
         array[j2++] = s.x[col][row]; 
    for(row=1; row<=n*8; row++)
         set_cell_value(row, 1, array[row]);
    for(col=2; col<9; col++)
    {    sprintf(arg[1], "%d", col+1);
         delete_column(arg, 1);
    }
    delete[] array;
}
*/
    int row,j2,col,nn;
    double *array;
    noofargs = 3;
    nn = s.maxrow[1]; 
    array = new double [nn*8+8];
    sprintf(arg[1], "1");
    sprintf(arg[2], "%d", nn * 8);    
    shrink(arg, noofargs);

    for(row=0,j2=0; row<=nn; row++)
    {    for(col=1; col<9; col++)              // read across columns first
             array[j2++] = s.x[col+1][row+1]; 
    }
    for(row=0; row<=nn*8; row++)
         set_cell_value(row+1, 2, array[row]);
    for(col=2; col<9; col++)
    {    sprintf(arg[1], "%d", col+1);
         delete_column(arg, 1);
    }
    delete[] array;
    s.maxrow[2] = s.maxrow[1];
}



//--------------------------------------------------------------------//
//  print_spreadsheet                                                 //
//  print on printer                                                  //
//--------------------------------------------------------------------//
void print_spreadsheet(void)
{ 
    Dialog dialog;
    int endrow=0;
    int k;
    for(k=0;k<s.cols;k++) endrow = max(endrow, s.maxrow[k]); 
    dialog.needrepeat = 0;  
    strcpy(dialog.title,"Print");
    dialog.radiono[0]=0;
    for(k=0;k<2;k++) dialog.boxset[k]=0;

    strcpy(dialog.boxes[0]," ");
    strcpy(dialog.boxes[1],"Printer");
    strcpy(dialog.boxes[2],"Print to File");
    dialog.boxtype[0]=LABEL;
    dialog.boxtype[1]=TOGGLESTRING;
        strcpy(dialog.answer[1], g.printer_command);
    dialog.boxtype[2]=TOGGLESTRING; 
        strcpy(dialog.answer[2], g.printer_filename);
    if(want_print_to_file) dialog.boxset[2]=1; else dialog.boxset[1]=1;

    dialog.noofradios=0;
    dialog.noofboxes=3;
    dialog.helptopic=0;
    dialog.keep=0;
    dialog.needrepeat=0;
    dialog.want_changecicb = 0;
    dialog.f1 = printcheck;
    dialog.transient = 1;
    strcpy(dialog.path, ".");

    ///////---///////
    dialogbox(dialog);
    ///////---///////

    do_print(&dialog);
    return;
}


//--------------------------------------------------------------------------//
//  printcheck                                                              //
//--------------------------------------------------------------------------//
void printcheck(dialoginfo *a, int radio, int radiobuttion, int box, int boxbutton)
{
    int k,wantbox;
    if(box == -1) 
    {    if(want_print_to_file) wantbox=2; else wantbox=1; }
    else wantbox = box;
    for(k=1; k<=2; k++) if(k!=wantbox)
    {    a->boxset[k] = 0;
         XmToggleButtonSetState(a->boxwidget[k][0],a->boxset[k],False);
    }
    a->boxset[wantbox] = 1;
}


//--------------------------------------------------------------------------//
//  do_print                                                                //
//--------------------------------------------------------------------------//
void do_print(dialoginfo *a)
{
    int col,row,status=0;
    int maxwidth=26;
    FILE *fp;
    char filename[FILENAMELENGTH] = "~temp.dat";
    char tempstring[2*FILENAMELENGTH];
    char errstring[4*FILENAMELENGTH];
    want_print_to_file = a->boxset[2]; 
    strcpy(g.printer_command, a->answer[1]);
    strcpy(g.printer_filename, a->answer[2]);

    if(want_print_to_file) strcpy(filename, g.printer_filename);
    if(cant_overwrite_file(filename)) return;

    if((fp=fopen(filename,"w")) == NULL)
    error_message(filename, errno);

    for(col=1; col<=s.maxcol; col++)
    for(row=0; row<=s.maxrow[col]; row++)
    {    if(s.string[col][row]) maxwidth=max(maxwidth, (int)strlen(s.text[col][row]));
    }

  
    for(col=1; col<=s.maxcol; col++)
         fprintf(fp,"%14s ", s.collabel[col]);
    fprintf(fp, "\n");
    for(row=0; row<s.maxrow[1]; row++)
    {   for(col=1; col<=s.maxcol; col++)
        {    if(s.string[col][row] || s.hex[col][row])
             {   ////  %*s = specify field width in a variable
                 fprintf(fp, "%*s", maxwidth, s.text[col][row]);
             }
             else
             {
                 fprintf(fp, "%14.14s ",s.text[col][row]);
             }
        }
        fprintf(fp, "\n");
    }
    fclose(fp);

    if(!want_print_to_file)
    {    sprintf(tempstring, "%s %s", g.printer_command, filename);
         printf("Executing %s\n",tempstring);
         status = system(tempstring);
         if(status)
         {    sprintf(errstring, "The command\n%s\nreturned error %d", tempstring, status);
              message(errstring, ERROR);
         }
         unlink("~temp.dat");
    }
}

//--------------------------------------------------------------------------//
// swapbytes                                                                //
// reverses the order of bytes  n=no. of bytes in array,size=bytes at a time//
//--------------------------------------------------------------------------//
void swapbytes(uchar* a, int n, int size)
{
   int k;
   switch(size)
   {   case 1: return;           // Too small, throw it back
       case 2:
          for(k=0;k<n-1;k+=2)    // For 16 bit ints
             swap(a[k],a[k+1]);
          break;
       case 3: 
          for(k=0;k<n-2;k+=3)    // For 24 bit pixels
               swap(a[k], a[k+2]);
          break;
       case 4:
          for(k=0;k<=n-4;k+=4)    // For 32 bit ints
          {     
                swap(a[k],a[k+3]);
#ifndef LITTLE_ENDIAN
                a[k+3] = 0;
#endif
                swap(a[k+1],a[k+2]);
          }
          break;
       case 6:
          for(k=0;k<=n-6;k+=6)    // For 48 bit ints
          {     swap(a[k],a[k+5]);
                swap(a[k+1],a[k+4]);
                swap(a[k+2],a[k+3]);
          } 
          break;
       case 8:
          for(k=0;k<=n-8;k+=8)    // For 64 bit ints
          {     swap(a[k],a[k+7]);
                swap(a[k+1],a[k+6]);
                swap(a[k+2],a[k+5]);
                swap(a[k+3],a[k+4]);
          } 
          break;
       default:
          fprintf(stderr,"Error in swapbytes (size=%d)\n",size);
   }
}

//--------------------------------------------------------------------------//
// delimiter_guess                                                          //
//--------------------------------------------------------------------------//
char delimiter_guess(char* filename)
{
  FILE *fp;
  int size = 200 * s.cols;
  int k;
  int space_count=0;
  int comma_count=0;
  int other_count=0;
  char other=' ';
  char delimiter=' ';
  char buffer[size];
  if((fp=fopen(filename,"rt")) == NULL)
      return(' ');
  else 
  {   do
      {     fgets(buffer, size, fp);
            if(feof(fp)) break;
            if(!strlen(buffer)) continue;
            if(buffer[0] == '#')
            {     continue;
            }else
            {     for(k=0; k<(int)strlen(buffer); k++)
                  {
                     if(buffer[k] == 10) continue;
                     if(buffer[k] == ',') comma_count++;
                     else if(buffer[k] == ' ') space_count++;
                  }
            }

      }while(!feof(fp));
      fclose(fp);
   }
   delimiter = ' ';
   if(comma_count>space_count) delimiter=',';
   if(other_count>space_count && other_count>comma_count) delimiter=other;   
   printf("delimiter is '%c'\n",delimiter);
   return(delimiter);
}


//--------------------------------------------------------------------//
// reformatmassspec                                                   //
// converts each datapoint graph to a vertical line                   //
// input: col.1=list of masses, col.2 = list of intensities           //
// output: expanded lists col. 1&2 suitable for graphing software     //
//--------------------------------------------------------------------//
void reformatmassspec(char **arg, int noofargs)
{
    int row,n;
    n = s.maxrow[0]; 
    double *array;
    array = new double [n*3];
    for(row=0; row<n; row++)
    {    array[row*3+0] = s.x[0][row]-0.00001;
         array[row*3+1] = s.x[0][row];
         array[row*3+2] = s.x[0][row]+0.00001;
    }
    for(row=0; row<n*3; row++)
         s.x[0][row] = array[row];
    for(row=0; row<n; row++)
    {    array[row*3+0] = 0;
         array[row*3+1] = s.x[1][row];
         array[row*3+2] = 0;
    }
    for(row=0; row<n*3; row++)
         s.x[1][row] = array[row];
    s.maxrow[0] = n*3;
    s.maxrow[1] = n*3;
    delete[] array;

    for(row=0; row<3*n; row++)
    {  s.number[0][row] = 1;
       s.string[0][row] = 0;
       s.hex[0][row] = 0;
       s.number[1][row] = 1;
       s.string[1][row] = 0;
       s.hex[1][row] = 0;
    }

    refreshspreadsheet();
}


//--------------------------------------------------------------------//
// findintegerdifferences                                             //
// for hdx - creates new file showing masses & intensities that differ//
// by integer values                                                  //
// input : spreadsheet created by findmass                            //
//         col 1 = filename                                           //
//         col 2 = observed mass                                      //
//         col 3 = matched mass                                       //
//         col 4 = intensity                                          //
//         col 5 = no of peptides                                     //
//         col 6 = sequence                                           //
// output: spreadsheet of masses matching m/z + integer +/- 0.02      //
//         where integer = 1,2, or 3 from deuteriums                  //
//         spreadsheet is written to a file, can be sorted by mass    //
//         each number of deuteriums is put in a different file       //
//         named hdx1.txt, hdx2.txt, etc.                             //
//         col 1 = sequence                                           //
//         col 2 = lower mass                                         //
//         col 3 = higher mass                                        //
//         col 4 = lower intensity                                    //
//         col 5 = higher intensity                                   //
//         col 6 = number of deuteriums                               //
//         col 7 = number of peptides found by turbosequest           //
//         col 8 = percent deuterium                                  //
//--------------------------------------------------------------------//
void findintegerdifferences(char **arg, int noofargs)
{
   char filename[1024] = "";
   char tempstring[2*FILENAMELENGTH] = "";
   char sequence[128] = "";
   FILE *fp;

   double threshold = 0.02;
   int maxdeuterium = 3;
   int col_obs = 1; 
   int col_mat = 2;
   int col_int = 3;
   int col_pep = 4;
   int col_seq = 5;
   int peptides = 0;
   
   int k,row,n,m;
   n = s.maxrow[0]; 
   double mass1, mass2, intensity1, intensity2;   


   //// Separate them into m=1, 2, 3, etc
   for(m=1; m<=maxdeuterium; m++)
   {  
       sprintf(filename, "hdx%d.txt", m);
       if ((fp=fopen(filename,"wt")) == NULL)
            error_message(filename, errno);
       else
    
       fprintf(fp, "## Masses with integer differences, threshold=%g \n",threshold);
       fprintf(fp, "#Sequence\n#Obs.mass1\n#Matched mass2\n#Intensity1\n#Intensity2\n#Delta m\n#No.peptides\n#PctDeut\n");

       for(row=0; row<n; row++)
       {   mass1      = s.x[col_obs][row];
           intensity1 = s.x[col_int][row];
           strcpy(sequence, s.text[col_seq][row]);
           for(k=row; k<n; k++)
           {  
                mass2      = s.x[col_mat][k];
                intensity2 = s.x[col_int][k];
                peptides   = s.x[col_pep][k];
                //// observed must be higher than matched by 1,2, or 3
                if((fabs(-mass1 + mass2 + (double )m) < threshold) && (intensity1 != 0.0))
                     fprintf(fp, "\"%s\"  %g  %g  %g  %g   %d %d  %g\n",
                     sequence, mass1, mass2, intensity1, intensity2, m, peptides,
                     intensity2/(intensity2+intensity1));
           }
       }
       sprintf(tempstring, "Results saved in\n%s", filename);
       fclose(fp);
   }

   message(tempstring);
}

//--------------------------------------------------------------------//
//  sselect = select a region using a command                         //
//--------------------------------------------------------------------//
void sselect(char **arg, int noofargs)
{
  static int ulcol=0;
  static int ulrow=0;
  static int lrcol=0;
  static int lrrow=0;
  Dialog dialog;
  int j,k,row,col;
  deselectallcells(arg, noofargs, 0);
  if(noofargs>=4) 
  {     ulrow = atoi(arg[1]); 
        ulcol = atoi(arg[2]); 
        lrrow = atoi(arg[3]); 
        lrcol = atoi(arg[4]); 
  }else
  {    
        dialog.radiono[0]=0;

        strcpy(dialog.boxes[0],"Region to select");
        strcpy(dialog.boxes[1],"Starting row");
        strcpy(dialog.boxes[2],"Starting column");
        strcpy(dialog.boxes[3],"Ending row");
        strcpy(dialog.boxes[4],"Ending column");

        dialog.boxtype[0]=LABEL;
        dialog.boxtype[1]=STRING; sprintf(dialog.answer[1], "%d", ulrow);
        dialog.boxtype[2]=STRING; sprintf(dialog.answer[2], "%d", ulcol);
        dialog.boxtype[3]=STRING; sprintf(dialog.answer[3], "%d", lrrow);
        dialog.boxtype[4]=STRING; sprintf(dialog.answer[4], "%d", lrcol);
        for(j=0;j<10;j++) for(k=0;k<20;k++) dialog.radiotype[j][k]=RADIO;
        for(j=0;j<10;j++) dialog.radiocomeback[j]=0;
        for(j=0;j<20;j++) dialog.boxcomeback[j]=0;
 
        dialog.noofradios=0;
        dialog.noofboxes=5;
        dialog.helptopic=0;
        dialog.keep=0;
        dialog.needrepeat=0;
        dialog.want_changecicb = 0;
        dialog.f1 = null;
        dialog.transient = 1;
        strcpy(dialog.path, ".");

        ///////---///////
        dialogbox(dialog);
        ///////---///////
 
        if(g.getout){ g.getout=0; return; }
        ulrow = atoi(dialog.answer[1]);
        ulcol = atoi(dialog.answer[2]);
        lrrow = atoi(dialog.answer[3]);
        lrcol = atoi(dialog.answer[4]);
  }
  for(row=ulrow; row<=lrrow; row++)
  for(col=ulcol; col<=lrcol; col++)
  {
      XbaeMatrixSelectCell(s.spreadsheet_widget, row, col);
      s.selected[col][row] = True;
  }
  refreshspreadsheet();
}
